require 'middleclass'

local curl = require 'curl'
local cjson = require 'cjson'

local M = class('db_client')

local function url_escape(s)
    return string.gsub(s, '([^A-Za-z0-9_])', function(c)
        return string.format('%%%02x', string.byte(c))
    end)
end

function M:initialize(settings)
    self:set(settings)

    self.cu = curl.easy_init()
    self.tx = nil
    self.rx = nil

    self.cu:setopt(curl.OPT_READFUNCTION,
        function(len)
            if len < string.len(self.tx) then
                local s = string.sub(self.tx, 1, len)
                self.tx = string.sub(self.tx, len + 1)
                return len, s
            else
                local s = self.tx
                self.tx = ''
                return string.len(s), s
            end
        end
    )
    self.cu:setopt(curl.OPT_WRITEFUNCTION,
        function(s, len)
            if len > 0 then
                if self.ofh then
                    self.ofh:write(s)
                else
                    self.rx[#self.rx + 1] = s
                end
            end
            return len, nil
        end
    )
end

function M:set(settings)
    for k, v in pairs(settings) do
        self[k] = v
    end
end

function M:http(url, opt)
    opt = opt or {}

    -- Default query stuff
    if self.query then
        opt.query = opt.query or {}
        for k, v in pairs(self.query) do
            if opt.query[k] == nil then
                opt.query[k] = v
            end
        end
    end

    if type(opt.data) == 'table' then
        opt.data = cjson.encode(opt.data)
    end
    self.tx = opt.data or ''

    local headers = {
    }
    local query = ''
    if opt.query then
        local t = {}
        for k, v in pairs(opt.query) do
            t[#t + 1] = url_escape(k) .. '=' .. url_escape(v)
        end
        query = '?' .. table.concat(t, '&')
    end
    local curl_opts = {
        OPT_URL = string.format('%s%s', url, query),
        OPT_TIMEOUT = 10,
        OPT_HTTPHEADER = headers,
        OPT_CUSTOMREQUEST = opt.method,
        OPT_SSL_VERIFYHOST = 0,
        OPT_SSL_VERIFYPEER = 0,
        OPT_UPLOAD = opt.data and 1 or 0,
    }

    if opt.data then
        curl_opts.OPT_INFILESIZE = string.len(self.tx)
    else
        headers[#headers + 1] = 'Content-Length: ' .. string.len(self.tx)
    end

    if opt.filename then -- KLUDGE allow Content-disposition???
        self.ofh = io.open(opt.filename, 'wb')
        curl_opts.OPT_TIMEOUT = 0 -- File transfers can be BIG
    end

    for k, v in pairs(opt.curl or {}) do
        curl_opts[k] = v
    end

    for k, v in pairs(curl_opts) do
        self.cu:setopt(curl[k], v)
    end

    self.rx = {}
    local got, err = self.cu:perform()
    if got ~= 0 then
        error(err)
    end

    if opt.filename and self.ofh then
        assert(self.ofh:close())
        self.ofh = nil
        -- KLUDGE check size here!!!
        return {ok=true}
    end

    local j = table.concat(self.rx, '')
    local v = nil
    pcall(function() v = cjson.decode(j) end)
    return v, j
end

function M:request(method, path, opt)
    opt = opt or {}
    opt.method = method
    local url = string.format('https://%s%s', self.host, path)
    return self:http(url, opt)
end

function M:raw_db(method, path, data)
    return self:request(method, '/api/raw_db/' .. path, {data=data})
end

function M:tracked_db(method, path, data)
    return self:request(method, '/api/tracked_db/' .. path, {data=data})
end

return M
