-- aux package
--
-- Provides an API to a VBM channel
--
-- API Functions:
--  aux:new(devname, firmware_path, master_freq)
--  aux:poll()
--  aux:ts_offset(ts)
--  aux:get_current_ts() [DEBUG]
--  aux:txdr_power_on()
--  aux:txdr_power_off()
--  aux:setup_vbm()
--  aux:setup_ecm(update_rate, start_time, initial_value, increment_value)
--  aux:setup_sync(clock_rate, start_time)
--  aux:setup_sync_ts()
--  aux:get_sync_ts()
--  aux:update_sync(static_data)
--  aux:rx_start(start_time, flush_interval, samples_per_bd, num_bds)
--  aux:rx_stop()
--  aux:tx_start(ring_buffer_id, msgs_per_bd, num_bds)
--  aux:tx_cmd_msg(ring_buffer_id, cmd_strings)
--  aux:allocate_msg_array(num_msgs)
--  aux:insert_msg(msg_array, index, ts, data_msg)
--  aux:tx_append_msgs(ring_buffer_id, msg_array, start_index)
--  aux:tx_stop(ring_buffer_id)
--  aux:event_setup(args)
--  aux:event_start(start_time, heartbeat_interval,
--                  samples_per_bd, num_bds)
--  aux:event_stop()
--  aux:get_cable_detect_state()
--
--
-- Notes on event_setup(...) parameters:
--   Set the following parameter names to true or non-zero to enable.
--       cable_unplugged
--       cable_plugged
--
--   For example: event_setup({cable_unplugged = true})
--   All parameters not listed are disabled by default.
--


local cc = require 'connctrl'
local ffi = require 'ffi'
local rb = require 'ring_buffers'
local ch = require 'channel'

local aux = setmetatable({}, { __index = ch })
local f = assert(io.open("/usr/include/fpga-firmware/vbm_structures.h", "r"))
local h = f:read("*all")
f:close()
ffi.cdef(h)


function aux:new(devname, firmware_path, master_freq)
    local aux_ch = self:new_channel(devname, firmware_path, "vbm_cmd_s")
    aux_ch.master_freq = master_freq
    aux_ch.tx_rb_msgs = {}
    return aux_ch
end

function aux:ts_offset(ts)
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_TS_OFFSET
    auxcmd.prm.ts_offset.ts_adjust = ts
    return self:send_command_buffer(auxcmd)
end

function aux:setup_sync_ts()
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_SETUP_SYNC_TS
    return self:send_command_buffer(auxcmd)
end

function aux:get_sync_ts()
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_GET_SYNC_TS
    return self:send_command_buffer(auxcmd)
end

function aux:get_current_ts()
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_GET_TS
    return self:send_command_buffer(auxcmd)
end

function aux:txdr_power_on()
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_TXDR_POWER_ON
    return self:send_command_buffer(auxcmd)
end

function aux:txdr_power_off()
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_TXDR_POWER_OFF
    return self:send_command_buffer(auxcmd)
end

function aux:setup_vbm()
    local smpl_freq_lo_speed = 115200 * 5     -- 5X oversampled clock required
    local smpl_freq_hi_speed = 1500000 * 5
    local smpl_freq_full_speed = 2000000 * 5
    local pwm_period = self.master_freq / 5000  -- Monitor INTR line at 5KHz
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_SETUP_VBM
    self:calculate_scg_parameters(auxcmd.prm.vbm.baud_rate_lo_speed,
                                  smpl_freq_lo_speed, 1, self.master_freq)
    self:calculate_scg_parameters(auxcmd.prm.vbm.baud_rate_hi_speed,
                                  smpl_freq_hi_speed, 1, self.master_freq)
    self:calculate_scg_parameters(auxcmd.prm.vbm.baud_rate_full_speed,
                                  smpl_freq_full_speed, 1, self.master_freq)
    self:calculate_scg_parameters(auxcmd.prm.vbm.pwm_monitor,
                                  pwm_period, 1, self.master_freq)
    auxcmd.prm.vbm.packet_timeout_ts_count = self.master_freq / 10;  -- 100ms
    return self:send_command_buffer(auxcmd)
end

function aux:setup_sync(clock_rate, start_time)
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_SETUP_SYNC
    if clock_rate == 0 and start_time == 0 then
        auxcmd.prm.sync.update_rate.base_period = 0
    else
        self:calculate_scg_parameters(auxcmd.prm.sync.update_rate,
                                      clock_rate, 1, self.master_freq)
    end
    auxcmd.prm.sync.start_time = start_time
    return self:send_command_buffer(auxcmd)
end

function aux:setup_ecm(update_rate, start_time, initial_value, increment_value)
    -- TODO: This code is NOT functional. Awaiting ECM Specification.
    local sync_pkt_baud_rate = 100000 * 5     -- 5X oversampled clock required
    local smpl_freq_lo_speed = 115200 * 5     -- 5X oversampled clock required
    local smpl_freq_hi_speed = 1500000 * 5
    local smpl_freq_full_speed = 2000000 * 5
    local pwm_period = self.master_freq / 5000  -- Monitor INTR line at 5KHz
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_SETUP_ECM
    -- VBM parameters
    self:calculate_scg_parameters(auxcmd.prm.ecm.vbm.baud_rate_lo_speed,
                                  smpl_freq_lo_speed, 1, self.master_freq)
    self:calculate_scg_parameters(auxcmd.prm.ecm.vbm.baud_rate_hi_speed,
                                  smpl_freq_hi_speed, 1, self.master_freq)
    self:calculate_scg_parameters(auxcmd.prm.ecm.vbm.baud_rate_full_speed,
                                  smpl_freq_full_speed, 1, self.master_freq)
    self:calculate_scg_parameters(auxcmd.prm.ecm.vbm.pwm_monitor,
                                  pwm_period, 1, self.master_freq)
    auxcmd.prm.ecm.vbm.packet_timeout_ts_count = self.master_freq / 10; -- 100ms
    -- SYNC parameters
    self:calculate_scg_parameters(auxcmd.prm.ecm.sync.baud_rate,
                                  sync_pkt_baud_rate, 1, self.master_freq)
    self:calculate_scg_parameters(auxcmd.prm.ecm.sync.update_rate,
                                  update_rate, 1, self.master_freq)
    auxcmd.prm.ecm.sync.start_time = start_time
    auxcmd.prm.ecm.sync.initial_value = initial_value
    auxcmd.prm.ecm.sync.increment_value = increment_value
    return self:send_command_buffer(auxcmd)
end

function aux:update_sync(static_data)
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_UPDATE_SYNC
    local byte_count = 0
    for i,v in ipairs(static_data) do
        auxcmd.prm.update.static_data[i - 1] = tonumber(v)
        byte_count = byte_count + 4
        if i == 6 then break end
    end
    auxcmd.prm.update.valid_bytes = byte_count
    return self:send_command_buffer(auxcmd)
end

function aux:rx_start(start_time, flush_interval, samples_per_bd, num_bds)
    local pa = self:add_ring_buffer('RX',
                                    "vbm_rxtx_buffer_t["..samples_per_bd.."]",
                                    "vbm_rxtx_buffer_t", num_bds)
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_RX_START
    auxcmd.prm.rx_start.phys_address = pa
    auxcmd.prm.rx_start.start_time = start_time
    self:calculate_scg_parameters(auxcmd.prm.rx_start.flush_interval,
                                  flush_interval, 1, self.master_freq)
    self:calculate_scg_parameters(auxcmd.prm.rx_start.vbm_clk_rate,
                                  20000, 1, self.master_freq)
    return self:send_command_buffer(auxcmd)
end

function aux:rx_stop()
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_RX_STOP
    return self:send_command_buffer(auxcmd)
end

function aux:tx_start(ring_buffer_id, msgs_per_bd, num_bds)
    local pa = self:add_ring_buffer('TX'..ring_buffer_id,
                                    "vbm_rxtx_buffer_t["..msgs_per_bd.."]",
                                    "vbm_rxtx_buffer_t", num_bds)
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_TX_START
    auxcmd.prm.tx_start_stop.phys_address = pa
    auxcmd.prm.tx_start_stop.channel_id = ring_buffer_id
    self.tx_rb_msgs['TX'..ring_buffer_id] = msgs_per_bd
    return self:send_command_buffer(auxcmd)
end

function aux:tx_cmd_msg(ring_buffer_id, cmd_strings)
    if type(cmd_strings) == 'string' then
        local cmd_string = cmd_strings
        cmd_strings = {}
        cmd_strings[1] = cmd_string
    end
    local msg_array = ffi.new("vbm_rxtx_buffer_t["..#cmd_strings.."]")
    for i=1,#cmd_strings do
        local bytes_to_copy = 64
        if #cmd_strings[i] < 64 then bytes_to_copy = #cmd_strings[i] end
        ffi.copy(msg_array[i - 1].msg, cmd_strings[i], bytes_to_copy)
        msg_array[i - 1].ts = 0
        msg_array[i - 1].is_control_frame = 1;
        msg_array[i - 1].msg_length = bytes_to_copy;
    end
    self:tx_append_msgs(ring_buffer_id, msg_array)
end

function aux:allocate_msg_array(num_msgs)
    return ffi.new("vbm_rxtx_buffer_t["..num_msgs.."]")
end

function aux:insert_msg(msg_array, index, ts, data_msg)
    local total_msgs = ffi.sizeof(msg_array) / ffi.sizeof("vbm_rxtx_buffer_t")
    assert(index > 0 and index <= total_msgs)
    assert(#data_msg <= 60)
    ffi.copy(msg_array[index - 1].msg + 4, data_msg, #data_msg)
    msg_array[index - 1].ts = ts
    msg_array[index - 1].is_control_frame = 0;
    msg_array[index - 1].msg_length = #data_msg + 4;
end

function aux:tx_append_msgs(ring_buffer_id, msg_array, start_index)
    local rb_name = 'TX'..ring_buffer_id
    local msgs_per_bd = self.tx_rb_msgs[rb_name]
    local total_msgs = ffi.sizeof(msg_array) / ffi.sizeof("vbm_rxtx_buffer_t")
    local msgs_queued = start_index or 0
    total_msgs = total_msgs - msgs_queued
    while total_msgs > 0 do
        local msgs_in_buffer
        if total_msgs > msgs_per_bd then
            msgs_in_buffer = msgs_per_bd
        else
            msgs_in_buffer = total_msgs
        end
        local buffer_size = msgs_in_buffer * ffi.sizeof("vbm_rxtx_buffer_t")
        local ptr = ffi.cast("vbm_rxtx_buffer_t *", msg_array) + msgs_queued
        if self:send_buffer(ptr, rb_name, buffer_size) then
            total_msgs = total_msgs - msgs_in_buffer
            msgs_queued = msgs_queued + msgs_in_buffer
        else
            return msgs_queued
        end
    end
    return msgs_queued
end

function aux:tx_stop(ring_buffer_id)
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_TX_STOP
    auxcmd.prm.tx_start_stop.channel_id = ring_buffer_id
    return self:send_command_buffer(auxcmd)
end

function aux:event_setup(arg)
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_EVENT_SETUP
    if arg.cable_unplugged then auxcmd.prm.events.cable_unplugged = 1 end
    if arg.cable_plugged then auxcmd.prm.events.cable_plugged = 1 end
    if arg.vbm_unplugged then auxcmd.prm.events.vbm_unplugged = 1 end
    if arg.vbm_plugged then auxcmd.prm.events.vbm_plugged = 1 end
    return self:send_command_buffer(auxcmd)
end

function aux:event_start(start_time, heartbeat_interval,
                         samples_per_bd, num_bds)
    local pa = self:add_ring_buffer('EV',
                                    "vbm_ev_buffer_t["..samples_per_bd.."]",
                                    "vbm_ev_buffer_t", num_bds)
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_EVENT_START
    auxcmd.prm.ev_start.phys_address = pa
    auxcmd.prm.ev_start.start_time = start_time
    self:calculate_scg_parameters(auxcmd.prm.ev_start.heartbeat_interval,
                                  heartbeat_interval, 1, self.master_freq)
    return self:send_command_buffer(auxcmd)
end

function aux:event_stop()
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_EVENT_STOP
    return self:send_command_buffer(auxcmd)
end

function aux:get_cable_detect_state()
    local auxcmd = ffi.new("vbm_cmd_s")
    auxcmd.cmd = ffi.C.VBM_GET_CABLE_SENSE
    return self:send_command_buffer(auxcmd)
end

return aux
