--[[
DIO Proxy Documentation

Channels
    Channels available for each input signal, 1-3, and (Logical Unit)
        Digital Input (Input Capture)
        Clamped Input Low Peak Voltage (Hysteresis Control)
        Clamped Input High Peak Voltage (Hysteresis Control)
        Low Threshold Voltage (Hysteresis Control)
        High Threshold Voltage (Hysteresis Control)
        Clamped Input Voltage (Hysteresis Control)
        Raw Pulse Period, Low (Pulse Counter)
        Raw Pulse Period, High (Pulse Counter)
        Raw Cumulative Pulse Period, Low (Pulse Counter)
        Raw Cumulative Pulse Period, High (Pulse Counter)
        Raw Pulse Count, Low (Pulse Counter)
        Raw Pulse Count, High (Pulse Counter)
        Pulse Period (Pulse Counter)
        Pulse Frequency (Pulse Counter)
        Pulse Duty Cycle (Pulse Counter)
        Pulse Rate (Pulse Counter)

    Channels available for the entire connector
        Quadrature Count (Quadrature Counter)
        Quadrature Latch Index Count (Quadrature Counter)
        Quadrature Direction (Quadrature Counter)
        Quadrature Latch Status (Quadrature Counter)

Channel Parameters
    Channel Parameters
        start_time                      Default now + 1 second
        sample_rate                     Default 20 sps
        transducer_power                Default Off
        threshold_lo                    Default 800 mV
        threshold_hi                    Default 1800 mV
        threshold_initial_lo            Default 800 mV
        threshold_initial_hi            Default 1800 mV
        threshold_mode                  Default 'static'
        threshold_sample_rate           Default 100 sps
        threshold_decay_rate_mv         Default 1 mV
        threshold_decay_divider         Default 2
        threshold_dac_update_divider    Default 2
        threshold_ratio_lo2hi           Default 0.6875
        threshold_ratio_hi2lo           Default 0.3125

        NOTE:
            threshold_low_mv is converted to threshold_lo
            threshold_high_mv is converted to threshold_hi
            threshold_initial_low_mv is converted to initial_threshold_lo
            threshold_initial_high_mv is converted to initial_threshold_hi
            threshold_mode is converted to dynamic_mode
        The conversion is needed to match what the dio api is expecting.
        Once universal agreement is reached on parameter names and meanings,
        the dio api can be updated to minimize or eliminate conversions.

    Quadrature Specific Channel Parameters
        invert_a                    Invert input A, default false
        invert_b                    Invert input B, default false
        enable_index                Use the 3rd input as an index, default true
        invert_index                Invert index, default false
        swap_a_b                    Swap A and B inputs, defaults false
        count_down_b_fall_a_one     Count down on FE of B with A=1, default true
        count_down_b_rise_a_zero    Count down on RE of B with A=0, default true
        count_down_a_fall_b_zero    Count down on FE of A with B=0, default true
        count_down_a_rise_b_one     Count down on RE of A with B=1, default true
        count_up_b_fall_a_zero      Count up on FE of B with A=0, default true
        count_up_b_rise_a_one       Count up on RE of B with A=1, default true
        count_up_a_fall_b_one       Count up on FE of A with B=1, default true
        count_up_a_rise_b_zero      Count up on RE of A with B=0, default true

    Parameter Notes
        Unspecified parameters will be defaulted in the proxy.

        The threshold parameters for quadrature channels apply to ALL of the
        input signals.

Logical Units
    There are 10 logical units, 3 each of the Input Capture, Hystereis Control,
    and Pulse Counter (one logical unit for each input signal) and one
    Quadrature Counter logical unit.  Collecting data for one channel within
    a logical unit will make available all other channels within that logical
    unit, whether or not the data is used or streamed out a network socket.

Comet Streams
    If no data streams are established, each channel will run using
    default parameters with values as listed above unless provided during
    comet stream initialization.  To change a parameter, the comet stream
    must be stopped and restarted.
    If a stream is established, data will be provided at 10 sps regardless
     of channel setup (the comet stream update rate).

Data Streams
    Data streaming works pretty much as before. If a data stream connection is
    attempted, any parameter mismatch will result in an error.

-- ** DFD
Prestarting Data Flow
    The Pulse Counter channels do not measure instantaneous values of frequency and
    pulse rate for example. Following the eDAQ convention, the values of these parameters
    at Test Start "Time_0" are based on timing / counting information from the previous
    sample period. As such, the Pulse Counter channels must be prestarted at least one sample
    period before Test Start "Time_0". The current implementation ensures that this will be
    the case by starting 1 second before time 0 (in the Decimal DRD). This is based on 1 Hz
    being the lowest sample rate used to source data, and relies on the fact that all Decimal
    DRD sample rates are an integer mutiplier of 1 Hz. (The other DRDs are handled analogously).
-- ** DFD

Rules
    Cell Rules
        sample_rate                     Rational between 0 and 100000 (sps)
        threshold_low_mv                Integer between -5000 and 5000 (mv)
        threshold_high_mv               Integer between -5000 and 5000 (mv)
        threshold_initial_low_mv        Integer between -5000 and 5000 (mv)
        threshold_initial_high_mv       Integer between -5000 and 5000 (mv)
        threshold_mode                  'static' or 'dynamic'
        threshold_sample_rate           Rational between 0 and 100000 (sps)
        threshold_decay_rate_mv         Integer between -5000 and 5000 (mv)
        threshold_decay_divider         Integer between 1 and 100000
        threshold_dac_update_divider    Integer between 1 and 100000
        threshold_ratio_lo2hi           Floating point between 0 and 1
        threshold_ratio_hi2lo           Floating point between 0 and 1
    Intra-Row Rules
        threshold_ratio_hi2lo < threshold_ratio_lo2hi
        threshold_low_mv < threshold_high_mv
        threshold_initial_low_mv < threshold_initial_high_mv
        sample_rate == threshold_sample_rate for HC channels
    Inter-Row Rules
        All threshold parameters must match between channels associated
        with the same input signal.
        All sample_rate parameters must match between channels in the same
        logical unit.

    If a data stream connection is attempted that would violate the rules,
    the connection attempt will fail with an appropriate error message.

Dynamic Threshold Algorithm
    Related Channel Parameters
        threshold_low_mv
        threshold_high_mv
        threshold_initial_low_mv
        threshold_initial_high_mv
        threshold_mode
        threshold_sample_rate
        threshold_decay_rate_mv
        threshold_decay_divider
        threshold_dac_update_divider
        threshold_ratio_lo2hi
        threshold_ratio_hi2lo

    Algorithm Description
        The ADC runs at the threshold_sample_rate collecting data.  The high
        and low peaks are updated with each new ADC sample.  The
        threshold_decay_divider is the number of ADC samples that occur before
        the threshold_decay_rate value is applied to the peaks (subtracted
        from the high peak and added to the low peak).  The
        threshold_dac_update_divider is the number of ADC samples that occur
        before the threshold DACs are updated.  The threshold DACs are only
        updated if the threshold_mode parameter is 'dynamic'.  The initial
        state of the threshold DACs is specified by threshold_lo and
        threshold_hi.  If the threshold_mode parameter is 'static', the DACs
        are never changed from these initial values.  When an update occurs,
        the new DAC values are calculated by finding the difference between
        the high and low peaks and applying the threshold_ratio_* proportional
        values to arrive at new thresholds as such:

          peak_to_peak = high_peak - low_peak
          new_low_threshold = low_peak + (threshold_ratio_lo2hi * peak_to_peak)
          new_high_threshold = low_peak + (threshold_ratio_hi2lo * peak_to_peak)
]]

local __ = require '__'
local ev = require 'ev'
local ffi = require 'ffi'
local ts = require 'timestamps'
local u = require 'unix'
local sample_clock = require 'sample_clock'
local rational = require 'rational'
local cjson = require 'cjson.safe'

local dio = require 'dio'
local channel_proxy = require 'channel_proxy'
local cca_handler = require 'cca_handler'

local log = require 'log'

local si514_rate = sample_clock.master_rate

-- size of buffer used to read "set output port" requests
local OUTPUT_REQ_BUF_LEN = 128

local on = true
local off = false
local dio_events

local function dbg(fmt, ...)
    log:log(log.DEBUG, fmt, ...)
end

local function bits_to_mv(bits)
--    return math.floor((((bits / 65536) * 2.5) - 1.25) * 5000 + 0.5)
    return math.floor(bits / 5.24288 - 6249.5)
end

--- NOTE: Call this function only if lun.lun_type == 'pc'
local function preprocess_lun_data(lun, bufdata)
    if bufdata.period_count_lo > 0 then
        lun.current_period_lo = bufdata.period_count_lo
    else
        if bufdata.partial_period_count_lo > lun.current_period_lo then
            lun.current_period_lo = bufdata.partial_period_count_lo
        end
    end
    if bufdata.period_count_hi > 0 then
        lun.current_period_hi = bufdata.period_count_hi
    else
        if bufdata.partial_period_count_hi > lun.current_period_hi then
            lun.current_period_hi = bufdata.partial_period_count_hi
        end
    end
    lun.total_period = lun.current_period_lo + lun.current_period_hi
end

-- Digital Input Channels (FPGA controller per physical Input pin 1, 2, and 3)
local dio_input_channels = {
    { 'input', 'ic', 'Digital Input', 'bit',
        function(lun, index, bufdata)
            return bufdata.pin_state
        end, true },
--[[ NOTE: DIO dynamic threshold mode deferred
    { 'input_low_peak', 'hc', 'Clamped Input Low Peak Voltage', 'mV',
        function(lun, index, bufdata)
            return bits_to_mv(bufdata.lo_peak)
        end },
    { 'input_high_peak', 'hc', 'Clamped Input High Peak Voltage', 'mV',
        function(lun, index, bufdata)
            return bits_to_mv(bufdata.hi_peak)
        end },
    { 'threshold_low', 'hc', 'Low Threshold Voltage', 'mV',
        function(lun, index, bufdata)
            return bits_to_mv(bufdata.threshold_lo)
        end },
    { 'threshold_high', 'hc', 'High Threshold Voltage', 'mV',
        function(lun, index, bufdata)
            return bits_to_mv(bufdata.threshold_hi)
        end },
]]
    { 'clamped_input_level', 'hc', 'Clamped Input Voltage', 'mV',
        function(lun, index, bufdata)
            return bits_to_mv(bufdata.adc_data)
        end },
    { 'raw_clamped_input_level', 'hc', 'Raw Clamped Input ADC Reading', 'bits',
        function(lun, index, bufdata)
            return bufdata.adc_data
        end },
    { 'raw_period_low', 'pc', 'Raw Pulse Period (Low)', 'ticks',
        function(lun, index, bufdata)
            return bufdata.period_count_lo
        end },
    { 'raw_period_high', 'pc', 'Raw Pulse Period (High)', 'ticks',
        function(lun, index, bufdata)
            return bufdata.period_count_hi
        end },
    { 'raw_sum_period_low', 'pc', 'Raw Cumulative Pulse Period (Low)', 'ticks',
        function(lun, index, bufdata)
            return bufdata.avg_period_count_lo
        end },
    { 'raw_sum_period_high', 'pc', 'Raw Cumulative Pulse Period (High)', 'ticks',
        function(lun, index, bufdata)
            return bufdata.avg_period_count_hi
        end },
    { 'raw_partial_period_low', 'pc', 'Raw Partial Pulse Period (Low)', 'ticks',
        function(lun, index, bufdata)
            return bufdata.partial_period_count_lo
        end },
    { 'raw_partial_period_high', 'pc', 'Raw Partial Pulse Period (High)', 'ticks',
        function(lun, index, bufdata)
            return bufdata.partial_period_count_hi
        end },
    { 'raw_edge_count_low', 'pc', 'Raw Pulse Count (Low)', 'counts',
        function(lun, index, bufdata)
            return bufdata.edge_count_lo
        end },
    { 'raw_edge_count_high', 'pc', 'Raw Pulse Count (High)', 'counts',
        function(lun, index, bufdata)
            return bufdata.edge_count_hi
        end },
    { 'period', 'pc', 'Pulse Period', 'seconds',
        function(lun, index, bufdata)
            return lun.v_cpt[index] / si514_rate
        end },
    { 'frequency', 'pc', 'Pulse Frequency', 'Hz',
        function(lun, index, bufdata)
            return si514_rate / lun.v_cpt[index]
        end },
    { 'duty_cycle', 'pc', 'Pulse Duty Cycle', 'percent',
        function(lun, index, bufdata)
            return lun.v_cph[index] / lun.v_cpt[index] * 100
        end },
    { 'pulse_rate', 'pc', 'Pulse Rate', 'Hz',
        function(lun, index, bufdata)
            return bufdata.edge_count_hi * lun.spec.rate
        end }
}

-- Quadrature Channels (FPGA controller - A (Input 1), B (Input 2), Z (Input 3))
local dio_qc_channels = {
    { 'quadrature_count', 'qc', 'Quadrature Count', 'counts',
        function(lun, index, bufdata)
            return tonumber(bufdata.quadrature_count)
        end },
    --[[{ 'quadrature_latched_index_count', 'qc',
      'Quadrature Latch Index Count', 'counts',
        function(lun, index, bufdata)
            return tonumber(bufdata.latched_index_count)
        end },--]]
    { 'quadrature_direction', 'qc', 'Quadrature Direction', '',
        function(lun, index, bufdata)
            if bit.band(bufdata.direction_and_latch_status, 1) == 1 then
                return -1
            else
                return 1
            end
        end },
--[[
    { 'quadrature_latch_status', 'qc', 'Quadrature Latch Status', '',
        function(lun, index, bufdata)
            return bit.rshift(bufdata.direction_and_latch_status, 1)
        end },
]]
}

-- Digital Output Channels (Provided by Quadrature's FPGA controller)
local dio_oc_channels = {
    { 'output_1', 'oc', 'Digital Output 1', '',
        function(lun, bufdata)
            if dio_events.force_closed0 == 1 then
                return 0
            elseif dio_events.force_open0 == 1 then
                return 1
            else
                return -1
            end
        end },
    { 'output_2', 'oc', 'Digital Output 2', '',
        function(lun, bufdata)
            if dio_events.force_closed1 == 1 then
                return 0
            elseif dio_events.force_open1 == 1 then
                return 1
            else
                return -1
            end
        end },
}

-- Table for all local DIO Channels
local dio_channels = {}

do
    local o = { }
    for i = 1, 3 do
        for _, ch in ipairs(dio_input_channels) do
            o[#o + 1] = {
                name = ch[1]:sub(1,3) == 'raw' and ch[1]..'_'..i..'.status'
                                                or ch[1]..'_'..i,
                lun = ch[2]..i,
                description = ch[3],
                electrical_units = ch[4],
                bit = i - 1,
                cca_index = i,
                lun_type = ch[2],
                extractor = ch[5],
                is_logical = ch[6] or false
            }
            o[o[#o].name] = o[#o]
        end
    end
    for _, ch in ipairs(dio_qc_channels) do
        o[#o + 1] = {
            name = ch[1],
            lun = ch[2],
            description = ch[3],
            electrical_units = ch[4],
            cca_index = 4,
            lun_type = ch[2],
            extractor = ch[5],
            is_logical = ch[6] or false
        }
        o[o[#o].name] = o[#o]
    end
    for _, ch in ipairs(dio_oc_channels) do
        o[#o + 1] = {
            name = ch[1],
            lun = ch[2],
            description = ch[3],
            electrical_units = ch[4],
            cca_index = 4,
            lun_type = ch[2],
            extractor = ch[5]
        }
        o[o[#o].name] = o[#o]
    end
    dio_channels = o
end

local channel_threshold_defaults = {
    threshold_high_mv = 1800,
    threshold_low_mv = 800,
    threshold_initial_high_mv = 1800,
    threshold_initial_low_mv = 800,
    threshold_mode = 'static',
    threshold_sample_rate =  100,
    threshold_decay_rate_mv =  1,
    threshold_decay_divider = 2,
    threshold_dac_update_divider = 2,
    threshold_ratio_lo2hi = 0.6875,
    threshold_ratio_hi2lo = 0.3125,
}

local quadrature_defaults = {
    invert_a = false,
    invert_b = false,
    enable_index = true,
    invert_index = true,
    swap_a_b = false,
    count_down_b_fall_a_one = true,
    count_down_b_rise_a_zero = true,
    count_down_a_fall_b_zero = true,
    count_down_a_rise_b_one = true,
    count_up_b_fall_a_zero = true,
    count_up_b_rise_a_one = true,
    count_up_a_fall_b_one = true,
    count_up_a_rise_b_zero = true,
}

local function conv_mv_to_v(v) return tonumber(v) / 1000 end

local hyst_args_parameter_translations = {
    threshold_high_mv = {'threshold_hi', conv_mv_to_v },
    threshold_low_mv = {'threshold_lo', conv_mv_to_v },
    threshold_initial_high_mv = {'initial_threshold_hi', conv_mv_to_v },
    threshold_initial_low_mv = {'initial_threshold_lo', conv_mv_to_v },
    threshold_mode = {'dynamic_mode', function(v) return v == 'dynamic' end },
    threshold_decay_rate_mv =  {'decay_rate', conv_mv_to_v },
    threshold_decay_divider = 'decay_divider',
    threshold_dac_update_divider = 'dac_update_divider',
    threshold_ratio_lo2hi = 'threshold_ratio_lo2hi',
    threshold_ratio_hi2lo = 'threshold_ratio_hi2lo',
}

local parameter_translations = {}
parameter_translations['hc'] = hyst_args_parameter_translations
parameter_translations['ic'] = {}   -- No parameters to translate/copy
parameter_translations['pc'] = {}   -- No parameters to translate/copy
parameter_translations['qc'] = {}
for k,_ in pairs(quadrature_defaults) do
    parameter_translations['qc'][k] = k    -- No translations/just copies
end

local dio_proxy = class('dio_proxy', channel_proxy)

function dio_proxy:initialize(name, devname)
    self.proxy_type = 'dio'
    self.controllers = {}
    local base_dev_num = devname:match('/dev/cca(%d+)')
    for i = 1,4 do
        local dn = string.format('/dev/cca%d', base_dev_num + i - 1)
        self.controllers[i] = {}
        local fw = 'DIO_' .. (i - 1) .. '.bin'
        if i == 4 then fw = 'DIO_QC_0.bin' end
        cca_handler.new(dn, self.controllers[i], dio, fw, si514_rate)
        self.controllers[i].buffer = function(proxy, msg) self:buffer(msg) end
    end

    self.streaming = false

    -- ** Following sized for worst case of buffer size of 512 based on <samples_per_bd>
    -- ** parameter defined in function dio_proxy:start_lun().
    self.vector_data = ffi.new('double[512]')
    self.vector_cpl = ffi.new('double[512]')
    self.vector_cph = ffi.new('double[512]')
    self.vector_cpt = ffi.new('double[512]')

    self.luns = {}
    self.value_funcs = {}
    self.txdr_power = 'off'

    for _,ch in ipairs(dio_channels) do
        if not self.luns[ch.lun] then
            self.luns[ch.lun] = { data_sinks = {},
                                  lun_type = ch.lun_type,
                                  name = ch.lun,
                                  bit = ch.bit,
                                  ccadev = self.controllers[ch.cca_index],
                                  cca_index = ch.cca_index,
                                  samples_to_discard = 0,
                                  v_cpl = self.vector_cpl,
                                  v_cph = self.vector_cph,
                                  v_cpt = self.vector_cpt,
                                  current_period_lo = 0,
                                  current_period_hi = 0,
                                  total_period = 0 }
        end
        self.value_funcs[ch.name] = ch.extractor
    end
    self.lun_lookup = {}

    channel_proxy.initialize(self, name)

end

function dio_proxy:halt(dump_trace)
    for i = 1,4 do self.controllers[i]:halt(dump_trace) end
end

function dio_proxy:buffer(msg)
    if msg.cmd == ffi.C.DIO_EVENT_START then
        local last_element = msg.element_count - 1
        local buf = msg.bufdata[0]
        local timestamp = buf.ts
        if not timestamp then return end
        local events = buf.events
        if not dio_events then
            dio_events = ffi.new('dio_cmd_event_vector_s')
        end
        -- NOTE: Add actions to the corresponding connector events
        if events.cable_unplugged == 1
                and dio_events.cable_unplugged ~= 1 then
            log:notice(__"DIO: Cable Unplugged")
        end
        if events.cable_plugged == 1
                and dio_events.cable_plugged ~= 1 then
            log:notice(__"DIO: Cable Plugged")
        end
        if events.force_closed0 == 1
                and dio_events.force_closed0 ~= 1 then
            -- output 1 is forced off
        end
        if events.force_closed1 == 1
                and dio_events.force_closed1 ~= 1 then
            -- output 2 is forced off
        end
        if events.force_open0 == 1
                and dio_events.force_open0 ~= 1 then
            -- output 1 is forced on
        end
        if events.force_open1 == 1
                and dio_events.force_open1 ~= 1 then
            -- output 2 is forced on
        end
        dio_events = events
        return
    end

    local lun = assert(self.lun_lookup[msg.pa])

    if not self.streaming then
        -- Idle
        return
    end

    lun.bufdata, lun.count = msg.bufdata, msg.element_count

    if lun.lun_type == 'pc' then
        for i=0,lun.count-1 do
            preprocess_lun_data(lun, lun.bufdata[i])
            self.vector_cpl[i] = lun.current_period_lo
            self.vector_cph[i] = lun.current_period_hi
            self.vector_cpt[i] = lun.total_period
        end
    end

    local sink_count = lun.count
    local sink_discards = lun.samples_to_discard
    if sink_discards > 0 then
        sink_discards = lun.samples_to_discard
        if sink_discards >= lun.count then
            sink_discards = lun.count
            sink_count = 0
        else
            sink_count = lun.count - sink_discards
        end
        lun.samples_to_discard = lun.samples_to_discard - sink_discards
    end

    lun.timestamp = lun.timestamp + sink_count
    for ch,ch_sinks in pairs(lun.data_sinks) do
        local value_func = assert(self.value_funcs[ch])
        for i=0,lun.count-1 do
            self.vector_data[i] = value_func(lun, i, lun.bufdata[i])
        end
        if sink_count then
            for _,sink in pairs(ch_sinks) do
                sink(lun.timestamp, self.vector_data + sink_discards, sink_count * 8)
            end
        end
    end
end

function dio_proxy:available_channels(hostname)
    local category
    local data_type
    local is_logical
    local chans = {}
    for _, ch in ipairs(dio_channels) do
        if ch.lun_type == 'oc' then
            category = 'sink/hbm_cx23r/digital_output'
            data_type = nil
            is_logical = false
        else
            category = 'source/hbm_cx23r/dio_input'
            data_type = 'sequential_float64'
            is_logical = ch.is_logical
        end
        table.insert(chans, {
            name = ch.name..'_'..hostname:gsub('[%.%-]', '_'),
            connector = ch.name..'@'..hostname,
            description = ch.description,
            electrical_units = ch.electrical_units,
            data_type = data_type,
            sample_rate = 100,
            category = category,
            is_logical = is_logical
        })
    end
    table.insert(chans, {
        name = 'control',
        connector = 'control.status@'..hostname,
        description = 'DIO Control Channel',
        data_type = 'message_raw',
        category = 'source/hbm_cx23r/dio_input',
    })
    self.channels = chans
    return chans
end

function dio_proxy:start_lun(lun, spec, md_sink)

    -- Fill in defaults for anything missing in 'spec'
    if not spec.start_ts then
        if spec.start_time then
            spec.start_ts = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
        else
            log:error(__"DIO: Start Time not provided for {0}", spec.connector)
        end
    end

    local prestart_ts = math.floor(0.5 + (spec.start_ts - self.start_offset))

    spec.sample_rate = spec.sample_rate or 20
    spec.rate, spec.rate_num, spec.rate_denom =
            rational:new(spec.sample_rate):tonumbers()
    spec.data_type = (not spec.category:match("output") and
        (spec.data_type or 'sequential_float64') or nil)

    lun.samples_to_discard = math.floor(0.5 + (self.prestart_seconds * spec.rate))

    -- Add extra stuff of dubious value...
    spec._timescale = spec.rate_denom..'/'..spec.rate_num

    -- Convert threshold parameters from table form to flat form
    if type(spec.threshold) == 'table' then
        spec.threshold_high_mv = spec.threshold.max
        spec.threshold_low_mv = spec.threshold.min
        spec.threshold_mode = spec.threshold.mode
    end

    -- Lun specific defaults
    for k,v in pairs(channel_threshold_defaults) do
        spec[k] = spec[k] or v
        spec[k] = spec[k] == '' and v or spec[k]
    end
    if lun.lun_type == 'qc' then
        for k,v in pairs(quadrature_defaults) do
            spec[k] = spec[k] or v
        end
    end

    -- If the data flow has already started, decide what to do
    if lun.spec then
        -- Nasty comparison, spec vs. lun.spec
        if not (spec.rate_num == lun.spec.rate_num and
                spec.rate_denom == lun.spec.rate_denom) then
            spec.error_message = 'Sample rate mismatch'
            if md_sink then md_sink(spec) end
            return
        end
        if spec.start_ts ~= lun.spec.start_ts then
            spec.error_message = 'Start time mismatch'
            if md_sink then md_sink(spec) end
            return
        end
        for k,v in pairs(channel_threshold_defaults) do
            if spec[k] ~= lun.spec[k] then
                spec.error_message = 'Threshold setting mismatch'
                if md_sink then md_sink(spec) end
                return
            end
        end
        -- All matched up, LUN already running
        if md_sink then md_sink(spec) end
        return
    end

    -- Before we start this lun, if it's not the hysteresis control lun,
    -- start that lun up first.  If it fails, we fail.
    if lun.lun_type ~= 'hc' then
        local hyst_spec = {}
        local hyst_lun
        for k,_ in pairs(channel_threshold_defaults) do
            hyst_spec[k] = spec[k]
        end
        hyst_spec.start_ts = spec.start_ts
        hyst_spec.sample_rate = spec.threshold_sample_rate
        if lun.lun_type ~= 'qc' then
            local new_name = lun.name:gsub(lun.lun_type, 'hc')
            hyst_lun = { new_name }
        else
            if spec.enable_index then
                hyst_lun = { 'hc1', 'hc2', 'hc3' }
            else
                hyst_lun = { 'hc1', 'hc2' }
            end
        end

        for _,l in ipairs(hyst_lun) do
            local lun = assert(self.luns[l])
            if hyst_spec.threshold_mode == 'static' then
                lun.ccadev:cmd_yield('hc_set_thresholds',
                                     conv_mv_to_v(hyst_spec.threshold_high_mv),
                                     conv_mv_to_v(hyst_spec.threshold_low_mv))
            elseif hyst_spec.threshold_mode == 'dynamic' then
                self:start_lun(lun, hyst_spec)
                if hyst_spec.error_message then
                    spec.error_message = hyst_spec.error_message
                    if md_sink then md_sink(spec) end
                    return
                end
            end
        end
    end

    -- Compute samples per buffer descriptor (bd) and buffer descriptor
    -- count (bdcnt) for ring buffer sizing
    local samples_per_bd = 1
    local rate_int = math.ceil(spec.rate)

    -- ** Warning: Limit on <samples_per_bd> of 512 cannot be changed here without changing
    -- ** size of <self.vector_data> in the dio_proxy:initialize() function
    if rate_int > 20 then
        -- Want new data at least 20 times per second (for display updates)
        samples_per_bd = math.floor(rate_int / 20)
        if samples_per_bd > 512 then samples_per_bd = 512 end
    end

    -- Buffering Time (counts) [ n * spec.rate = n seconds ]
    local bdcnt = math.ceil((spec.rate * 2) / samples_per_bd)

    -- Limit buffering allocation for rates above 10 KHz (to 20000 / 512) to
    -- prevent FPGA memory allocation error.
    if bdcnt > 40 then bdcnt = 40 end

    -- Translate spec keywords into xx_start parameters
    local args = {}
    for k,v in pairs(parameter_translations[lun.lun_type]) do
        local orig_value = spec[k]
        if type(v) == 'table' then
            args[v[1]] = v[2](orig_value)
        else
            args[v] = orig_value
        end
    end

    if spec.transducer_power then
        if spec.transducer_power == '5v' then
            self.controllers[4]:cmd_yield('txdr_power_on_5v')
            self.txdr_power = '5v'
        elseif spec.transducer_power == '12v' then
            self.controllers[4]:cmd_yield('txdr_power_on_12v')
            self.txdr_power = '12v'
        elseif spec.transducer_power == 'off' then
            self.controllers[4]:cmd_yield('txdr_power_off')
            self.txdr_power = 'off'
        end
    end

    -- If this is a digital output, set it's initial state
    if spec.initial_state then
        local num = tonumber(spec.connector:match('^output_(%d+)'))
        if spec.initial_state == 'high' then
            self.controllers[4]:cmd('set_output', num, true)
        else
            self.controllers[4]:cmd('set_output', num)
        end
    end

    -- Sanity check on the test start time
    spec.start_time_delay = spec.start_time_delay or 3
    local cts = self.controllers[4]:cmd_block('get_current_ts')
    if cts and cts.ts and cts.ts > prestart_ts then
        md_sink{
            error_message = "Test start time in the past"
        }
        return
    end

    -- Start data flow
    local msg = lun.ccadev:cmd_yield(lun.lun_type..'_start',
                                     spec.rate_num, spec.rate_denom,
                                     prestart_ts, samples_per_bd, bdcnt, args)

    -- Save parameters
    lun.pa = msg.pa
    lun.spec = spec
    lun.timestamp = 0
    lun.total_period = 0
    lun.current_period_lo = 0
    lun.current_period_hi = 0
    self.lun_lookup[msg.pa] = lun

    if md_sink then md_sink(spec) end
end

function dio_proxy:stop_lun(lun)
    local msg = lun.ccadev:cmd_yield(lun.lun_type..'_stop')
    -- Wipe out state now that it's stopped
    lun.spec = nil
    if lun.pa then
        self.lun_lookup[lun.pa] = nil
    end
    lun.pa = nil
    lun.timestamp = nil
    lun.bufdata = nil
    lun.count = nil
end

function dio_proxy:start_comet_stream(w, channels, sink)
-- required channel_proxy callback function
end

function dio_proxy:start_channel_stream(w, spec, md_sink, sink)
    local name = spec.connector:match('([^@]+)@')
    local lun = self.luns[dio_channels[name].lun]
    local func = function() self:start_lun(lun, spec, md_sink) end
    if not self.streaming then
        -- Setup FPGA Start time offset based on sample_rate_domain
        -- Decimal: 40,000,000 counts (1.0 seconds)
        -- Binary:  50,000,000 counts (1.25 seconds)
        -- Legacy:  39989590.64483642578125 counts (0.99973976612091064453125 seconds)
        self.sample_rate_domain = spec.sample_rate_domain or 'decimal'
        self.start_offset = 40000000
        self.prestart_seconds = 1.0
        if self.sample_rate_domain == 'binary' then
            self.start_offset = 50000000
            self.prestart_seconds = 1.25
        elseif  self.sample_rate_domain == 'hbm_legacy' then
            self.start_offset = 39989590.64483642578125
            self.prestart_seconds = 0.99973976612091064453125
        end
        log:notice(__"DIO: Started in ({0}) Sample Rate Domain",
                   self.sample_rate_domain)
        dio_events = nil
        self.streaming = true
        self.controllers[4]:cmd('event_setup', {
                 cable_plugged = true,
                 cable_unplugged = true,
                 output1_forced_on = true,
                 output1_forced_off = true,
                 output2_forced_on = true,
                 output2_forced_off = true,
                 })
        self.controllers[4]:cmd('event_start',
                 ts.ts_to_stamp(ts.get_ts()) + 4000000,
                 10, 1, 128)
    end
    lun.ccadev:launch_coroutine(func)
    lun.data_sinks[name] = lun.data_sinks[name] or {}
    lun.data_sinks[name][w.socket] = sink
    w.cleanup = function()
        if self.streaming then
            self.streaming = false
            self.controllers[4]:cmd('event_stop')
        end
        lun.data_sinks[name][w.socket] = nil
        if not next(lun.data_sinks[name]) then
            lun.data_sinks[name] = nil
        end
        if not next(lun.data_sinks) then
            local func = function() self:stop_lun(lun) end
            lun.ccadev:launch_coroutine(func)
        end
    end
end

function dio_proxy:start_channel_sink(w, spec, md_sink)
    if not spec.connector:find("^output_") then
        md_sink{
            error_message =
                string.format("Channel %s can not be used as a sink",
                    spec.name),
        }
        return
    end

    local num = tonumber(spec.connector:match("^output_(%d+)"))
    self:set_transducer_power(spec.transducer_power)
    self:set_output(num, spec.initial_state == 'high' and true or false)

    local function reader(loop, io, revents)
            local got = u.read(w.socket, 8192)
            if not got or got == '' then
                w.server.close(loop, io)
                return
            end
            local cj = cjson.decode(got)
            if cj and cj.state then
                self:set_output(num, cj.state == 1 and true or false)
            end
        end

    w.cleanup = function()
            if spec.final_state == 'high' then
                self:set_output(num, true)
            elseif spec.final_state == 'low' then
                self:set_output(num, false)
            end
        end

    w:callback(reader)
end

function dio_proxy:set_transducer_power(state)
    if state ~= self.txdr_power then
        if state == 'off' or  state == '5v' or state == '12v' then
            local api
            if state == 'off' then
                api = 'txdr_power_' .. state
            else
                api = 'txdr_power_on_' .. state
            end
            self.controllers[4]:cmd_block(api)
            self.txdr_power = state
            log:notice(__"DIO: {0}", api)
        else
            local err = string.format("unknown transducer power ({0})", state)
            return self.txdr_power, err
        end
    end
    return state
end

function dio_proxy:set_output(num, on)
    self.controllers[4]:cmd('set_output', num, on)
end

function dio_proxy:control(loop, w, http, spec, cb)
    if not spec.request then
        return {
            msg = string.format('Control request not specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not spec.connector then
        return {
            msg = string.format('Control with no connector specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    for _, c in ipairs(self.channels) do
        if c.connector == spec.connector then
            if not self[spec.request] then
                return {
                    msg = string.format("Control request %s undefined on %s",
                                        spec.request, c.connector),
                    result = 'error',
                }
            end
            if type(self[spec.request]) ~= 'function' then
                return {
                    msg = string.format("Control request %s invalid on %s",
                                        spec.request, c.connector),
                    result = 'error',
                }
            end
            return self[spec.request](self, spec.params)
        end
    end
    return {
        msg = string.format("Control connector %s not found",
                            spec.connector),
        result = 'error',
    }
end

function dio_proxy:digital_output(params)
    if not params[1] then
        return {
            msg = string.format("digital output number not specified for %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if params[1] ~= 1 and params[1] ~= 2 then
        return {
            msg = string.format("digital output number must be 1 or 2 for %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not params[2] then
        return {
            msg = string.format("digital output state not specified for %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    local state
    if params[2] == 'on' then
        state = true
    elseif params[2] == 'off' then
        state = false
    else
        return {
            msg = string.format("digital output state %s is invalid for %s",
                                params[2]:upper(), self.proxy_hostname),
            result = 'error',
        }
    end
    self:set_output(params[1], state)
    return {
        msg = string.format("digital output %d is %s for %s", params[1],
                            params[2]:upper(), self.proxy_hostname),
        result = 'success',
    }
end

function dio_proxy:transducer_power(params)
    if not params[1] then
        return {
            msg = string.format("transducer power not specified for %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    local state, err = self:set_transducer_power(params[1])
    if err then
        log:notice(__"DIO: {0}", err)
        return {
            msg = string.format("%s for %s", err, self.proxy_hostname),
            result = 'error',
        }
    end
    return {
        msg = string.format("transducer power is %s for %s", state,
                            self.proxy_hostname),
        result = 'success',
    }
end

return dio_proxy
