

local aux = require 'aux'
local can = require 'can'
local ffi = require 'ffi'
local listener = require 'listener'
local dbus = require 'dbus'
local sample_clock = require 'sample_clock'
local bus = dbus.bus_get('system')
local timing = assert(bus:auto_proxy('com.somat.armada.Timing',
                                     '/com/somat/armada/Timing'))

local clkfreq = sample_clock.master_rate

local cmd_timeout = 0.01

local can_address = 0x93426665



local function process_vbm_elements(channel_obj, name, pktstruct)
    if pktstruct.is_control_frame == 1 then return end
    local s = ffi.string(pktstruct.msg, pktstruct.msg_length)
    local pc = s:byte(1)
    local len = pc
    if pc >= 192 then
        -- Extended code
        len = pc - 192
        pc = pc - len
    else
        if pc >= 128 then
            -- Rx message
            len = pc - 128
            pc = pc - len
        else
            if pc >= 64 then
                -- Tx message
                len = pc - 64
                pc = pc - len
            else
                -- Unknown message
            end
        end
    end
    if pc == 128 then
        can_id = s:sub(2,5)
        if can_id:byte(1) >= 128 then
            can_id = string.char(can_id:byte(1) - 128, can_id:byte(2, 4))
        end
        if can_id == can_address then
            local data_rcved = s:sub(6, -1)
            local data = channel_obj.verify_fd:read(#data_rcved)
            if data ~= data_rcved then
                print("VBM Mismatch!", data, data_rcved)
                channel_obj.verify_errors = channel_obj.verify_errors + 1
            end
            channel_obj.verify_count = channel_obj.verify_count + #data
        end
    end
end


local function process_can_elements(channel_obj, name, pktstruct)
    local ff = pktstruct.frame_format
    local ext_frame = false
    if ff > 127 then
        ff = ff - 128
        ext_frame = true
    end
    local rtr = false
    if ff > 63 then
        ff = ff - 64
        rtr = true
    end
    local direction = "RX"
    if ff > 15 then
        ff = ff - 16
        direction = "TX"
    end
    local can_id = pktstruct.can_id
    if ext_frame and (can_id == can_address) then
        local data_rcved = ffi.string(pktstruct.can_message, ff)
        if direction == "RX" then
            local data = channel_obj.verify_rx_fd:read(#data_rcved)
            if data ~= data_rcved then
                print("CAN Mismatch!", direction, data, data_rcved)
                channel_obj.verify_rx_errors = channel_obj.verify_rx_errors + 1
            end
            channel_obj.verify_rx_count = channel_obj.verify_rx_count + #data
        else
            local data = channel_obj.verify_tx_fd:read(#data_rcved)
            if data ~= data_rcved then
                print("CAN Mismatch!", direction, data, data_rcved)
                channel_obj.verify_tx_errors = channel_obj.verify_tx_errors + 1
            end
            channel_obj.verify_tx_count = channel_obj.verify_tx_count + #data
        end
    end
end


local function stress_test(can_devname, aux_devname, filename)
    local read_fd = io.open(filename, 'rb')

    -- Setup CAN
    local can_channel = can:new(can_devname, "can.bin", clkfreq)
    can_channel.verify_rx_fd = io.open(filename, 'rb')
    can_channel.verify_tx_fd = io.open(filename, 'rb')
    can_channel.verify_rx_errors = 0
    can_channel.verify_tx_errors = 0
    can_channel.verify_rx_count = 0
    can_channel.verify_tx_count = 0
    can_channel.element_action_func = process_can_elements
    can_channel:bus_setup({baud_rate = 250000, sjw = 2, tseg1 = 13, tseg2 = 2 })
    assert(listener({can_channel}, cmd_timeout))
    can_channel:enable_terminator()
    assert(listener({can_channel}, cmd_timeout))
    can_channel:get_current_ts()
    local msg = assert(listener({can_channel}, cmd_timeout))
    local can_start_time = msg.ts + clkfreq / 2
    can_channel:rx_start(can_start_time - 3, 1, 16, 128)
    assert(listener({can_channel}, cmd_timeout))
    can_channel:tx_start_buf(0, 16, 128)
    assert(listener({can_channel}, cmd_timeout))
    listener({can_channel}, 1.0)

    -- Setup AUX port for VBM mode
    local aux_channel = aux:new(aux_devname, "aux.bin", clkfreq)
    aux_channel.verify_fd = io.open(filename, 'rb')
    aux_channel.verify_errors = 0
    aux_channel.verify_count = 0
    aux_channel.element_action_func = process_vbm_elements
    aux_channel:setup_vbm()
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:txdr_power_on()
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:get_current_ts()
    local msg = assert(listener({aux_channel}, cmd_timeout))
    local aux_start_time = msg.ts + clkfreq / 2
    aux_channel:rx_start(aux_start_time - 3, 1, 16, 128)
    assert(listener({aux_channel}, cmd_timeout))
    aux_channel:tx_start(0, 16, 128)
    assert(listener({aux_channel}, cmd_timeout))

    listener({aux_channel, can_channel}, 1.0)

    aux_channel:tx_cmd_msg(0, {"HIGH_SPEED=YES",
                               "TERMINATION_ENABLE=YES",
                               "CAN_MODE=INITIALIZE",
                               "SYNC_JUMP_WIDTH=3",
                               "BAUD_RATE_PRESCALER=4",
                               "TIME_SEGMENT_1=11",
                               "TIME_SEGMENT_2=4",
                               "SAMPLES_PER_BIT=3",
                               "LISTEN_ONLY_MODE=NO",
                               "CAN_MODE=NORMAL",
                               "APPSTATE?",
                               "RUN=YES"
                               })
    assert(listener({aux_channel, can_channel}, 3.0, 12))

    listener({aux_channel, can_channel}, 2.0)

    local start_clock = os.time()
    local max_read_size = 1048
    local bytes_read = 0
    while true do
        local data = read_fd:read(max_read_size)
        if data == nil then break end
        bytes_read = bytes_read + #data
        local msg_count = math.ceil(#data / 8)
        local msgs = can_channel:allocate_msg_array(msg_count)
        local spos = 1
        for mi=1,msg_count do
            can_channel:insert_msg(msgs, mi, 0, can_address,
                                   data:sub(spos, spos + 7))
            spos = spos + 8
        end
        local start_index = 0
        io.write('+') io.flush()
        local retries = 0
        while true do
            start_index = can_channel:tx_append_msgs(0, msgs, start_index)
            if start_index == msg_count then break end
            listener({aux_channel, can_channel}, 0.1)
            io.write('.') io.flush()
            retries = retries + 1
            if retries == 10 then
                print("\nRetries exceeded!")
                can_channel:halt(true)
                can_channel:dump_ring_buffers()
                os.exit(1)
            end
        end
    end
    local elapsed_time = os.time() - start_clock

    listener({aux_channel, can_channel}, 2.0)

    print("Cleaning up")

    aux_channel:tx_stop(0)
    assert(listener({aux_channel}, cmd_timeout))
    can_channel:tx_stop(0)
    assert(listener({can_channel}, cmd_timeout))
    aux_channel:rx_stop()
    assert(listener({aux_channel}, cmd_timeout))
    can_channel:rx_stop()
    assert(listener({can_channel}, cmd_timeout))

    aux_channel:txdr_power_off()
    assert(listener({aux_channel}, cmd_timeout))

    aux_channel:halt()
    can_channel:halt()

    print("Done")
    print("Bytes Read from Source File = "..bytes_read)
    print("CAN RX Errors = "..can_channel.verify_rx_errors)
    print("CAN TX Errors = "..can_channel.verify_tx_errors)
    print("AUX RX Errors = "..aux_channel.verify_errors)
    print("CAN RX Bytes Transferred = "..can_channel.verify_rx_count)
    print("CAN TX Bytes Transferred = "..can_channel.verify_tx_count)
    print("AUX RX Bytes Transferred = "..aux_channel.verify_count)
    print("Elapsed Time = "..elapsed_time)
    local rate = bytes_read / elapsed_time
    print("Throughput = "..rate.." bytes/sec")
    os.exit(0)
end




if #arg < 3 then
    print("Usage can2vbm_stress_test <can device name> <aux device name>"..
          " <file to transfer>")
    os.exit(1)
end


stress_test(arg[1], arg[2], arg[3])

