--
-- Copyright (C) 2015 Hottinger Baldwin Measurements, Inc.
--

local channel_proxy = require 'channel_proxy'
local task_chain    = require 'task_chain'

local mxChannelProxy = class('mxChannelProxy', channel_proxy)

-- The MX471B will send us data values at least at this interval, in
-- seconds. If this time elapses without the MX471B receiving a new
-- CAN message, it will send us a "no data" value.
local MIN_REPEAT = 1

-- The XML-escaped preamble statement of an XML document.
local XML_PREAMBLE_ESCAPED = '&lt;?xml version=&quot;1.0&quot;' ..
    ' encoding=&quot;UTF-8&quot;?&gt;'

function mxChannelProxy:initialize(mxDevice, info)

    self.mxdev              = mxDevice
    self.proxy_type         = 'mxChannelProxy'
    self.can_info           = info
    self.num_signals        = 0
    self.signal_slots       = {}
    self.queued_channels    = {}

    self.xpath_root = '/AdditionalConnector'

end

function mxChannelProxy:available_channels(hostname)

    return {
        {
            name        = 'raw_' .. hostname:gsub('%W', '_'),
            connector   = '@' .. hostname,
            description = 'CAN messages',
            data_type   = 'message_float64',
            category    = 'source/hbm_mx471b',
            sample_rate = 10
        }
    }

end

function mxChannelProxy:start_comet_stream(w, chans, sink)

end

function mxChannelProxy:start_channel_stream(w, spec_array, md_sink, sink)

    for _, spec in ipairs(spec_array) do

        -- If this is a new test ID, discard any queued channels.
        self.mxdev.queued_channels = self.mxdev.queued_channels or {}
        if spec._test_id ~= self.mxdev.test_id then
            local virt_chans = self.mxdev.mx.control.device.virtual_channels
            for _, virt_chan in ipairs(virt_chans) do
                virt_chan.proxy.queued_channels = {}
            end
            self.mxdev.queued_channels = {}
            self.mxdev.test_id = spec._test_id
        end

        -- Queue up all channel requests until all of them have been received.
        -- We're maintaining TWO lists here: self.queued_channels, which has all
        -- channels for a given CAN connector, and self.mxdev.queued_channels,
        -- which is a global queue of all channels in the test for this MX471B
        -- device. We need the latter to enforce the 128-per-connector limit, and
        -- we need the former to know when we've received start_channel_stream()
        -- calls for all the channels so we can start up.

        local queued_channel = {
            w                   = w,
            spec                = spec,
            sink                = sink,
            md_sink             = md_sink,
            stream_reference    = string.format(
                                    'CanReceiver_Connector%d_Decoder%d.Signal1',
                                    self.can_info.channel_num,
                                    #self.queued_channels + 1)
        }

        self.queued_channels[#self.queued_channels + 1] = queued_channel
        self.mxdev.queued_channels[#self.mxdev.queued_channels + 1] =
            queued_channel

        if #self.mxdev.queued_channels == spec._channel_count then
            self:_start_test(md_sink)
        end

    end

    -- When the test ends, detach all channel sinks and tell libmxstream
    -- to unsubscribe from all of the signals.
    w.cleanup = function()
        self.mxdev.active_channels = self.mxdev.active_channels - #spec_array
        if self.mxdev.active_channels == 0 then
            local tasks = task_chain:new()
            for i, channel in ipairs(self.mxdev.queued_channels) do
                self.mxdev.mx.hbmstream:remove_subscription(
                    channel.stream_reference)
            end
            self.mxdev.mx.hbmstream:update_subscriptions()
            local decoder_nums = {}
            for i = 1, 128 do decoder_nums[i] = i end
            for i = 1, 4 do
                tasks:add(function(on_success, on_error)
                    return self:_activate_all_decoders(i, decoder_nums,
                        '0', on_success, on_error) end)
            end
            tasks:run()
        end
    end

end

local function tag(name, value)
    return '&lt;' .. name .. '&gt;' .. value .. '&lt;/' .. name .. '&gt;'
end

local function open_tag(name)
    return '&lt;' .. name .. '&gt;'
end

local function close_tag(name)
    return '&lt;/' .. name .. '&gt;'
end

local function empty_tag(name)
    return '&lt;' .. name .. ' /&gt;'
end

function mxChannelProxy:_start_test(md_sink)

    -- Validate setup for each of the four CAN connectors
    local virt_chans = self.mxdev.mx.control.device.virtual_channels
    for _, virt_chan in ipairs(virt_chans) do

        -- Only support up to 128 signal slots per CAN connector
        local chans = virt_chan.proxy.queued_channels or {}
        if #chans > 128 then
            md_sink({ error_message = "Cannot support more than " ..
                "128 CAN signals on a single MX471B channel" })
            return
        end

        if #chans > 0 then
            -- All signals on a connector must have same bitrate/termination
            for _, channel in ipairs(chans) do
                if channel.spec.bit_rate ~= chans[1].spec.bit_rate then
                    md_sink({ error_message = "CAN signals on a single " ..
                        "connector must have the same bit rate" })
                    return
                end
                if channel.spec.internal_termination ~=
                    chans[1].spec.internal_termination then
                    md_sink({ error_message = "CAN signals on a single " ..
                        "connector must have the same termination setting" })
                    return
                end
            end
        end

    end

    -- Set up a chain of tasks/callbacks to perform.
    local tasks = task_chain:new()
    local xml = nil

    local virt_chans = self.mxdev.mx.control.device.virtual_channels

    -- Send XML configuration for the baud rate and termination settings
    -- for each enabled connector, one by one.
    for connector_num, virt_chan in ipairs(virt_chans) do
        for _, chan in ipairs(virt_chan.proxy.queued_channels) do
            tasks:add(function(on_success, on_error)
                return self:_set_canbus_settings(connector_num,
                    chan.spec.bit_rate, chan.spec.internal_termination,
                    on_success, on_error) end)
            break
        end
    end

    -- Send XML configuration for each channel.
    local activations = {}
    for connector_num, virt_chan in ipairs(virt_chans) do
        for decoder_num, chan in ipairs(virt_chan.proxy.queued_channels) do
            tasks:add(function(on_success, on_error)
                return self:_set_decoder_settings(connector_num, decoder_num,
                    chan.spec, on_success, on_error) end)
            tasks:add(function(on_success, on_error)
                return self:_set_async_signal_settings(connector_num,
                    decoder_num, on_success, on_error) end)
            activations[connector_num] = activations[connector_num] or {}
            activations[connector_num][decoder_num] = decoder_num
        end
    end

    -- Set all used decoders to active.
    for connector_num, decoder_nums in pairs(activations) do
        tasks:add(function(on_success, on_error)
            return self:_activate_all_decoders(connector_num, decoder_nums,
                '1', on_success, on_error) end)
    end

    -- Next, tell libhbmstream to subscribe to all of the channels.
    tasks:add(
        function(on_success, on_error)
            for _, channel in ipairs(self.mxdev.queued_channels) do
                self.mxdev.mx.hbmstream:add_subscription(
                    channel.stream_reference, channel.w.socket)
            end
            self.mxdev.mx.hbmstream:update_subscriptions()
        end)

    -- Initialize this for mx_proxy's benefit (though we won't be using
    -- the start_channel_list at all for our purposes).
    self.mxdev.start_channel_list = self.mxdev.start_channel_list or {}

    -- Run all the tasks we just queued up.
    self.mxdev.active_channels = #self.mxdev.queued_channels
    tasks:run()

end

function mxChannelProxy:_set_configuration(xpath, xml, on_success, on_error)

    return self.mxdev.mx.control:set_view_data(8, xpath, xml,
        function(results)
            if results.result == 'success' then
                if on_success then return on_success() end
            else
                if on_error then return on_error(results.message) end
            end
        end)

end

function mxChannelProxy:_set_canbus_settings(connector_num,
    bit_rate, termination, on_success, on_error)

    local xpath = 'AdditionalConnector'
    xpath = xpath .. string.format(
        '/Connector[@address=&quot;%d&quot;]',
        connector_num)
    xpath = xpath .. '/Mode[@choice=&quot;CanBus&quot;]'

    local termination_str = '0'
    if termination and termination ~= 0 then termination_str = 1 end

    local xml = XML_PREAMBLE_ESCAPED
    xml = xml .. open_tag('CanBus')
    xml = xml .. tag('Baudrate', bit_rate)
    xml = xml .. tag('Termination', termination_str)
    xml = xml .. tag('ListenOnly', '0')
    xml = xml .. close_tag('CanBus')

    return self:_set_configuration(xpath, xml, on_success, on_error)

end

function mxChannelProxy:_set_decoder_settings(
    connector_num, decoder_num, spec, on_success, on_error)

    local xpath = 'AdditionalConnector'
    xpath = xpath .. string.format(
        '/Connector[@address=&quot;%d&quot;]',
        connector_num)
    xpath = xpath .. '/Mode[@choice=&quot;CanBus&quot;]'
    xpath = xpath .. '/CanBus'
    xpath = xpath .. string.format(
        '/CanInChannel[@address=&quot;%d&quot;]',
        decoder_num)

    local msg_id = bit.band(tonumber(spec.db_msg_id), 0x1fffffff)

    local extended_frame = 0
    if msg_id >= 2048 then extended_frame = 1 end

    local bit_sequence = 0
    if spec.db_data_format:find('LSB') then bit_sequence = 1 end

    -- 0 = unsigned 32 bit
    -- 1 = signed 32 bit
    -- 2 = unsigned 64 bit
    -- 3 = signed 64 bit
    -- 4 = 32 bit real
    -- 5 = 64 bit real

    local raw_value_format = 0
    if spec.db_data_format:find('^U') then
        if spec.db_bit_length > 32 then
            raw_value_format = 2
        else
            raw_value_format = 0
        end
    elseif spec.db_data_format:find('^S') then
        if spec.db_bit_length > 32 then
            raw_value_format = 3
        else
            raw_value_format = 1
        end
    elseif spec.db_data_format:find('^F') then
        if spec.db_bit_length > 32 then
            raw_value_format = 5
        else
            raw_value_format = 4
        end
    else
        return on_error('Unrecognized database data format ' ..
            spec.db_data_format)
    end

    local scale = spec.db_scale
    local offset = spec.db_offset
    local value_min = spec.db_min
    local value_max = spec.db_max

    if spec.user_scale ~= nil and spec.user_offset ~= nil then
        scale = scale * spec.user_scale
        offset = spec.user_offset + offset * spec.user_scale
        value_min = value_min * spec.user_scale + spec.user_offset
        value_max = value_max * spec.user_scale + spec.user_offset
    end

    local xml = XML_PREAMBLE_ESCAPED

    xml = xml .. open_tag('CanSensor')
    xml = xml .. tag('Identifier', msg_id)
    xml = xml .. tag('SignalName', spec.name)
    xml = xml .. tag('OriginOfName', '1')
    xml = xml .. tag('ExtendedFrame', extended_frame)
    xml = xml .. tag('BitSequence', bit_sequence)
    xml = xml .. tag('SignalType', '0')
    xml = xml .. tag('RawValueFormat', raw_value_format)
    xml = xml .. tag('StartBit', spec.db_bit_start)
    xml = xml .. tag('SignalLength', spec.db_bit_length);
    xml = xml .. tag('ScalingFactor', scale)
    xml = xml .. tag('ScalingOffset', offset)
    xml = xml .. tag('ModeBitSequence', '1')
    xml = xml .. tag('ModeStartbit', '0')
    xml = xml .. tag('ModeLength', '2')
    xml = xml .. tag('ModeValueFormat', '0')
    xml = xml .. tag('ModeValue', '3')
    xml = xml .. tag('PhysUnit', '100000')
    xml = xml .. empty_tag('PhysUnitUsr')
    xml = xml .. tag('Min', value_min)
    xml = xml .. tag('Max', value_max)
    xml = xml .. close_tag('CanSensor')

    return self:_set_configuration(xpath, xml, on_success, on_error)

end

function mxChannelProxy:_set_async_signal_settings(
    connector_num, decoder_num, on_success, on_error)

    local xpath = 'AdditionalConnector'
    xpath = xpath .. string.format(
        '/Connector[@address=&quot;%d&quot;]',
        connector_num)
    xpath = xpath .. '/Mode'
    xpath = xpath .. '/CanBus'
    xpath = xpath .. string.format(
        '/CanInChannel[@address=&quot;%d&quot;]',
        decoder_num)
    xpath = xpath .. '/AsyncSignal'

    local xml = XML_PREAMBLE_ESCAPED
    xml = xml .. open_tag('AsyncSignal')
    xml = xml .. tag('MaxRepTime', MIN_REPEAT)
    xml = xml .. tag('Decimals', '4')
    xml = xml .. tag('SignalFormat', '4')
    xml = xml .. close_tag('AsyncSignal')

    return self:_set_configuration(xpath, xml, on_success, on_error)

end

function mxChannelProxy:_activate_all_decoders(
    connector_num, decoder_nums, activate, on_success, on_error)

    local xpath = 'AdditionalConnector'
    xpath = xpath .. string.format(
        '/Connector[@address=&quot;%d&quot;]',
        connector_num)
    xpath = xpath .. '/Mode'

    local xml = XML_PREAMBLE_ESCAPED
    xml = xml .. open_tag('CanBus')
    for _, decoder_num in ipairs(decoder_nums) do
        xml = xml .. '&lt;CanInChannel address=&quot;' ..
            decoder_num .. '&quot;&gt;'
        xml = xml .. tag('Active', activate)
        xml = xml .. close_tag('CanInChannel')
    end
    xml = xml .. close_tag('CanBus')

    return self:_set_configuration(xpath, xml, on_success, on_error)

end

return mxChannelProxy
