--! \file

local common        = require 'mx.device_common'
local underscore    = require 'underscore'
local ffi           = require 'ffi'
local C             = ffi.C

require 'xpath'

local m = {}

-- XML used to do initial configuration of channels when proxy starts up
local default_xml = [[<?xml version="1.0" encoding="UTF-8"?>
                <Amplifier>
                    <Connector#conn_num#>
                        <Mode choice="AnalogInChannel">
                            <AnalogInChannel>
                                <Sensor>
                                    <TedsOptions>
                                        <UseTeds>2</UseTeds>
                                    </TedsOptions>
                                </Sensor>
                                <SyncSignal address="1">
                                    <OutputRate>9600</OutputRate>
                                </SyncSignal>
                                <OutputUnit>
                                    <UnitId>1100</UnitId>
                                </OutputUnit>
                            </AnalogInChannel>
                        </Mode>
                    </Connector#conn_num#>
                </Amplifier>
]]

-- Function used to return the XML used to set the channel to it's default
-- configuration
function m:get_default_xml(channel_num)
    return default_xml:gsub('#conn_num#', tostring(channel_num))
end

m.json2xpath = {
    transducer_power = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Supply',
    excitation_range = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/ExciteAmpl',
    excitation_mode = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/ExciteFreq',
    _bridge_type = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/BridgeType',
    bridge_resistance = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/Impedance',
    _bridge_wiring = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype[@choice="#sensor_input_mode#"]/#sensor_input_mode#/Wiring',
    _pt_wiring = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype[@choice="Pt"]/Pt/Wiring',
    _pt_type = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/Pt/PtType',
}

underscore.extend(m.json2xpath, common.json2xpath)

m.phase_delay = {}
m.phase_delay.fixed = 398
m.phase_delay.bessel_iir = {}
m.phase_delay.butterworth_iir = {}
m.phase_delay.linear_phase_fir = {}
m.phase_delay.butterworth_fir = {}

m.physical_unit_map = {
    Voltage = 'V',
    Current = 'mA',
    IEPE = 'C',
}

-- List of parameters in a channel setup that we do NOT want sent off to the MX
-- device
m.ignore_list = {}
underscore.extend(m.ignore_list, common.ignore_list)

-- List of parameters returned by MX device that we don't verify after setting
-- up a channel.
m.dont_verify_list = {}
underscore.extend(m.dont_verify_list, common.dont_verify_list)

function m:sensor_input_type(p, fromxml)
    return m.translate({
            ['1'] = 'quarter',
            ['2'] = 'half',
            ['4'] = 'full',
        }, p, fromxml)
end

function m:excitation_mode(p, fromxml)
    return m.translate({
            ['1'] = 'dc',
            ['5'] = 'ac_4800_hz_sine',
        }, p, fromxml)
end

function m:excitation_range(p, fromxml)
    if fromxml then
        return string.format("%.1fv", tonumber(p))
    end
    return p:match('(.+)[V|v]$')
end

m.translator = {
    sensor_input_type = m.sensor_input_type,
    excitation_mode = m.excitation_mode,
    excitation_range = m.excitation_range
}

function m:to_internal(spec)
    local spec = spec

    -- Do some processing based on sensor input mode (if it exists)
    if spec.sensor_input and spec.sensor_input.mode then

        -- If this is a bridge, set the bridge wiring
        if (spec.sensor_input.mode == 'bridge' or 
            spec.sensor_input.mode == 'inductive_bridge') and 
            spec.excitation_range then
            -- If this is NOT a TEDS device do this stuff
            if not spec.teds_sensor or not spec.teds_sensor.bindata then
                local bridge_max_levels = {
                    ['7.5'] = 4,
                    ['5.0'] = 4,
                    ['2.5'] = 8,
                    ['1.0'] = 20,
                }
                local er = string.format("%.1f",
                    tonumber(spec.excitation_range:match('(.*)[V|v]') or 0))
                if not bridge_max_levels[er] then
                    return nil, string.format("Invalid excitation range '%s'",
                        spec.excitation_range)
                end
                spec.excitation_range = er
                spec.sensor_input.max = spec.sensor_input.max or
                    bridge_max_levels[er]
                if spec.sensor_input.type == 'full' and
                    tonumber(spec.sensor_input.wires) == 4 then
                    spec.sensor_input.wires = '6'
                elseif spec.sensor_input.type == 'half' and
                    tonumber(spec.sensor_input.wires) == 3 then
                    spec.sensor_input.wires = '5'
                elseif spec.sensor_input.type == 'quarter' and
                    tonumber(spec.sensor_input.wires) == 3 and
                    not spec.sensor_input.regulation then
                    spec.sensor_input.wires = '9'
                end
            end
            spec._bridge_type =
                m.sensor_input_type(self, spec.sensor_input.type)
            spec.sensor_input.min = spec.sensor_input.max * -1
            spec._bridge_wiring = spec.sensor_input.wires
            spec.sensor_input.wires = nil
            spec.sensor_input.type = nil
        end
    end

    return spec
end

local set_shunt_xml_template =[[
<?xml version="1.0"?>
<SrvParams>
<Control>SHUNTCALIBRATION</Control>
<ControlParams>&lt;?xml version="1.0"
encoding="UTF-8"?&gt;&lt;ControlParams&gt;&lt;Par1&gt;#channel&lt;/Par1&gt;&lt;Par2&gt;#state&lt;/Par2&gt;&lt;/ControlParams&gt;</ControlParams>
</SrvParams>
]]

m.controls = {}
m.controls.channel = {}
m.controls.device = {}

-- Function used to control shunts on Quantum devices that have them.
-- Note, self in this function is the ChannelProxy object
function m.controls.channel:shunt(params, cb)
    local chan = self.index
    local enable = params and params[1]
  local xml = set_shunt_xml_template
    xml = string.gsub(xml, '#channel', chan)
    if type(enable) == 'boolean' then
        xml = string.gsub(xml, '#state', (enable and 1 or 0))
    else
        xml = string.gsub(xml, '#state', enable)
    end
    self.mxdev.mx.control:_send_transaction_request(C.TR_SCHEMA, C.TR_XML,
        function(obj, ecode, data, len)
            cb{result = 'success', msg = 'ok'}
        end, xml)
end

-- Device level shunt control that enables/disables shunts on all channels
function m.controls.device:shunt(params, cb)
    local enable = params and params[1]
    local xml = set_shunt_xml_template
    xml = string.gsub(xml, '#channel', 0)
    if type(enable) == 'boolean' then
        xml = string.gsub(xml, '#state', (enable and 1 or 0))
    else
        xml = string.gsub(xml, '#state', enable)
    end
    self.mx.control:_send_transaction_request(C.TR_SCHEMA, C.TR_XML,
        function(obj, ecode, data, len)
            cb{result = 'success', msg = 'ok'}
        end, xml)
end

local function escape_xml(xml)
    xml = xml:gsub('<', '&lt;')
    xml = xml:gsub('>', '&gt;')
    xml = xml:gsub('"', '&quot;')
    return xml
end

local connectors_off_xml = escape_xml([[
<Amplifier>
  <Connector1><Mode choice="Off" /></Connector1>
  <Connector2><Mode choice="Off" /></Connector2>
  <Connector3><Mode choice="Off" /></Connector3>
  <Connector4><Mode choice="Off" /></Connector4>
</Amplifier>
]])

local highspeed_connectors_on_xml = escape_xml([[
<Amplifier>
  <Connector1><Mode choice="AnalogInChannel"><AnalogInChannel>
    <SpeedMode>1</SpeedMode></AnalogInChannel></Mode></Connector1>
  <Connector2><Mode choice="AnalogInChannel"><AnalogInChannel>
    <SpeedMode>1</SpeedMode></AnalogInChannel></Mode></Connector2>
</Amplifier>
]])

local lowspeed_connectors_on_xml = escape_xml([[
<Amplifier>
  <Connector1><Mode choice="AnalogInChannel"><AnalogInChannel>
    <SpeedMode>0</SpeedMode></AnalogInChannel></Mode></Connector1>
  <Connector2><Mode choice="AnalogInChannel"><AnalogInChannel>
    <SpeedMode>0</SpeedMode></AnalogInChannel></Mode></Connector2>
  <Connector3><Mode choice="AnalogInChannel"><AnalogInChannel>
    <SpeedMode>0</SpeedMode></AnalogInChannel></Mode></Connector3>
  <Connector4><Mode choice="AnalogInChannel"><AnalogInChannel>
    <SpeedMode>0</SpeedMode></AnalogInChannel></Mode></Connector4>
</Amplifier>
]])

-- Control API function to switch the MX411B-R device into or out of
-- high-speed mode. In high speed mode, connectors 3 and 4 are disabled
-- and connectors 1 and 2 support signal rates up to 200 kHz.
function m.controls.device:change_speed_mode(params, cb)
    local enable = params and params[1] == 'high'
    local on_xml = enable and
        highspeed_connectors_on_xml or
        lowspeed_connectors_on_xml
    self.mx.control:set_view_data(5,
        'Amplifier',
        connectors_off_xml,
        function(results)
            if results.result == 'success' then
                self.mx.control:set_view_data(5,
                    'Amplifier',
                    on_xml,
                    function(results)
                        if results.result == 'success' then
                            self.current_speed_mode = 1
                            cb{result = 'success', msg = 'ok'}
                            os.exit(0)
                        else
                            cb{result = 'error', msg = results.message}
                        end
                    end)
            else
                cb{result = 'error', msg = results.message}
            end
        end)
end

function m:to_external(spec)
    local spec = spec

    -- Don't return this stuff to the external entities
    spec.wiring = nil

    if spec.sensor_input.mode and 
    spec.sensor_input.mode ~= 'bridge' and
    spec.sensor_input.mode ~= 'inductive_bridge' then
        spec.bridge_bridge_factor = nil
        spec.bridge_gage_factor = nil
        spec.bridge_resistance = nil
        spec.excitation_mode = nil
        spec.excitation_range = nil
    else
        spec.sensor_input.type =
            m.sensor_input_type(self, spec._bridge_type, true)
        spec.sensor_input.wires = spec._bridge_wiring
    end

    spec._bridge_wiring = nil
    spec._bridge_type = nil

    return spec
end

-- Specification of the default settings of this channel type
m.default_spec = {
    sample_rate = '10',
    sensor_input = {
        mode = 'volt',
        min = '-10',
        max = '10.000000',
    },
    calibration = {
        define_source_units = {
            purpose = 'define_source_units',
            units = 'V',
        },
        sensor_scaling = {
            purpose = 'sensor_scaling',
            mode = 'defined_slope_intercept',
            slope = '1',
            intercept = '0',
            units = 'V',
        },
        units_conversion = {
            purpose = 'units_conversion',
            mode = 'defined_slope_intercept',
            units = 'V',
            intercept = '0',
            slope = '1',
        },
        rezero = {
            purpose = 'rezero',
            mode = 'defined_slope_intercept',
            intercept = '0.00000',
        },
    },
    transducer_power = '0.000000',
    digital_filter_type = 'none',
    sensor = {
        use_teds = 2,
    },
}

local speed_mode_map = {
    [0] = 'normal',
    [1] = 'high'
}

m.current_speed_mode = 0

function m:create_additional_metachannels(device)

    local n = device.name:gsub('-', '_')
    device.device_channels['speedmode_status'] = {
        name = string.format('%s_speed_status', n),
        connector = string.format('speed.status@%s', device.proxy_hostname),
        description = 'MX High Speed Mode',
        data_type = 'message_raw',
        output_units = 'state',
        category = string.format('source/hbm_%s', device.category_name),
        message_rate = 1,
        is_logical = false,
        comet_sinks = {},
        valid_controls = {
            change_speed_mode = true,
        },
    }

end

function m:additional_metachannels_comet_service(device)

    local c = device.device_channels.speedmode_status
    for _, sink in pairs(c.comet_sinks) do
        local key = self.current_speed_mode or 0
        local msg = speed_mode_map[key] or 'unknown'
        sink(0, c.connector, msg)
    end

end

function m:additional_info(device, info)

    local key = self.current_speed_mode or 0
    local msg = speed_mode_map[key] or 'unknown'
    info.current_speed_mode = msg
    return info

end

function m:read_initial_state(device)

    local xml = device.control.view['Analog input connector parameters']

    local connectors = {}
    self.speed_modes = {}

    for i = 1, 4 do
        local conn_mode = xpath.selectNodes(xml,
            '/Amplifier/Connector' .. i .. '/Mode/@choice')[1]
        local speed_mode = xpath.selectNodes(xml,
            '/Amplifier/Connector' .. i ..
            '/Mode/AnalogInChannel/SpeedMode/text()')[1]
        connectors[i] = {
            mode = conn_mode, speed = speed_mode }
        self.speed_modes[i] = speed_mode
    end

    if connectors[1].mode == 'AnalogInChannel' and
        connectors[2].mode == 'AnalogInChannel' and
        connectors[1].speed == '1' and
        connectors[2].speed == '1' then
        self.current_speed_mode = 1
    else
        self.current_speed_mode = 0
    end

end

return m
