local ffi = require 'ffi'
local C = ffi.C
local gmp = require 'gmp'
local rational = require 'rational'
local ts = require 'timestamps'
local config = require 'system_config'

gmp.f_set_default_prec(1024)

local logger = {
    log = function() end,
}

local m = {}

-- Find the greatest common divisor between two non-zero values
local function calc_gcd(x, y)
    local a = ffi.new('unsigned long', (tonumber(x) >= tonumber(y)) and x or y)
    local b = ffi.new('unsigned long', (tonumber(x) >= tonumber(y)) and y or x)
    local remainder = ffi.new('unsigned long')

    while true do
        remainder = a % b
        if tonumber(remainder) == 0 then
            return b
        end
        a = b
        b = remainder
    end
end

-- Function used to find the greatest common divisor of a series of values
local function find_gcd(count, values)
    if count < 2 then return values[0] end
    for i = 1, count - 1 do
        values[0] = calc_gcd(values[i], values[0])
    end
    return values[0]
end

local gmp_print_buf = ffi.new('char[1024]')

-- Function used to get the greatest common divisor across the sample rates of
-- all channels.
function m.get_gcd(drd, channels)
    local mult_factor = ffi.new('unsigned long', (drd == 'binary' and 5 or 1))

    local work = ffi.new('unsigned long[?]', #channels)
    for i, c in ipairs(channels) do
        work[i - 1] = mult_factor *
            ffi.new('unsigned long',
                rational:new(c.sample_rate_legacy_uncorrected):tonumber())
    end
    local gcd_rate = gmp.new_f()
    gmp.f_set_ui(gcd_rate, find_gcd(#channels, work))

    local gcd_period = gmp.new_f()
    if drd == 'hbm_legacy' then
        gmp.f_mul_ui(gcd_rate, gcd_rate, 8388608)
        gmp.f_div_ui(gcd_rate, gcd_rate, 8386425)
    end
    gmp.f_ui_div(gcd_period, 1, gcd_rate)

    if config.mx_phase_debug_messages then
        gmp.sprintf(gmp_print_buf, "GCD Rate: %.20Ff, GCD Period: %.20Ff",
            gcd_rate, gcd_period)
        logger:log(logger.INFO, ffi.string(gmp_print_buf))
    end

    return gcd_rate, gcd_period
end

-- This is the epoch on MX devices
local mx_epoch = os.time({
    year = 2000,
    month = 1,
    day = 1,
    hour = 0,
    min = 0,
    sec = 0
    })

-- Convert an ISO formatted time string (requested start time) to an MX time
-- base time.
local function iso_to_mx_ts(iso)
    local time, fraction = { }, nil
    time.year, time.month, time.day, time.hour, time.min, time.sec, fraction =
        iso:match('(%d%d%d%d)%-(%d%d)%-(%d%d)T(%d%d)%:(%d%d)%:(%d%d)%.?(%d*)')
    local seconds = os.time(time)
    local ns = 0
    if fraction then
        local frac_digits = fraction:len()
        ns = fraction * 10^(9 - frac_digits)
    end
    seconds = seconds - mx_epoch
    seconds = seconds + (ns / 1000000000)
    return seconds
end


-- Function used to adjust the test start time based on any/all MX channels
-- being used in the test
function m.get_adjusted_start_time(requested_time, setup, log)
    logger = log or logger

    -- Find the data rate domain for the test
    local drd
    for _, section in ipairs(setup) do
        if section.category and section.category == 'summary' and
            section.sample_rate_domain then
            drd = section.sample_rate_domain
            break
        end
    end

    -- If there is no data rate domain specified, ERROR!
    if not drd then
        logger:log(logger.ERR, "No data rate domain specified in test")
        return requested_time
    end

    -- Get the channels we care about for this
    local channels, mx_types
    for _, section in ipairs(setup) do
        if section.category and section.category:find('hbm_mx') then
            channels = channels or {}
            mx_types = mx_types or {}
            if not section.is_async then
                channels[#channels + 1] = section
            end
            local t = section.category:match('hbm_(.+)')
            mx_types[t:lower()] = true
        end
    end

    -- If there are no MX channels in this test, just return the requestsed
    -- start time
    if not channels or #channels == 0 then return requested_time end

    if config.mx_phase_debug_messages then
        logger:log(logger.INFO,
            string.format("Requested start time: %s", requested_time))
    end

    -- First we need to convert the requested time in ISO format to seconds
    -- Add .35 seconds for each MX channel in the test.  NOTE:  This value was
    -- determined via experimentation using and MX1601.
    local req_time = ts.iso_to_epoch(requested_time) +
        (#channels * (config.per_mx_channel_setup_delay or .06))

    -- Then we need to convert the time from CX timebase to MX timebase
    req_time = req_time - mx_epoch

    -- Get the GCD rate and period of the set of sample rates
    -- NOTE:  These values are GMP F values
    local gcd_rate, gcd_period = m.get_gcd(drd, channels)

    -- Figure out how many GCD periods exist in the requested time
    local gcd_period_count = gmp.new_f()
    gmp.f_set_d(gcd_period_count, req_time)
    gmp.f_div(gcd_period_count, gcd_period_count, gcd_period)

    -- Round down to the nearest integer
    gmp.f_floor(gcd_period_count, gcd_period_count)

    -- Multiply by the period to get a new start time in MX time base
    local new_start_time = gmp.new_f()
    gmp.f_mul(new_start_time, gcd_period_count, gcd_period)

    if config.mx_phase_debug_messages then
        gmp.sprintf(gmp_print_buf, "Predicted MX TS: %.20Ff", new_start_time)
        logger:log(logger.INFO, ffi.string(gmp_print_buf))
    end

    -- At this point we have an MX device specific timestamp that should be
    -- common to all of the sample rates being used on the MX chanels being
    -- used on all of the MX devices of the MX device specified.

    -- Convert from MX timebase to CX timebase by adding in the difference in
    -- the epochs between a CX and an MX
    local mxe = gmp.new_f()
    gmp.f_set_d(mxe, mx_epoch)
    gmp.f_add(new_start_time, new_start_time, mxe)

    -- Convert to ISO format
    local new_start_time_iso = ts.epoch_to_iso(new_start_time:get_d())

    if config.mx_phase_debug_messages then
        logger:log(logger.INFO,
            string.format("New start time: %s", new_start_time_iso))
    end

    -- Return new start time in ISO format
    return new_start_time_iso
end

return m
