require('lj_include')(
    'stdlib',
    'unistd',
    'time'
)

require'middleclass'
local net = require'net'
local ev = require'ev'
local ffi = require'ptpv2-ffi/cdata'
local C = ffi.C
require'ptpv2-ffi/internal-time'

C.nice(-5);

local ptpClock = class('ptpClock')

ptpClock.msg_type_map = {
    [C.PTP_MSG_TYPE_SYNC] = 'SYNC',
    [C.PTP_MSG_TYPE_DELAY_REQ] = 'DELAY REQUEST',
    [C.PTP_MSG_TYPE_PDELAY_REQ] = 'PEER DELAY REQUEST',
    [C.PTP_MSG_TYPE_PDELAY_RESP] = 'PEER DELAY RESPONSE',
    [C.PTP_MSG_TYPE_FOLLOW_UP] = 'FOLLOW UP',
    [C.PTP_MSG_TYPE_DELAY_RESP] = 'DELAY RESPONSE',
    [C.PTP_MSG_TYPE_PDELAY_RESP_FOLLOW_UP] = 'PEER DELAY RESPONSE FOLLOW UP',
    [C.PTP_MSG_TYPE_ANNOUNCE] = 'ANNOUNCE',
    [C.PTP_MSG_TYPE_SIGNALING] = 'SIGNALING',
    [C.PTP_MSG_TYPE_MANAGEMENT] = 'MANAGEMENT',
}

ptpClock.state_map = {
    [C.PTP_STATE_LISTENING] = 'LISTEN',
    [C.PTP_STATE_MASTER] = 'MASTER',
    [C.PTP_STATE_SLAVE] = 'SLAVE',
    [C.PTP_STATE_INITIALIZING] = 'INITIALIZING',
    [C.PTP_STATE_PRE_MASTER] = 'PREMASTER',
    [C.PTP_STATE_PASSIVE] = 'PASSIVE',
    [C.PTP_STATE_FAULTY] = 'FAULTY',
}

function ptpClock:to_faulty()
    local clock_info = self.clock_info
    clock_info.state = C.PTP_STATE_FAULTY
end

function ptpClock:set_log_level(level, enable)
    self.log_level[level] = enable
end

function ptpClock:log(level, msg)
    if level == 'error' then
        self.logger:log(self.logger.ERR, msg)
    elseif level == 'debug' then
    else
        self.logger:log(self.logger.INFO, msg)
    end
    if level == 'error' then
        self:to_faulty()
    end
end

function ptpClock:fixup_correction(hll)
    hll.fractional_nanoseconds = C.ntohs(hll.fractional_nanoseconds)
    hll.nanoseconds.msb = C.ntohs(hll.nanoseconds.msb)
    hll.nanoseconds.lsb = C.ntohl(hll.nanoseconds.lsb)
    return hll
end

-- Table of function handlers for handling incoming messages
ptpClock.handler = {}

-- Table of functons used to send PTP messages
ptpClock.send = {}

local hinfo = {
    [C.PTP_MSG_TYPE_ANNOUNCE] = 'announce',
    [C.PTP_MSG_TYPE_SYNC] = 'sync',
    [C.PTP_MSG_TYPE_FOLLOW_UP] = 'followup',
    [C.PTP_MSG_TYPE_MANAGEMENT] = 'management',
    [C.PTP_MSG_TYPE_SIGNALING] = 'signaling',
    [C.PTP_MSG_TYPE_DELAY_REQ] = 'delay-request',
    [C.PTP_MSG_TYPE_DELAY_RESP] = 'delay-response',
}
for t, f in pairs(hinfo) do
    local _f = 'ptpv2-ffi/' .. f
    local m = require(_f)
    ptpClock.handler[t] = m.handle
    ptpClock.send[t] = m.send
end

local bmc = require('ptpv2-ffi/bmc')
for k, v in pairs(bmc) do
    ptpClock[k] = v
end

local function normalize_time(s, ns)
    while ns <= -1000000000 do
        ns = ns + 1000000000
        s = s - 1
    end
    while ns >= 1000000000 do
        ns = ns - 1000000000
        s = s + 1
    end
    if s < 0 and ns > 0 then
        ns = ns - 1000000000
        s = s + 1
    elseif s > 0 and ns < 0 then
        ns = ns + 1000000000
        s = s - 1
    end
    return {seconds = s, nanoseconds = ns}
end

function ptpClock:add_time(t1, t2)
    t1.seconds = t1.seconds + t2.seconds
    t1.nanoseconds = t1.nanoseconds + t2.nanoseconds
    return normalize_time(t1.seconds, t1.nanoseconds)
end

-- Function used to determine whether a received message is from the current
-- master
function ptpClock:is_from_current_master()
    local ci = self.clock_info
    local hdr = ffi.cast('msg_hdr_t *', ci.ibuf)
    if ci.parent_port_identity.port_number ==
        hdr.source_port_identity.port_number and
        C.memcmp(ci.parent_port_identity.clock_identity,
            hdr.source_port_identity.clock_identity,
            ffi.sizeof(ci.parent_port_identity.clock_identity)) == 0 then
        return true
    end
end

-- Function used to process ALL incoming data
function ptpClock:process_message(buf, blen, host, serv, time)
    local clock_info = self.clock_info
    local msg_hdr = clock_info.msg_hdr

--[=[
    -- Convert the time stamp to time.  If nothing is returned, the the time
    -- stamp was a raw hw timestamp and the conversion happened within the
    -- argument passed.
    local s, ns = self.timing_iface:raw_ts_to_time(time)
    if s and ns then
        time.seconds = s
        time.nanoseconds = ns
    end
]=]

    local version = bit.band(msg_hdr.reserved_version_ptp, 0xf)
    if version ~= clock_info.version_number then
        -- Ignore the message if it isn't the version we expect
        return
    end
    if msg_hdr.domain_number ~= clock_info.domain_number then
        -- Ignore the message if it isn't from the domain we expect
        return
    end

    local msg_is_from_me
    if clock_info.port_identity.port_number ==
        msg_hdr.source_port_identity.port_number and
        C.memcmp(clock_info.port_identity.clock_identity,
            msg_hdr.source_port_identity.clock_identity,
            ffi.sizeof(clock_info.port_identity.clock_identity)) == 0 then
        msg_is_from_me = true
    else
        time = time + self.inbound_latency
    end

    local msg_type = bit.band(msg_hdr.transport_specific_message_type, 0xf)

    local handler = self.handler[msg_type]
    if handler then
        handler(self, time, msg_is_from_me)
    end
end

-- Function used to get the MAC address of the specified interface and store it
-- in an array
local function get_mac(iface, to)
    local fh = io.popen('ip link show dev ' .. iface)
    if fh then
        for line in fh:lines() do
            local t =
                {line:match('link/ether (.-):(.-):(.-):(.-):(.-):(.-) brd')}
            if t[1] then
                for i, v in ipairs(t) do
                    to[i - 1] = bit.band(tonumber('0x' .. v), 0xff)
                end
                fh:close()
                return true
            end
        end
        fh:close()
    end
end

function ptpClock:init_clock()
    local ci = self.clock_info

    ci.master_to_slave_delay.seconds = 0
    ci.master_to_slave_delay.nanoseconds = 0
    ci.slave_to_master_delay.seconds = 0
    ci.slave_to_master_delay.nanoseconds = 0
--[[
    -- From original
    -- We don't want to do this because it causes nasty jumps in sync
    if ci.no_adjust == 0 then
        self:adj_freq(0)
    end
]]
end

function ptpClock:build_msg_header()
    local ci = self.clock_info

    local hdr = ffi.cast('msg_hdr_t *', ci.ibuf)

    ffi.fill(hdr, ffi.sizeof('msg_hdr_t'))
    hdr.transport_specific_message_type = 0
    hdr.reserved_version_ptp = ci.version_number
    hdr.domain_number = ci.domain_number
    if ci.two_step_flag ~= 0 then
        hdr.flag_field[0] = C.PTP_TWO_STEP_FLAG
    end
    ffi.copy(hdr.source_port_identity.clock_identity,
        ci.port_identity.clock_identity,
        ffi.sizeof(hdr.source_port_identity.clock_identity))
    hdr.source_port_identity.port_number = ci.port_identity.port_number
    hdr.log_message_interval = 0x7f
end

local state = require'ptpv2-ffi/state'
local th = require'ptpv2-ffi/timer-handlers'

ptpClock.to_state = state.to_state


ptpClock.announce_receipt_timer_handler = th.announce_receipt_timer_handler
ptpClock.announce_interval_timer_handler = th.announce_interval_timer_handler
ptpClock.sync_interval_timer_handler = th.sync_interval_timer_handler
ptpClock.delay_req_interval_timer_handler = th.delay_req_interval_timer_handler

ptpClock.send_sync = ptpClock.send[C.PTP_MSG_TYPE_SYNC]

function ptpClock:update_delay(correction_field)
    local ci = self.clock_info

    if ci.offset_first_updated ~= 0 then
        self.delay_sm = self.delay_req_receive_time -
            self.delay_req_send_time - correction_field

--[[
        -- Code from original
        if ci.max_delay ~= 0 then
            if delay_sm.seconds ~= 0 then
                self:log("error", "update delay aborted, delay > 1 second")
            elseif delay_sm.nanoseconds > ci.max_delay then
                self:log("error", string.format(
                    "update delay aborted, delay %d greater than administratively set maximum of %d", delay_sm.nanoseconds, ci.max_delay))
            end
            return
        end
]]

        self.mean_path_delay = (self.delay_sm + self.delay_ms) / 2
        ci.offset_first_updated = 0

        if self.stat_cb then
            self.stat_cb(self:get_stats())
        end
    end
end

function ptpClock:update_clock()
    local ci = self.clock_info
    local nanos = math.abs((self.offset_from_master.time.seconds * 1000000000) +
       self.offset_from_master.time.nanoseconds)
    local jump_amount = 100000
    local s = self[C.EVENT_PORT]
    local hw_stamped = s:is_hw_timestamped()
    if not hw_stamped then
        jump_amount = 10000000
    end
    -- If off by more than jump amount, jump time
    if nanos >= jump_amount then
        self.timing_iface:PTPServo(0, true)
        self.timing_iface.timing:AdjustClock(
            -self.offset_from_master.time.seconds,
            -self.offset_from_master.time.nanoseconds, 'PTP')
        self:init_clock()
        if self.stat_cb then
            self.stat_cb(self:get_stats())
        end
        return
    end
    self.timing_iface:PTPServo(self.offset_from_master.time.nanoseconds)
    if self.stat_cb then
        self.stat_cb(self:get_stats())
    end
end

function ptpClock:update_offset(send_time, correction_field)
    local ci = self.clock_info
    local recv_time = self.sync_receive_time
    local master_to_slave_delay = recv_time - send_time
--[[
    -- Code from original
    if ci.max_delay ~= 0 then
        if master_to_slave_delay.seconds ~= 0 then
            self:log("error", "update offset aborted, delay > 1 second")
        elseif master_to_slave_delay.nanoseconds > ci.max_delay then
            self:log("error", string.format(
                "update offset aborted, delay %d greater than administratively set maximum of %d", master_to_slave_delay.nanoseconds, ci.max_delay))
        end
        return
    end
]]

    self.master_to_slave_delay = master_to_slave_delay - correction_field

    self.delay_ms = self.master_to_slave_delay

    if ci.e2e_mode == 0 then
        self.offset_from_master = self.master_to_slave_delay -
            self.peer_mean_path_delay
    else
        self.offset_from_master = self.master_to_slave_delay -
            self.mean_path_delay
    end

    if ci.offset_first_updated == 0 then
        ci.offset_first_updated = 1
    end
end

function ptpClock:get_stats()
    local ci = self.clock_info
    return {
        state = self.state_map[self.clock_info.state],
        offset_from_master = {self.offset_from_master:get_time()},
        one_way_delay = {self.mean_path_delay:get_time()},
        slave_to_master_delay = {self.delay_sm:get_time()},
        master_to_slave_delay = {self.delay_ms:get_time()},
    }
end

function ptpClock:init_sockets()
    for _, si in ipairs(self.sock_info) do
        if self[si.port] then
            C.close(self[si.port].sock)
            self[si.port] = nil
        end

        local socket = net.socket{port = si.port, type = C.SOCK_DGRAM}
        if not socket then
            self:log("error",
                string.format("Unable to create socket on port %d",
                    si.port))
            return
        end

        self[si.port] = socket
        socket:start(ev.Loop.default)
        socket:bind{interface = self.iface}
        if si.port == C.EVENT_PORT then
            -- TX timestamps needed on event port
            socket:timestamping(true, true)
        else
            socket:timestamping(true)
        end
        socket.ttl = 5
        socket:join_multicast_group(self.mcast_group)

        local function read_func(loop, io)
            local s = socket
            local buf, blen, host, serv, time, rawcount = s:recvfrom(true)

            -- We get a nil buffer if an EAGAIN from the recvfrom happens OR
            -- we got a TX packet timestamp.  In that case, the transmit packet
            -- timestamp callback function passed in the last sendto call would
            -- have been called BEFORE recvfrom returns.  NOTE, buf is a native
            -- C buffer.
            if buf == nil then return end

            if rawcount then
                time = InternalTime:new(
                    self.timing_iface:raw_ts_to_time(rawcount))
            end

            if blen < ffi.sizeof('msg_hdr_t') then
                self:log("error",
                    string.format("Received less than sizeof header"))
                return
            end
            ffi.copy(self.clock_info.ibuf, buf, blen)
            self:process_message(buf, blen, host, serv, time)
        end
        local rcb = ev.IO.new(read_func, socket.sock, ev.READ)
        rcb:start(self.loop)
        self.rcb[si.name] = rcb

        -- Set the address we are going to be sending to
        socket:set_sendto_addr("ff02::181", si.port)
    end

--[[
    Ok, so you may be wondering what the code below this comment is all about.
    This bit of code is a workaround for what I believe is a issue with
    transmit packet timestamps in one or more of, gianfar driver, ip stack,
    udp stack, sockets, libev, or lua-ev.

    We were experiencing an extremely rare occurrance of the PTP daemon suddenly
    using all available processor time.  I was fortunate enough to be able to
    run strace when one of these instances occurred and noted that read events
    indicating data available on the "event" socket.  However, when the read
    from the socket was attempted, it would fail with EAGAIN.  Then another
    read event would immediately occur and the daemon would essentially be
    stuck in this infinite loop.

    As noted above, this issue was occurring on the "event" socket.  That
    particular socket is the socket that has transmit packet timestamping
    enabled.  Transmit packet timestamps in linux are retrieved from a sockets
    error queue.  Availability of the transmit packet timestamp is indicated
    via a read event.  When the read event occurs, readmsg with MSG_ERRQUEUE
    must be used to retrieve the timestamp.

    What I noticed happening was the PTP daemon was expecting the next read
    event to be for a transmit packet timestamp but recvmsg was returning an
    error indicating that no error (timestamp) message (ENOMSG) was available.

    I thought there could be a race condition whereby after transmitting a
    packet, an incoming packet on the same socket got queued before the error
    message packet was queued.  I added code that would attempt to read a
    normal packet if ENOMSG was returned when we attempted to read a transmit
    packet timestamp.  However, that also resulted in an EAGAIN return code
    from recvmsg.  Basically, at this point no matter whether the daemon tried
    to read, data or the error queue from the socket, recvmsg always returned
    either EAGAIN or ENOMSG.

    As stated above, I believe the issue is actually an issue with one or more
    of the layers under the PTP daemon.

    While working on another issue, I stumbled across something that would
    force the fail condition to occur every time.  It was a simple matter of
    sending SIGSTOP, wait a bit, then send SIGCONT to the PTP daemon.  Doing
    this caused a condition that, if not identical, was very similar to the
    observed rarely occurring problem discussed above.

    With the ability to cause the problem to occur I was able to determine that
    my previous observations about read events occuring but no error queue or
    normal data being returned by recvmsg was actually happening and the only
    theoritical explanation was some issue with underlying software.

    It occurred to me at that point that all of this communication is UDP and
    thus by defintion, is not reliable.  I decided at that point that it may be
    possible to recover from the situation by simply closing the offending
    sockets and opening new ones in their places without no affect the state of
    the PTP clock.

    So, that's what the code below is all about.  When ENOMSG is detected on the
    "event" socket, we simply close the existing sockets and replace them with
    new ones.  This all seems to work for the test case.  Now, one thing to
    keep in mind, all of this assumes that the rarely occurring asynchronous
    case is actually the same sort of thing as what is caused by my "forced to
    failure" case.  Hopefully it is and this workaround will solve the problem
    once and forever.

    It's probably worth mentioning that when we get time, it may be worth
    trying to figure out if/what/where the issue is in the underlying software.
    Until then, this seems to get done what needs to be done.
]]
    local socket = self[C.EVENT_PORT]
    socket.error_cb = function(errno)
            if errno == C.ENOMSG then
                self:log("info", "PTP Creating new sockets due to ENOMSG")
                for _, rcb in pairs(self.rcb) do
                    rcb:stop(self.loop)
                end
                self:init_sockets()
            end
        end
end

--function ptpClock:initialize(loop, timing, iface, mcast_group)
function ptpClock:initialize(args)
    assert(args.loop ~= nil,
        'loop is required for ptp clock')
    assert(args.timing ~= nil,
        'timing interface object is required for ptp clock')
    self.loop = args.loop
    self.timing_iface = args.timing
    self.rcb = {}
    local iface = args.iface or 'eth1'
    self.iface = iface
    self.mcast_group = args.mcast_group or "ff02::181"
    self.log_level = args.log_level or {error = true, stats = true}
    self.log_level.error = true
    self.logger = args.log

    self.pdelay_req_receive_time = InternalTime:new(0, 0)
    self.pdelay_req_send_time = InternalTime:new(0, 0)
    self.pdelay_resp_receive_time = InternalTime:new(0, 0)
    self.pdelay_resp_send_time = InternalTime:new(0, 0)
    self.sync_receive_time = InternalTime:new(0, 0)
    self.delay_req_receive_time = InternalTime:new(0, 0)
    self.delay_req_send_time = InternalTime:new(0, 0)

    self.delay_sm = InternalTime:new(0, 0)
    self.delay_ms = InternalTime:new(0, 0)

    self.peer_mean_path_delay = InternalTime:new(0, 0)
    self.mean_path_delay = InternalTime:new(0, 0)

    self.offset_from_master = InternalTime:new(0, 0)

    self.clock_info = ffi.cast('struct ptp_clock_info *',
        C.calloc(1, ffi.sizeof('struct ptp_clock_info')))

    local clock_info = self.clock_info
    clock_info.state = C.PTP_STATE_INITIALIZING
    clock_info.e2e_mode = 1

    clock_info.msg_hdr = ffi.cast('msg_hdr_t *', clock_info.ibuf)
    clock_info.announce_msg = ffi.cast('announce_msg_t *', clock_info.ibuf)
    clock_info.msg_hdr_save = ffi.cast('msg_hdr_t *',
        C.calloc(1, ffi.sizeof('msg_hdr_t')))

    if self.mcast_group:find(':') then
        self.clock_info.transport_is_ipv6 = 1
    end

    if not get_mac(iface, clock_info.port_uuid_field) then
        -- FIXME:  Error
        return
    end

    clock_info.two_step_flag = C.PTP_TWO_STEP_FLAG
    ffi.copy(clock_info.clock_identity, clock_info.port_uuid_field, 3)
    clock_info.clock_identity[3] = 0xff
    clock_info.clock_identity[4] = 0xff
    ffi.copy(clock_info.clock_identity + 5, clock_info.port_uuid_field + 3, 3)

    clock_info.number_ports = C.DEFAULT_NUMBER_PORTS

    local options = args.clock_options
    if not options then
        local ofile = loadfile('/etc/ptp-options.lua')
        if ofile then
            options = ofile()
        end
    end
    options = options or {}

    clock_info.clock_quality.clock_accuracy =
        options.clock_accuracy or C.DEFAULT_CLOCK_ACCURACY
    clock_info.clock_quality.clock_class =
        options.clock_class or C.DEFAULT_CLOCK_CLASS
    clock_info.clock_quality.offset_scaled_log_variance =
        options.clock_offset_scaled_log_variance or
        C.DEFAULT_CLOCK_VARIANCE
    clock_info.priority1 = options.priority1 or C.DEFAULT_PRIORITY1
    clock_info.priority2 = options.priority2 or C.DEFAULT_PRIORITY2
    clock_info.domain_number = options.domain_number or C.DEFAULT_DOMAIN_NUMBER
    clock_info.slave_only = options.slave_only or 0
    if clock_info.slave_only ~= 0 then
        options.clock_class = 255
    end
    clock_info.inbound_latency.seconds = options.inbound_latency_s or
        C.DEFAULT_INBOUND_LATENCY_S
    clock_info.inbound_latency.nanoseconds = options.inbound_latency_ns or
        C.DEFAULT_INBOUND_LATENCY_NS
    clock_info.outbound_latency.seconds = options.outbound_latency_s or
        C.DEFAULT_OUTBOUND_LATENCY_S
    clock_info.outbound_latency.nanoseconds = options.outbound_latency_ns or
        C.DEFAULT_OUTBOUND_LATENCY_NS

    self.inbound_latency = InternalTime:new(clock_info.inbound_latency.seconds,
        clock_info.inbound_latency.nanoseconds)
    self.outbound_latency =
        InternalTime:new(clock_info.outbound_latency.seconds,
        clock_info.outbound_latency.nanoseconds)

    ffi.copy(clock_info.port_identity.clock_identity, clock_info.clock_identity,
        ffi.sizeof(clock_info.port_identity.clock_identity))
    clock_info.port_identity.port_number = C.DEFAULT_NUMBER_PORTS

    -- Clear clockInfo.parent_port_identity

    clock_info.log_min_delay_req_interval = C.DEFAULT_DELAYREQ_INTERVAL

    -- Clear clock_info.peer_mean_path_delay
    -- Clear clock_info.mean_path_delay
    clock_info.log_announce_interval =
        options.log_announce_interval or C.DEFAULT_ANNOUNCE_INTERVAL
    clock_info.announce_receipt_timeout = C.DEFAULT_ANNOUNCE_RECEIPT_TIMEOUT
    clock_info.log_sync_interval =
        options.log_sync_interval or C.DEFAULT_SYNC_INTERVAL
    clock_info.delay_mechanism = C.DEFAULT_DELAY_MECHANISM
    clock_info.log_min_pdelay_req_interval = C.DEFAULT_PDELAYREQ_INTERVAL
    clock_info.version_number = C.PTP_VERSION_2

    C.srand(C.time(nil))
    clock_info.R = (C.rand() * 1.0) / C.RAND_MAX

    clock_info.number_foreign_records = 0
    clock_info.max_foreign_records =
        options.max_foreign_records or C.DEFAULT_MAX_FOREIGN_RECORDS
    clock_info.foreign = ffi.cast('foreign_master_record_t *',
        C.calloc(clock_info.max_foreign_records,
        ffi.sizeof('foreign_master_record_t')))

    clock_info.current_utc_offset = options.utc_offset or C.DEFAULT_UTC_OFFSET

    clock_info.sent_sync_sequence_id = 0
    clock_info.sent_announce_sequence_id = 0
    clock_info.waiting_for_follow = 0

    -- Message sizes adjusted to account for IPv6 padding
    clock_info.sizeof_sync_msg = ffi.offsetof('sync_msg_t', 'ip6_pad')
    clock_info.sizeof_followup_msg = ffi.offsetof('followup_msg_t', 'ip6_pad')
    clock_info.sizeof_announce_msg = ffi.offsetof('announce_msg_t', 'ip6_pad')
    clock_info.sizeof_delay_req_msg = ffi.offsetof('delay_req_msg_t', 'ip6_pad')
    clock_info.sizeof_delay_resp_msg =
        ffi.offsetof('delay_resp_msg_t', 'ip6_pad')

    self.sock_info = {
        {port = C.GENERAL_PORT, name = 'general_socket'},
        {port = C.EVENT_PORT, name = 'event_socket'},
    }

    self:init_sockets()

    -- Adjust message sizes if we are NOT using IPv6 transport
    if clock_info.transport_is_ipv6 == 0 then
        clock_info.sizeof_sync_msg = clock_info.sizeof_sync_msg - 2
        clock_info.sizeof_announce_msg = clock_info.sizeof_announce_msg - 2
        error("FINISH THIS")
    end

    -- Instantiate the assorted timers needed for the protocol
    local t = clock_info.announce_receipt_timeout *
        math.pow(2, clock_info.log_announce_interval)
    self.announce_receipt_timer = ev.Timer.new(function(l, t, e)
            local r, e = pcall(ptpClock.announce_receipt_timer_handler, self)
            if not r then
                self:log("error", string.format(
                    "Error in announce_receipt_timer_handler\n%s", e))
                self:to_state(C.PTP_STATE_FAULTY)
            end
        end, t)

    local t = math.pow(2, clock_info.log_sync_interval)
    self.sync_interval_timer = ev.Timer.new(function(l, t, e)
            local r, e = pcall(ptpClock.sync_interval_timer_handler, self)
            if not r then
                self:log("error", string.format(
                    "Error in sync_interval_timer_handler\n%s", e))
                self:to_state(C.PTP_STATE_FAULTY)
            end
        end, t)

    local t = math.pow(2, clock_info.log_announce_interval)
    self.announce_interval_timer = ev.Timer.new(function(l, t, e)
            local r, e = pcall(ptpClock.announce_interval_timer_handler, self)
            if not r then
                self:log("error", string.format(
                    "Error in announce_interval_timer_handler\n%s", e))
                self:to_state(C.PTP_STATE_FAULTY)
            end
        end, t)

    local t = math.pow(2, clock_info.log_min_delay_req_interval)
    self.delay_request_interval_timer = ev.Timer.new(function(l, t, e)
            local r, e = pcall(ptpClock.delay_req_interval_timer_handler, self)
            if not r then
                self:log("error", string.format(
                    "Error in delay_req_interval_timer_handler\n%s", e))
                self:to_state(C.PTP_STATE_FAULTY)
            end
        end, t)

    if options.p2p_mode then
        local t = math.pow(2, clock_info.log_min_pdelay_req_interval)
        self.pdelay_request_interval_timer = ev.Timer.new(function(l, t, e)
                local r, e =
                    pcall(ptpClock.pdelay_req_interval_timer_handler, self)
                if not r then
                    self:log("error", string.format(
                        "Error in pdelay_req_interval_timer_handler\n%s", e))
                    self:to_state(C.PTP_STATE_FAULTY)
                end
            end, t)
    end

    clock_info.foreign_record_best = 0
    clock_info.foreign_record_i = 0

    -- Clear clock_info.grandmaster_clock_quality
    -- Clear clock_info.grandmaster_clock_idenity

    clock_info.sent_delay_req_sequence_id = 0
    clock_info.sent_pdelay_req_sequence_id = 0

    -- Clear clock_info.delay_sm
    -- Clear clock_info.delay_ms
    -- Clear clock_info.last_sync_correction_field
    -- Clear clock_info.offset_from_master
    -- Clear clock_info.grandmaster_identity
    -- Clear clock_info.master_to_slave_delay
    -- Clear clock_info.slave_to_master_delay
    -- Clear clock_info.pdelay_req_receive_time
    -- Clear clock_info.pdelay_req_send_time
    -- Clear clock_info.pdelay_resp_receive_time
    -- Clear clock_info.pdelay_resp_send_time
    -- Clear clock_info.sync_receive_time
    -- Clear clock_info.delay_req_receive_time
    -- Clear clock_info.delay_req_send_time

    clock_info.time_source = C.PTP_CLOCK_TYPE_INTERNAL_OSCILLATOR

    self:m1()
    self:to_state(C.PTP_STATE_LISTENING)

    return self
end

function ptpClock:shutdown()
    for _, v in ipairs(self.sock_info) do
        local s = self.rcb[v.name]
        if s then
            s:stop(self.loop)
        end
        local s = self[v.port]
        if s then
            s:stop(self.loop)
            s:close()
        end
    end
    for _, v in ipairs({
            'announce_receipt_timer',
            'sync_interval_timer',
            'announce_interval_timer',
            'delay_request_interval_timer',
            'pdelay_request_interval_timer',
        }) do
        local s = self[v]
        if s then
            s:stop(self.loop)
        end
    end
    if self.clock_info then
        C.free(self.clock_info)
    end
end

return ptpClock
