--[[
	Lua module file for exporting translated sdb sensors
	to an .sxr file for cx system use
--]]


local e = {}


function e:run( run_info ,fnames ,self_version)

	local ordered_blob ,sxr_blob = self:format_blob( run_info ,fnames ,self_version)
	print( "Written translated sensors into JSON blob, now finishing...")

	local fns = {}
	fns[ "/tmp/"..fnames.root] = sxr_blob
	fns[ "/tmp/sensor_delta"] = ordered_blob
	domovoi:write_files( fns)

	if run_info.trash > 0 then
		domovoi:incomplete( fnames.root_with_ext)
	end
	return sxr_blob
end


function e:format_blob( run_info ,fnames ,self_version)
	local date = os.date( '"%Y-%m-%d_%H:%M:%S"' )
	date = date:gsub( "_" ," " )

	--some lines are currently commented out awaiting a new
	--GUI version

	local summary = '\t{\n\t\t'
	.. '"category": "summary"'
	.. ',"database": "HBM sensor database"'
	.. ',"database_source_filename":"' .. fnames.root_with_ext

	summary = summary .. '","total_sensors": ' .. (run_info.passed + run_info.failed) .. ','
	.. '"supported_sensors": ' .. run_info.passed
	.. ',"unsupported_sensors": ' .. run_info.failed
	.. ',"import_time": ' .. date
	.. ',"database_version": '..run_info.about.Version
	.. ',"id": "summary"'
	.. ',"sxr_generator_version": "'..self_version..'"'
	.. ',"test_name": "'..fnames.root

	if arg[2] then
		summary = summary .. '","setup_id": "'..arg[2]..'"}'
	else
		summary = summary ..'"}'
	end

	local content = '"' 
	.. fnames.root ..  '":'
	.. self:lua_table_to_json( run_info.blob ,run_info.groups)
	local sxr_blob = '['..summary..',{'..content..'}]'
	local command_blob = '{'..content

	return command_blob ,sxr_blob
end


--Have to use table concat, really large catman files cause a 3min load time
--since concat (..) is expensive stack-wise
function e:lua_table_to_json( tble ,groups )
	local end_f = '\n'
	local chain = {}
	table.insert( chain ,'\n\t{\n\t\t')

	for i ,value in ipairs( tble ) do
		self:index_format( chain ,value ,end_f..'\t' ,groups ,i)
	end

	local result = table.concat( chain)
	result = result:match( '(.+),' ) .. '\n\t}\n'
	return result
end

--=== Helpers =====
	function e:index_format( chain ,value ,end_f ,groups ,index)
		if index == 'gi' or index == 'pi' then return end
		local format
		local s = ""

		if type(index) == 'number' then
			if value.gi then --sub GroupIndex for group name
				index = groups[ value.gi ].Name
				format = 'group'
			else
				index = value.description
				format = 'sensor'
			end
		else
			format = 'generic'
		end

		if not index:match( "Missing Group" ) then
			if not index:match( '".-"' ) then
				s = s .. '"' .. index .. '": '
			else
				s = s .. index .. ': '
			end
			table.insert( chain ,s)			
			self:value_format( chain ,value ,end_f..'\t' ,groups ,format )
		end
	end

	function e:value_format( chain ,value ,end_f ,groups ,format)
		local dtype = type( value )
		local s

		if dtype == "number" or dtype == "boolean" then
			s = table.concat( { tostring(value) ,end_f ,','})
		elseif dtype == "string" and not value:match( '".-"' ) then
			s = table.concat( { '"' ,value ,'"' ,end_f ,','})
		elseif dtype == 'string' then
			s = table.concat( { value ,end_f ,','})
		elseif dtype == "table" then
			self:table_format( chain, value , end_f .. '\t' , groups ,format)
			return
		end
		table.insert( chain ,s)
	end

	function e:table_format( chain ,tble ,end_f ,groups ,format)
		local count = 0
		local is_iv = false
		if type( format) == 'number' then format = 'generic' end

		for k,v in pairs( tble) do
			if k ~= 'gi' and k~='pi' then --gi and pi aren't counted as values
				count = count + 1
			end
			if type(k) == 'number' then
				is_iv = true
				if count > 2 then break end
			end
		end
		local is_empty = count == 0

		local setups = {
			group 	=	{ca='{' ,cb='}' ,sort=ipairs ,fn=self.index_format}
			,sensor =	{ca='{' ,cb='}' ,sort=pairs ,fn=self.index_format}
			,gen_iv	=	{ca='[' ,cb=']' ,sort=ipairs ,fn=self.value_format}
			,gen_kv	=	{ca='{' ,cb='}' ,sort=pairs ,fn=self.index_format}
		}

		local setup
		if format == 'generic' then
			if is_iv then
				setup = setups.gen_iv
			else
				setup = setups.gen_kv
			end
		else
			setup = setups[ format]
		end

		local s
		if not is_empty then
			s = table.concat({ setup.ca ,end_f })
			table.insert( chain ,s)
			for si ,sv in setup.sort( tble) do
				setup.fn( self ,chain ,sv ,end_f:sub(0,(#end_f-1)) ,groups ,si)
			end
			chain[#chain] = string.match( chain[#chain] ,"(.+)\t,")..setup.cb
				..end_f:sub(0,(#end_f-1))..','

		else
			s = '[]'..end_f..','
			table.insert( chain ,s )  --empty list/array
		end
	end


return e
