--[[
    Holds majority of the sxr import rules, per document SomatXR HBM SDB Sensor Import (Rev 10), 25-Oct-15 (DFD)
    supported_bridges are the supported bridge type enums
    scale_types are the different types of sensor scaling we support. Each has it's own set of rules.
    same for ss. Each supported sensor has a number of scaling types associated with it, hence the previous data struct.
--]]

local ss

--================== Supported Bridge Bridge Enumerations =====
local supported_bridges =
{ 362 ,363 ,364 ,365 ,366 ,367 ,368 ,369 ,371 ,372 ,350 ,351 ,352 ,353 ,354}

local mx1601 = "source/hbm_mx1601b" 
local mx1615 = "source/hbm_mx1615b"
local mx1609 = "source/hbm_mx1609b" 
local mx840 = "source/hbm_mx840b"
local mx411 = "source/hbm_mx411b"

--loops and calls all overlaps -> categories -> fields
--for _category_specific setup
--there's some extra stuff here to check for successive failures 
--in the same field- since we can only fail a sensor once
--it's failed in the same field for each category (if multiple)
local function setup_category_specifics( sensor ,sxr_row ,args)
  sxr_row._category_specific = {} 
  local failures = {}
  local cc = 0 
  for category ,fields in pairs( ss[ sensor].overlaps) do
    sxr_row._category_specific[ category] = {}
    cc = cc + 1

    for field ,fn in pairs( fields) do
      sxr_row[ field] = nil
      local success = fn( sxr_row ,args)
      if not success then
        failures[ category] = true
      end
    end
  end
  print("Category Count: " ,cc)

  for category,_ in pairs( failures) do
    cc = cc - 1
    sxr_row._category_specific[ category] = nil
  end

  print( "Passed Categories: " ,cc)
  return cc == 0
end


--================== Supported Scale Types ======================
local scale_types = {
  StrainGage = {    enums = {0}
    ,sxr_name="strain_gage"
    ,required = function( cal ,l)
      local failed = false
      cal.gage_factor = l.GageFactor
      cal.bridge_factor = l.BridgeFactor
      if cal.gage_factor < 0.1
      or cal.gage_factor > 10 
      or cal.bridge_factor == 0 then
        failed = true
        print( "Failed strain gage check.")
      end
      return failed
    end
  }

  ,ZeroSpan = {     enums = {0,1}
    ,sxr_name="defined_zero_span" 
    ,required = function( cal ,l)
        local failed = false
        cal.electrical_zero = l.P1X
        cal.electrical_span = l.P2X - l.P1X
        cal.physical_span = l.P2Y
        if cal.electrical_span == 0 or l.P2Y == 0 then
            failed = true
            print( "Failed ZeroSpan check." )
        end
    --cal.max_electrical = 
--      math.abs(cal.electrical_span) +
--      math.abs(cal.electrical_zero)
      return failed
    end
  }

  ,Table2 = {   enums = {2}
    ,sxr_name = "defined_two_point"
    ,required = function( cal , l )
      local failed = false
      print( "Scale data: " ,l.ScaleData1 )
      local i,e1,p1,e2,p2 = string.match( l.ScaleData1 
        ,'([-.%d]+)~([-.%d]+)~([-.%d]+)~([-.%d]+)~([-.%d]+)' )
        print(tostring(i).."\t"..tostring(e1).."\t"..tostring(p1).."\t"..tostring(e2).."\t"..tostring(p2))
      if i ~= '4' then
        print( 'Failed Table2 2Point check, ',i )
        return true
      end
      cal.electrical = {}
      cal.physical = {}
      cal.electrical[1] = tonumber( e1 )
      cal.electrical[2] = tonumber( e2 )
      cal.physical[1] = tonumber( p1 )
      cal.physical[2] = tonumber( p2 )
      print("CHECK: " , e1 , e2 , p1 , p2 )
      if e1 == e2 or p1 == p2 then
        failed = true
        print( "Failed Table2 check." )
      end
      return failed
    end
  }

  ,TwoPoint = {     enums = {32} 
    ,sxr_name="defined_two_point" 
    ,required = function( cal , l )
      local failed = false
      cal.electrical = {}
      cal.electrical[1] = l.P1X
      cal.electrical[2] = l.P2X
      cal.physical = {}
      cal.physical[1] = l.P1Y
      cal.physical[2] = l.P2Y
      if l.P1X == l.P2X or l.P1Y == l.P2Y then
        failed = true
        print( "Failing TwoPoint check.")
      end
      return failed
    end
  }

  ,Gain =       {   enums = {64}
    ,sxr_name = "defined_slope_intercept"
    ,required = function( cal , l )
      print( l.ScaleData5)
      cal.intercept ,cal.slope = l.ScaleData5:match(  '([%d%.]+)~([%d%.]+)')
      cal.intercept = tonumber( cal.intercept)
      cal.slope = tonumber( cal.slope)
      return false
    end
  }

  ,Internal =   {   enums = {0,256,512}
    ,sxr_name = "internal"
    ,required = function( cal , l ) --internal has no requirements, per spec's given by Dennis
      return false
    end
  }

}

--==================== Supported Sensor Types =======================================

--Supported Sensors
ss = {}

ss.StrainGage = {   
  name = "StrainGage"
  ,enum = 1
  ,categories = { mx1615 ,mx840, mx411}
  ,elec_units = "mV/V"
  ,s_types = {
    scale_types.ZeroSpan
    ,scale_types.StrainGage
    ,scale_types.TwoPoint
    ,scale_types.Table2
    ,scale_types.Gain
  }
  ,overlaps = {
    [mx1615] = {
      excitation_range = function( s ,args)
        local l = args.l
        local er
        local sdb_ex
        local ex
        local max
        local t = l.SensorType
        local ref = {}
        --print_r(l)

        if l.Resistance == 0 then l.Resistance = 350
        end
         if l.ScaleType == 0 then
            if l.GageFactor == 0 then -- ZeroSpan
                max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
                s.calibration[3] = nil
            else -- StrainGage
                if l.Range > 1000 then max = 4
                else max = l.Range end
            end
        elseif  l.ScaleType == 64 then --Gain
            max = l.Range
            s.calibration[3] = nil
        elseif l.ScaleType == 1 then --ZeroSpan
            max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
            s.calibration[3] = nil
        elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
            max = l.P2X
            s.calibration[3] = nil
        else
            for k,v in pairs(s._supported_categories) do
                if v == mx1615 then table.remove(s._supported_categories,k)end
            end
            return false
        end

        if l.Excitation >= 5 then sdb_ex = 5.0
        elseif l.Excitation >= 2.5 then sdb_ex = 2.5
        elseif l.Excitation >= 1.0 then sdb_ex = 1.0
        elseif l.Excitation >= 0.5 then sdb_ex = 0.5
        else 
          print("Failed excitation range 1615 " .. l.Description .. "\texcitation=\t" .. l.Excitation)
            for k,v in pairs(s._supported_categories) do
                if v == mx1615 then table.remove(s._supported_categories,k)end
            end
          return false
        end

        if max > 40 then
          print("Failed max_electrical. max > 40 max=  ".. tostring(max) .. l.Description )
            for k,v in pairs(s._supported_categories) do
                if v == mx1615 then table.remove(s._supported_categories,k)end
            end
          return false
        elseif max > 20 then
          if t==368 or t==369 or t==371 or t==372 or t==352 then --quarter bridges
              ex = 0.5
          else
          print("max > 20 and not quarter " .. l.Description )
            for k,v in pairs(s._supported_categories) do
                if v == mx1615 then table.remove(s._supported_categories,k)end
            end
            return false
          end
        elseif max > 8 then
          if sdb_ex < 1.0 then ex = sdb_ex
          else ex = 1.0
          end
        elseif max > 4 then
          if sdb_ex < 2.5 then ex = sdb_ex
          else ex = 2.5
          end
        else
          if sdb_ex < 5.0 then ex = sdb_ex
          else ex = 5.0
          end
        end

        if s.bridge_resistance < 110 and sdb_ex > 1.0 then ex = 1.0
        elseif s.bridge_resistance < 300 and sdb_ex > 2.5 then ex = 2.5
        end

        if (t~=368 and t~=369 and t~=371 and t~=372 and t~=352) and ex == 0.5 then --not quarter
            print("not quarter bridge w/ excitation = .5")
            for k,v in pairs(s._supported_categories) do
                if v == mx1615 then table.remove(s._supported_categories,k)end
            end
            return false
        end

        if ex == 5.0 then
          er = "5.0v"
          ref.max = 4
          ref.min = -4
        elseif ex == 2.5 then
          er = "2.5v"
          ref.max = 8
          ref.min = -8
        elseif ex == 1.0 then
          er = "1.0v"
          ref.max = 20
          ref.min = -20
        elseif ex == 0.5 then
          er = "0.5v"
          ref.max = 40
          ref.min = -40
        end

        ref.mode = s.sensor_input.mode
        ref.type = s.sensor_input.type
        ref.regulation = s.sensor_input.regulation
        ref.wires = s.sensor_input.wires
        s._category_specific[ mx1615].excitation_range = er
        s._category_specific[ mx1615].sensor_input = ref
        return true     
      end
      ,excitation_mode = function( s ,args)
        local l = args.l
        local em
        if l.ExcitationFrequency == -1 then
          em = 'dc'
        else
          em = 'ac_1200_hz_square'
        end
        s._category_specific[ mx1615].excitation_mode = em
        return true
      end
    }
    ,[mx840] = {
      excitation_range = function( s ,args)
        local l = args.l
        local er
        local sdb_ex
        local ex
        local gain_low, gain_low_low = false, false
        local max 
        local t = l.SensorType
        local ref = {}

        if l.Resistance == 0 then l.Resistance = 350
        end
        if l.ScaleType == 0 then
            if l.GageFactor == 0 then -- ZeroSpan
                max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
                s.calibration[3] = nil
            else -- StrainGage
                if l.Range > 1000 then max = 4
                else max = l.Range end
            end
        elseif  l.ScaleType == 64 then --Gain
            max = l.Range
            s.calibration[3] = nil
        elseif l.ScaleType == 1 then --ZeroSpan
            max = math.abs(l.P2X - l.P1X) + math.abs(l.P1X)
            s.calibration[3] = nil
        elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
            max = l.P2X
            s.calibration[3] = nil
        else
            for k,v in pairs(s._supported_categories) do
                if v == mx840 then table.remove(s._supported_categories,k)end
            end
            return false
        end
        if t == 368 or t == 369 or t == 371 or t == 372 or t == 352 then --quarter bridge
            for k,v in pairs(s._supported_categories) do
                if v == mx840 then table.remove(s._supported_categories,k)end
            end
          print("quarter bridge 840")
          return false
        end

        if l.ExcitationFrequency == -1 and (t==362 or t==363 or t==364 or t==353 or t==350) then--full/dc
          if l.Excitation >= 2.5 then 
            sdb_ex = 2.5
          elseif l.Excitation >= 1.0 then
            sdb_ex = 1.0
          else 
            print("Failed excitation range 840 " .. l.Description .. "\texcitation=\t" .. l.Excitation)
            for k,v in pairs(s._supported_categories) do
                if v == mx840 then table.remove(s._supported_categories,k)end
            end
            return false
          end

          if max > 1000 then
            print("max > 1000")
            for k,v in pairs(s._supported_categories) do
                if v == mx840 then table.remove(s._supported_categories,k)end
            end
            return false
          elseif max > 120 then
            gain_low_low = true
            if sdb_ex < 2.5 then
                print("max < 2.5")
                for k,v in pairs(s._supported_categories) do
                    if v == mx840 then table.remove(s._supported_categories,k)end
                end
                return false
            end
            ex = 2.5
          elseif max > 5 then
            gain_low = true
            if sdb_ex < 2.5 then
                print("excitation < 2.5") 
                for k,v in pairs(s._supported_categories) do
                    if v == mx840 then table.remove(s._supported_categories,k)end
                end
                return false 
            end
            ex = 2.5
          else 
            if sdb_ex < 2.5 then
               ex = sdb_ex
            else
               ex = 2.5
            end
          end
        print("Resistance:\t" .. tostring(l.Resistance) .. "\tex\t" .. tostring(ex) .. "1")

          if gain_low or gain_low_low then
            if s.bridge_resistance < 300 then
                print("gain_low and resistance < 300")
                for k,v in pairs(s._supported_categories) do
                    if v == mx840 then table.remove(s._supported_categories,k)end
                end
                return false
            end
          else 
            if s.bridge_resistance < 300 then
              if ex > 1.0 then
                ex = 1.0
              end
            end
          end

          if ex == 2.5 then
            if gain_low_low then
              ref.max = 1000
              ref.min = -1000
            elseif gain_low then
              ref.max = 120
              ref.min = -120
            else 
              ref.max = 5 
              ref.min = -5
            end
          else 
            ref.max = 12.5
            ref.min = -12.5
          end

          if ex == 2.5 then er = "2.5v"
          else er = "1.0v"
          end

          ref.mode = s.sensor_input.mode
          ref.type = s.sensor_input.type
          ref.regulation = s.sensor_input.regulation
          ref.wires = s.sensor_input.wires
          s._category_specific[ mx840].excitation_range = er
          s._category_specific[ mx840].sensor_input = ref
          return true   
        else
          if l.Excitation >= 2.5 then 
            sdb_ex = 2.5
          elseif l.Excitation >= 1.0 then
            sdb_ex = 1.0
          else 
            print("Failed excitation range 840 " .. l.Description .. "\texcitation=\t" .. l.Excitation)
            for k,v in pairs(s._supported_categories) do
                if v == mx840 then table.remove(s._supported_categories,k)end
            end
            return false
          end

          if max > 12.5 then
            for k,v in pairs(s._supported_categories) do
                if v == mx840 then table.remove(s._supported_categories,k)end
            end
            return false
          elseif max > 5 then
            ex = 1.0
          else
            if sdb_ex < 2.5 then
              ex = sdb_ex
            else
              ex = 2.5
            end
          end
        print("Resistance:\t" .. tostring(l.Resistance) .. "\tex\t" .. tostring(ex))

          if s.bridge_resistance < 300 then
            if ex > 1.0 then
              ex = 1.0
            end
          end

          if ex == 2.5 then
            ref.max = 5
            ref.min = -5
          else
            ref.max = 12.5
            ref.min = -12.5
          end

          if ex == 2.5 then er = "2.5v"
          else er = "1.0v"
          end

          ref.mode = s.sensor_input.mode
          ref.type = s.sensor_input.type
          ref.regulation = s.sensor_input.regulation
          ref.wires = s.sensor_input.wires
          s._category_specific[ mx840].excitation_range = er
          s._category_specific[ mx840].sensor_input = ref
          return true   
        end 
      end
      ,excitation_mode = function( s ,args)
        local l = args.l
        local em
        print("excitation Freq =" .. l.ExcitationFrequency)
        if l.ExcitationFrequency == -1 then
          em = 'dc'
        else
          em = 'ac_4800_hz_sine'
        end
        s._category_specific[ mx840].excitation_mode = em
        return true
      end
    }
    ,[mx411] = {
      excitation_range = function( s ,args)
        local l = args.l
        local er
        local sdb_ex
        local ex
        local gain_low = false
        local max
        local ref = {}
        local t = l.SensorType

        if l.Resistance == 0 then l.Resistance = 350
        end
         if l.ScaleType == 0 then
            if l.GageFactor == 0 then -- ZeroSpan
                max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
                s.calibration[3] = nil
            else -- StrainGage
                if l.Range > 1000 then max = 4
                else max = l.Range end
            end
        elseif  l.ScaleType == 64 then --Gain
            max = l.Range
            s.calibration[3] = nil
        elseif l.ScaleType == 1 then --ZeroSpan
            max = math.abs(l.P2X - l.P1X) + math.abs(l.P1X)
            s.calibration[3] = nil
        elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
            s.calibration[3] = nil
            max = l.P2X
        else
            for k,v in pairs(s._supported_categories) do
                if v == mx411 then table.remove(s._supported_categories,k)end
            end
            return false
        end
        if t == 368 or t == 369 or t == 371 or t == 372 or t == 352 then --quarter bridge
            print("quarter bridge 411")
            for k,v in pairs(s._supported_categories) do
                if v == mx411 then table.remove(s._supported_categories, k)end
            end
        return false
        end

        if l.ExcitationFrequency == -1 then
          if l.Excitation >= 7.5 then
            sdb_ex = 7.5
          elseif l.Excitation >= 5 then 
            sdb_ex = 5.0
          elseif l.Excitation >= 2.5 then 
            sdb_ex = 2.5
          elseif l.Excitation >= 1.0 then
            sdb_ex = 1.0
          else 
            for k,v in pairs(s._supported_categories) do
                if v == mx411 then table.remove(s._supported_categories, k)end
            end
            print("Failed excitation range 411 " .. l.Description .. "\texcitation=\t" .. l.Excitation)
            return false
          end

          if max > 100 then 
            return false
          elseif max > 50 then
            gain_low = true
            if sdb_ex < 2.5 then 
                for k,v in pairs(s._supported_categories) do
                    if v == mx411 then table.remove(s._supported_categories,k)end
                end
              return false
            end
            ex = 2.5
          elseif max > 20 then
            gain_low = true
            if sdb_ex < 5.0 then
              ex = sdb_ex
              if sdb_ex < 2.5 then
                for k,v in pairs(s._supported_categories) do
                    if v == mx411 then table.remove(s._supported_categories,k)end
                end
                return false
              end
            else 
              ex = 5.0
            end
          elseif max > 10 then
            ex = 1.0
          elseif max > 4 then
            if sdb_ex < 2.5 then
              ex = sdb_ex
            else 
              ex = 2.5
            end
          else
            if sdb_ex < 7.5 then 
              ex = sdb_ex
            else
              ex = 7.5
            end
          end

          if gain_low then
            if s.bridge_resistance < 110 then
                for k,v in pairs(s._supported_categories) do
                    if v == mx411 then table.remove(s._supported_categories,k)end
                end
              return false
            elseif s.bridge_resistance < 300 then
              if ex > 2.5 then
                ex = 2.5
              end
            end
          else
            if s.bridge_resistance < 110 then
              if ex > 1.0 then
                ex = 1.0
              end
            elseif s.bridge_resistance < 300 then
              if ex > 2.5 then
                ex = 2.5
              end
            end
          end

          if ex == 7.5 then
            ref.max = 4
            ref.min = -4
            er = "7.5v"
          elseif ex == 5.0 then
            if gain_low then
              ref.max = 50
              ref.min = -50
            else
              ref.max = 4
              ref.min = -4
            end
            er = "5.0v"
          elseif ex == 2.5 then
            if gain_low then
              ref.max = 100
              ref.min = -100
            else
              ref.max = 10
              ref.min = -10
            end
            er = "2.5v"
          elseif ex == 1.0 then
            ref.max = 20
            ref.min = -20
            er = "1.0v"
          end

          ref.mode = s.sensor_input.mode
          ref.type = s.sensor_input.type
          ref.regulation = s.sensor_input.regulation
          ref.wires = s.sensor_input.wires
          s._category_specific[ mx411].excitation_range = er
          s._category_specific[ mx411].sensor_input = ref
          return true
        else
          if l.Excitation >= 5.0 then
            sdb_ex = 5.0
          elseif l.Excitation >= 2.5 then
            sdb_ex = 2.5
          elseif l.Excitation >= 1.0 then
            sdb_ex = 1.0
          else
            for k,v in pairs(s._supported_categories) do
                if v == mx411 then table.remove(s._supported_categories,k)end
            end
            print("Failed excitation range 411 " .. l.Description .. "\texcitation=\t" .. l.Excitation)
            return false
          end

          if max > 250 then
            for k,v in pairs(s._supported_categories) do
                if v == mx411 then table.remove(s._supported_categories,k)end
            end
            return false
          elseif max > 100 then
            gain_low = true
            ex = 1.0
          elseif max > 20 then
            gain_low = true
            if sdb_ex < 2.5 then
              ex = sdb_ex
            else
              ex = 2.5
            end
          elseif max > 8 then
            ex = 1.0
          elseif max > 4 then
            if sdb_ex < 2.5 then 
              ex = sdb_ex
            else 
              ex = 2.5
            end
          else
            ex = sdb_ex
          end

          if s.bridge_resistance < 110 then
            if ex > 1.0 then
              ex = 1.0
            end
          elseif s.bridge_resistance < 300 then
            if ex > 2.5 then
              ex = 2.5
            end
          end

          if ex == 5.0 then
            ref.max = 4
            ref.min = -4
            er = "5.0v"
          elseif ex == 2.5 then
            if gain_low then
              ref.max = 100
              ref.min = -100
            else
              ref.max = 8
              ref.min = -8
            end
            er = "2.5v"
          elseif ex == 1.0 then
            if gain_low then
              ref.max = 250
              ref.min = -250
            else
              ref.max = 20
              ref.min = -20
            end
            er = "1.0v"
          end

          ref.mode = s.sensor_input.mode
          ref.type = s.sensor_input.type
          ref.regulation = s.sensor_input.regulation
          ref.wires = s.sensor_input.wires
          s._category_specific[ mx411].excitation_range = er
          s._category_specific[ mx411].sensor_input = ref
          return true                           
        end
      end
      ,excitation_mode = function( s ,args)
        local l = args.l
        local em
        if l.ExcitationFrequency == -1 then
          em = 'dc'
        else
          em = 'ac_4800_hz_sine'
        end
        s._category_specific[ mx411].excitation_mode = em
        return true
      end
    }
  }
  ,req = function ( s ,l )
    local supported = false
    for k,v in ipairs( supported_bridges ) do
      if v == l.SensorType then
        supported = true
      end
    end
    return supported
  end
  ,special_setup = function( info ,s ,l )
    s.sensor_input.type = info.type
    if l.Wiring > 0 then
      s.sensor_input.wires = tonumber( l.Wiring )
    else
      s.sensor_input.wires = info.set_wire
    end
    for i,wire in pairs( info.comp_wires ) do
      if s.sensor_input.wires == wire then
        return false
      end
    end
    print( "failed wire check "..l.Description)
    return true
  end

  ,setup = function( s , l ) --this is really messy, but the rules are kind of all over the place.
    local t = l.SensorType
    local failed = false
    if ss.StrainGage.req( s , l ) then
      s.sensor_input = 
      { mode="bridge" ,regulation=true ,type="full" ,wires=6}
      s.bridge_resistance = 350
      s.bridge_lead_wire_resistance = 0

      if t==365 or t==366 or t==367 or t==354 then
        s.sensor_input.type = 'half'
        s.sensor_input.wires = 5
      elseif t==368 or t==369 or t==371 or t==372 then
        s.sensor_input.type = 'quarter'
        if t==368 or t==369 then
          s.sensor_input.wires = 4
        else s.sensor_input.wires = 3 
            if l.ExcitationFrequency ~= -1 then -- and l.Regulation == true then
                l.ExcitationFrequency = -1    
            end
        end
      end

      if t==362 or t==365 or t==368 or t==371 then
        s.bridge_resistance = 120
      elseif t==364 or t==367 then
        s.bridge_resistance = 700
      end

      if t==353 or t==354 or t==350 or t==351 or t==352 then
        s.bridge_resistance = l.Resistance
        if s.bridge_resistance == 0 then s.bridge_resistance = 350; end

        local special_bridge_types = 
        {   { num=350,type="full",set_wire=6,comp_wires={6}} 
          ,{ num=351,type="half",set_wire=5,comp_wires={5}} 
          ,{ num=352,type="quarter",set_wire=4,comp_wires={3,4}}
        }

        for i,info in pairs( special_bridge_types ) do
          if info.num == t then
            failed = ss.StrainGage.special_setup( info ,s ,l)
          end
        end
      end

      if setup_category_specifics( 'StrainGage' ,s ,{l=l}) then
        failed = true
        print( 'Failed category specific check: ' ,l.Description)
      end

      if (s.sensor_input.type == 'half' or s.sensor_input.type == 'full')
      and (s.bridge_resistance < 90 or s.bridge_resistance > 5000) then
        failed = true
        print( "Failed Bridge Resistance " .. l.Description)
      elseif s.sensor_input.type == 'quarter' 
      and ( s.bridge_resistance ~= 120 and s.bridge_resistance ~= 350 ) then
        failed = true
        print( "Failed Bridge Resistance " .. l.Description )
      end
    else
      failed = true
      print( "Failed general check " .. l.Description)
    end
--    s.sensor_input = nil
    return failed
  end
}

--Bridge and StrainGage are identical, save for their enums
--I'm keeping them seperate just in case something comes up in the future,
--otherwise I'll eventually merge them and adjust the system accordingly
ss.Bridge = copy_table( ss.StrainGage ) 
ss.Bridge.name = "Bridge"
ss.Bridge.enum = 2

ss.Voltage = {
  name="Voltage"
  ,enum = 4
  ,categories = { mx1601 , mx840, mx411, mx1615 }
  ,elec_units = "V"
  ,range_min = -10000
  ,range_max = 10000
  ,s_types = {
    scale_types.ZeroSpan
    ,scale_types.TwoPoint
    ,scale_types.Table2
    ,scale_types.Gain
  }
  ,overlaps = {
    [mx1601] = {
      sensor_input = function( s ,args)
        local max = args.max
        local ref = {}
        if max <= 0.1 then
          max = 0.1
        elseif max <= 10 then
          max = 10
        elseif max <= 60 then
          max = 60
        else
          max = nil
        end
        if max then
          ref.mode = "volt"
          ref.max = max
          ref.min = max * -1
          s._category_specific[ mx1601].sensor_input = ref
          return true
        else
          table.remove(s._supported_categories, 1)
          return false
        end
      end
    }
    ,[mx840] = {
      sensor_input = function( s ,args)
        local max = args.max
        local ref = {}
        if max <= 0.1 then
          max = 0.1
        elseif max <= 10 then
          max = 10
        elseif max <= 60 then
          max = 60

        -- DFD - Not supported in GUI yet 25-Oct-15
        -- elseif max <= 300 then
        --  max = 300

        else
          max = nil
        end
        if max then
          ref.mode = "volt"
          ref.max = max
          ref.min = max * -1
          s._category_specific[ mx840].sensor_input = ref
          return true
        else
          table.remove(s._supported_categories, 2)
          return false
        end
      end
    }
    ,[mx1615] = {
      sensor_input = function( s ,args)
        local max = args.max
        local ref = {}
        if max >= 0.6 and max <= 60 then
          max = 15
        else
          max = nil
        end
        if max then
          ref.mode = "volt"
          ref.max = max
          ref.min = max * -1
          s._category_specific[ mx1615].sensor_input = ref
          return true
        else
          table.remove(s._supported_categories, 4)
          return false
        end
      end
    }
    ,[mx411] = {
      sensor_input = function( s ,args)
        local max = args.max
        local ref = {}
        if max >= 0.1 and max <= 10 then
          ref.mode = "volt"
          ref.max = 10
          ref.min = -10
          s._category_specific[ mx411].sensor_input = ref
          return true
        else
          table.remove(s._supported_categories, 3)
          return false
        end
      end
    }
  }
  ,req = function( s , l )
    if not tonumber(l.ActiveSupplyVoltage) then
        l.ActiveSupplyVoltage = 0
    else 
        l.ActiveSupplyVoltage = tonumber(l.ActiveSupplyVoltage)
    end
    local asv = tonumber(l.ActiveSupplyVoltage)
    if l.SensorType == 420 then
      if type(asv) == 'string' then
        return false
      elseif l.ActiveSupplyVoltage == 0 then
        return true
      elseif l.ActiveSupplyVoltage >= 5 then
        if l.ActiveSupplyVoltage <= 24 then
          s._supported_categories[4] = nil
          --cut out 1615, since invalid in this range
          return true
        end
        if l.ActiveSupplyVoltage > 24 and l.ActiveSupplyVoltage <= 30 then
          s._supported_categories[2] = nil
          s._supported_categories[3] = nil
          s._supported_categories[4] = nil
          --cut out 1615, 411, and 840,  since invalid in this range
          return true
        end
      end
    end
    return false
  end
  ,setup = function( s , l )
    local failed = false

    if ss.Voltage.req( s , l ) then
      s.transducer_power = l.ActiveSupplyVoltage
      s.calibration[3] = nil

      local max = 0
      if l.ScaleType == 1 then --ZeroSpan
        max = math.abs(s.calibration[1].electrical_span) + math.abs(s.calibration[1].electrical_zero)
      elseif l.ScaleType == 2 or l.ScaleType == 32 then --Table2/TwoPoint
        max = s.calibration[1].electrical[2]
      elseif l.ScaleType == 64 then --Gain
        max = l.Range
      end

      if setup_category_specifics( 'Voltage' ,s ,{max=max}) then
        failed = true
        print( 'Failed category specific: ' ,l.Description)
      end

    else
      print( "Failed general check " .. l.Description )
      failed = true
    end
    return failed
  end
}

ss.Current = {  
  name="Current"
  ,enum = 5
  ,categories = { mx1601, mx840, mx411 } 
  ,elec_units = "mA"
  ,range_min = -40
  ,range_max = 240
  ,s_types = { 
    scale_types.ZeroSpan
    ,scale_types.TwoPoint 
    ,scale_types.Table2 
    ,scale_types.Gain 
  }
  ,req = function( s , l )
    if not tonumber(l.ActiveSupplyVoltage) then
        l.ActiveSupplyVoltage = 0
    else 
        l.ActiveSupplyVoltage = tonumber(l.ActiveSupplyVoltage)
    end
    local asv = l.ActiveSupplyVoltage
    return l.SensorType == 421 
    and type(asv) == 'number'
    and ( (l.ActiveSupplyVoltage >= 5 
        and l.ActiveSupplyVoltage <= 30) 
      or l.ActiveSupplyVoltage == 0 )
        
  end
  ,setup = function( s , l )
    local failed = false

    if ss.Current.req( s , l ) then
      s.sensor_input.mode = "4-20ma"
      s.transducer_power = l.ActiveSupplyVoltage
      s.calibration[3] = nil
    else
      failed = true
      print( "Failed general check " .. l.Description )
    end
    return failed
  end
}

ss.Thermocouple = {
  name="Thermocouple"
  ,enum = 8
  ,categories = { mx1609} 
  ,elec_units = "°C"
  ,range_min = -100
  ,range_max = 100
  ,s_types = {
    scale_types.Internal
  }
  ,req = function( s , l )
    return l.SensorType == 451 and
    ( l.EngUnit:match( '°C') or l.EngUnit:match('°F') or l.EngUnit:match('°R'))
    and ( l.ScaleType == 0 or l.ScaleType == 256 )
  end
  ,setup = function( s , l )
    local failed = false

    if ss.Thermocouple.req( s , l ) then
      print( s.source_physical_units 
        ,s.electrical_units , s.output_units )
      s.sensor_input.mode = "thermocouple"
      s.sensor_input.type = 'k'
      s.source_physical_units = '°C'
    else
      failed = true
      print( "Thermo Failed general check " 
        .. l.Description , l.EngUnit , l.SensorType )
    end
    return failed
  end

}

ss.RTD = {  
  name="RTD"
  ,enum = 7
  ,categories = { mx840, mx1615 } 
  ,elec_units = "°C"
  ,range_min = -100
  ,range_max = 100
  ,s_types = {
    scale_types.Internal
  }
  ,req = function( s , l )
    if l.SensorType == 501 and l.SADPara == 26 and
    ( l.EngUnit:match( '°C') or l.EngUnit:match('°F') or l.EngUnit:match('°R'))
    and ( l.ScaleType == 0 or l.ScaleType == 256 ) then
      return true
    elseif l.SensorType == 503 and l.SADPara == 26 and
    ( l.EngUnit:match( '°C') or l.EngUnit:match('°F') or l.EngUnit:match('°R'))
    and ( l.ScaleType == 0 or l.ScaleType == 256 ) then
      table.remove(s._supported_categories, 2)
      --cut out 1615, since invalid for the pt1000
      return true
    else return false
    end
  end
  ,setup = function( s , l )
    local failed = false

    if ss.RTD.req( s , l ) and l.SensorType == 501 then
      s.sensor_input.mode = "pt100"
      s.source_physical_units = '°C'
--      s.calibration[3] = nil
    elseif ss.RTD.req( s , l ) and l.SensorType == 503 then
      s.sensor_input.mode = "pt1000"
      s.source_physical_units = '°C'
--      s.calibration[3] = nil
    else
      failed = true
      print( "RTD Failed general check " 
        .. l.Description , l.EngUnit , l.SensorType )
    end
    return failed
  end

}

ss.Resistance =     {   
  name="Resistance"
  ,enum = 10
  ,categories = { mx1615, mx840} 
  ,elec_units = "Ohm"
  ,range_min = 0
  ,range_max = 2500
  ,s_types = {
    scale_types.ZeroSpan
    ,scale_types.TwoPoint
    ,scale_types.Table2
    ,scale_types.Gain
  }
  ,req = function( s , l )
    return l.SensorType == 475 and l.SADPara == 26
  end
  ,setup = function( s , l )
    local failed = false

    if ss.Resistance.req( s , l ) then
      s.sensor_input.mode = "resistor"
      s.calibration[3] = nil
    else
      failed = true
      print( "Failed general check")
    end
    return failed
  end
}

ss.Potentiometric =     {   
  name="Potentiometric"
  ,enum = 14
  ,categories = { mx1615, mx840}
  ,elec_units = "mV/V"
  ,range_min = -400
  ,range_max = 400
  ,s_types = {
    scale_types.ZeroSpan
    ,scale_types.TwoPoint
    ,scale_types.Table2
    ,scale_types.Gain
  }
  ,overlaps = {
    [mx1615] = {
      excitation_range = function( s ,args)
        local l = args.l
        local er
        er = "1.0v"
        s._category_specific[ mx1615].excitation_range = er
        return true
      end
    }
    ,[mx840] = {
      excitation_range = function( s ,args)
        local l = args.l
        local er
        er = "2.5v"
        s._category_specific[ mx840].excitation_range = er
        return true
      end
    }
  }

  ,req = function( s , l )
    return l.SensorType == 385 and l.SADPara == 28
  end

  ,setup = function( s , l )
    local failed = false

    if ss.Potentiometric.req( s , l ) then
      s.sensor_input.mode = 'potentiometer'
      s.calibration[3] = nil
      s.excitation_mode = "dc"

    else
      failed = true
      print( "Failed general check" ,l.SensorType , l.SADPara)
    end
    if setup_category_specifics( "Potentiometric" ,s ,{l=l}) then
      failed = true
      print( 'Failed category specific: ' ,l.Description)
    end
    return failed
  end

}

ss.IEPE = { 
  name="IEPE"
  ,enum = 31
  ,categories = { mx1601, mx840, mx411 }
  ,elec_units = "V"
  ,range_min = -100
  ,range_max = 100
  ,s_types = {
    scale_types.ZeroSpan
    ,scale_types.TwoPoint
    ,scale_types.Table2
    ,scale_types.Gain
  }
  ,req = function( s , l )
    return l.SensorType == 581
  end
  ,setup = function( s , l )
    local failed = false

    if ss.IEPE.req( s , l ) then
      s.sensor_input.mode = "iepe"
      s.calibration[3] = nil
    else
      failed = true
    end
    return failed
  end
}

ss.IndBridge = {    
  name="IndBridge"
  ,enum = 11
  ,categories = { mx840, mx411 }
  ,elec_units = "mV/V"
  ,s_types = {
    scale_types.ZeroSpan
    ,scale_types.TwoPoint
    ,scale_types.Table2
    ,scale_types.Gain
  }
  ,overlaps = {
    [mx840] = {
      excitation_range = function( s ,args)
        local l = args.l
        local max
        local er        
        local ref = {}
        if l.ScaleType == 0 or l.ScaleType == 64 then --StrainGage or Gain
            max = l.Range
        elseif l.ScaleType == 1 then --ZeroSpan
            max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
        elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
            max = l.P2X
        else
            return false
        end
        if l.SensorType == 356 then
          if max > 1000 or l.Excitation < 1.0 then
            print("Failed excitation range " .. l.Description )
            return false
          elseif max > 300 then
            er = "1.0v"
            ref.max = 1000
            ref.min = -1000
          elseif max > 120 then
            er = "1.0v"
            ref.max = 300
            ref.min = -300
          else
            if l.Excitation >= 2.5 then
              er = "2.5v"
              ref.max = 120
              ref.min = -120
            else
              er = "1.0v"
              ref.max = 300
              ref.min = -300
            end 
          end 
          ref.mode = s.sensor_input.mode
          ref.type = s.sensor_input.type
          ref.wires = s.sensor_input.wires
          s._category_specific[ mx840].sensor_input = ref
          s._category_specific[ mx840].excitation_range = er
          return true       
        elseif l.SensorType == 357 then    
          if max > 300 or l.Excitation < 1.0 then
            print("Failed excitation range " .. l.Description )
            return false
          elseif max > 120 then
            er = "1.0v"
            ref.max = 300
            ref.min = -300
          else
            if l.Excitation >= 2.5 then
              er = "2.5v"
              ref.max = 120
              ref.min = -120
            else
              er = "1.0v"
              ref.max = 300
              ref.min = -300
            end 
          end 
          ref.mode = s.sensor_input.mode
          ref.type = s.sensor_input.type
          ref.wires = s.sensor_input.wires
          s._category_specific[ mx840].sensor_input = ref
          s._category_specific[ mx840].excitation_range = er
          return true       
        else
          print("Failed excitation range " .. l.Description )
          return false
        end
      end
    }
    ,[mx411] = {
      excitation_range = function( s ,args)
        local l = args.l
        local max
        local er
        local ref = {}
        if l.ScaleType == 0 or l.ScaleType == 64 then --StrainGage or Gain
            max = l.Range
        elseif l.ScaleType == 1 then --ZeroSpan
            max = math.abs(l.P2X-l.P1X) + math.abs(l.P1X)
        elseif l.ScaleType == 2 or l.ScaleType == 32 then -- Table2 or  TwoPoint
            max = l.P2X
        else
            return false
        end
        if max > 250 or l.Excitation < 1.0 then
          print("Failed excitation range " .. l.Description )
          return false
        elseif max > 100 then
          er = "1.0v"
          ref.max = 250
          ref.min = -250
        else
          if l.Excitation >= 2.5 then
            er = "2.5v"
            ref.max = 100
            ref.min = -100
          else
            er = "1.0v"
            ref.max = 250
            ref.min = -250
          end 
        end 
        ref.mode = s.sensor_input.mode
        ref.type = s.sensor_input.type
        ref.wires = s.sensor_input.wires
        s._category_specific[ mx411].sensor_input = ref
        s._category_specific[ mx411].excitation_range = er
        return true     
      end
    }
  }
  ,req = function( s , l )
    return l.SensorType == 356 or l.SensorType == 357
  end
  ,setup = function( s , l )
    local failed = false

    if ss.IndBridge.req( s , l ) then
      s.calibration[3] = nil
      s.excitation_mode = "ac_4800_hz_sine"
      if l.SensorType == 356 then
        s.sensor_input.mode = "inductive_bridge"
        s.sensor_input.type = "full"
        s.sensor_input.wires = "6"
      else
        s.sensor_input.mode = "inductive_bridge"
        s.sensor_input.type = "half"
        s.sensor_input.wires = "5"

      end
      if setup_category_specifics( 'IndBridge' ,s ,{l=l}) then
        failed = true
        print( 'Failed category specific check: ' ,l.Description)
      end
    else
      failed = true
    end
    return failed
  end
}

ss.LVDT = { 
  name="LVDT"
  ,enum = 3
  ,categories = { mx840}
  ,elec_units = "mV/V"
  ,range_min = -100
  ,range_max = 100
  ,s_types = {
    scale_types.ZeroSpan
    ,scale_types.TwoPoint
    ,scale_types.Table2
    ,scale_types.Gain
  }
  ,req = function( s , l )
    return l.SensorType == 380
  end
  ,setup = function( s , l )
    local failed = false

    if ss.LVDT.req( s , l ) then
      s.sensor_input.mode = "lvdt"
        s.excitation_range = "1.0v"
        s.excitation_mode = "ac_4800_hz_sine"
      s.calibration[3] = nil
    else
      failed = true
    end
    return failed
  end
}

ss.Frequency = {    
  name="Frequency"
  ,enum = 6
  ,categories = { mx840 }
  ,elec_units = "kHz"
  ,range_min = -10000
  ,range_max = 10000
  ,s_types = {
    scale_types.ZeroSpan
    ,scale_types.TwoPoint
    ,scale_types.Table2
    ,scale_types.Gain
  }
  ,req = function( s , l )
    local voltage
    if not tonumber(l.ActiveSupplyVoltage) then 
        voltage = 0 
        l.ActiveSupplyVoltage = 0
    else
        voltage = tonumber(l.ActiveSupplyVoltage)
    end
    if type(voltage) == "number" then
      if voltage == 0 or (voltage >= 5 and voltage <= 24) then
        l.ActiveSupplyVoltage = voltage
        return l.SensorType == 520 or l.SensorType == 522 or l.SensorType == 523
      else
        print("Active Supply Voltage failure:\t" .. l.ActiveSupplyVoltage .. "\t" ..  tostring(voltage))
        return false
      end
    else 
        print("Active Supply Voltage type failure:\t" .. tostring(l.ActiveSupplyVoltage) .. "\t" .. tostring(voltage))
        return false
    end
  end
  ,setup = function( s , l )
    local failed = false
    if ss.Frequency.req( s , l ) then
      s.calibration[3] = nil
        s.transducer_power = l.ActiveSupplyVoltage
      if l.SensorType == 520 then
        s.sensor_input.mode = "pulse_frequency"
      elseif l.SensorType == 522 then
        s.sensor_input.mode = "encoder_frequency"
        s.sensor_input.quadrature = "false"
      else
        s.sensor_input.mode = "encoder_frequency"
        s.sensor_input.quadrature = "true"
      end    
    else
      failed = true
    end
    return failed
  end
}

ss.Incremental = {  
  name="Incremental"
  ,enum = 12
  ,categories = { mx840 }
  ,elec_units = "kImp"
  ,range_min = -360
  ,range_max = 360
  ,s_types = {
    scale_types.ZeroSpan
    ,scale_types.TwoPoint
    ,scale_types.Table2
    ,scale_types.Gain
  }
  ,req = function( s , l )
    if not tonumber(l.ActiveSupplyVoltage) then 
       l.ActiveSupplyVoltage = 0 
    else
        l.ActiveSupplyVoltage = tonumber(l.ActiveSupplyVoltage)
    end
    if l.ActiveSupplyVoltage == 0 or (l.ActiveSupplyVoltage >= 5 and l.ActiveSupplyVoltage <= 24) then
        print(l.SensorType == 525 or l.SensorType == 522 or l.SensorType == 523)
        return l.SensorType == 525 or l.SensorType == 522 or l.SensorType == 523
    else
        print("Sensor type not supported")
      return false
    end
  end
  ,setup = function( s , l )
    local failed = false
    print("Counter")
    if ss.Incremental.req( s , l ) then
        s.transducer_power = l.ActiveSupplyVoltage
      s.calibration[3] = nil
      if l.SensorType == 525 then
        s.sensor_input.mode = "pulse_counter"
      elseif l.SensorType == 522 then
        s.sensor_input.mode = "encoder_counter"
        s.sensor_input.quadrature = "false"
        if l.CounterIndexEnabled == "null" then
            l.CounterIndexEnabled = 0
        end
        if not tonumber(l.CounterIndexDivisor) then
            l.CounterIndexDivisor = 1
        end
        if l.CounterIndexEnabled and (l.CounterIndexDivisor >= 1 or l.CounterIndexDivisor <= 4) then
          if l.CounterIndexEnabled == 1 then
            s.sensor_index_reset = true
            s.encoder_index_divisor = l.CounterIndexDivisor
          elseif l.CounterIndexEnabled == 0 then
            s.sensor_index_reset = false
          else 
            print("sensor index failed")
            failed = true
          end
        else
            print("counter index failed")
          failed = true
        end
      else
        s.sensor_input.mode = "encoder_counter"
        s.sensor_input.quadrature = "true"
        if l.CounterIndexEnabled and (l.CounterIndexDivisor >= 1 or l.CounterIndexDivisor <= 4) then
          if l.CounterIndexEnabled == 1 then
            s.sensor_index_reset = true
            s.encoder_index_divisor = l.CounterIndexDivisor
          elseif l.CounterIndexEnabled == 0 then
            s.sensor_index_reset = false
          else 
            print("sensor index failed 2")
            failed = true
          end
        else
            print("counter index failed 2")
          failed = true
        end
      end    
    else
        print("req failed ")
      failed = true
    end
    return failed
  end
}

return ss


