-- spi_flash package of SPI FLASH memory access functions
--
-- API Functions:
--   verify_part()
--   read_block(offset, byte_count)
--   erase_all()
--   erase_region(offset, byte_count)
--   write_block(offset, data_bytes<as Lua string>)
--   
-- Internal Device Functions:
--   write_enable()
--   write_disable()
--   read_identification()
--   read_status_register()
--   write_status_register(status)
--   read_data_bytes(offset, count)
--   page_program(address, data)
--   sector_erase(address)
--   bulk_erase()
--   deep_power_down()
--   release_from_deep_power_down()
--
-- Note: fast reads are not supported, the FPGA logic doesn't run that fast.
--


local u = require 'unix'
local bit = require 'bit'
local connctrl = require 'connctrl'
local ffi = require 'ffi'

local spi_flash = {}

local page_size = 256
local sector_size = 65536
local device_size = 16777216

local function write_enable()
    connctrl.spi_transaction(0x06)
    connctrl.end_spi_command()
end

local function write_disable()
    connctrl.spi_transaction(0x04)
    connctrl.end_spi_command()
end

local function read_identification()
    connctrl.spi_transaction(0x9F)
    local result = {}
    result.manf_id = connctrl.spi_transaction(0)
    result.mem_type = connctrl.spi_transaction(0)
    result.mem_cap = connctrl.spi_transaction(0)
    local uid_length = connctrl.spi_transaction(0)
    if uid_length > 0 then
        result.uid = {}
        for i = 1,uid_length do
            result.uid[i] = connctrl.spi_transaction(0)
        end
    end
    connctrl.end_spi_command()
    return result
end

local function read_status_register()
    connctrl.spi_transaction(0x05)
    local status = connctrl.spi_transaction(0)
    connctrl.end_spi_command()
    return status
end

local function write_status_register(status)
    connctrl.spi_transaction(0x01)
    connctrl.spi_transaction(status)
    connctrl.end_spi_command()
end

local function send_address(address)
    local addr1 = bit.band(address, 0xFF)    
    local addr2 = bit.band(bit.rshift(address, 8), 0xFF)
    local addr3 = bit.band(bit.rshift(address, 16), 0xFF)
    connctrl.spi_transaction(addr3)
    connctrl.spi_transaction(addr2)
    connctrl.spi_transaction(addr1)
end

local function read_data_bytes(address, count)
    local buffer = ffi.new("uint8_t[?]", count)
    connctrl.spi_transaction(0x03)
    send_address(address)
    for i = 1,count do
        buffer[i - 1] = connctrl.spi_transaction(0)
    end
    connctrl.end_spi_command()
    return ffi.string(buffer, count)
end

local function page_program(address, data)
    connctrl.spi_transaction(0x02)
    send_address(address)
    for i=1,#data do connctrl.spi_transaction(data:byte(i, i)) end
    connctrl.end_spi_command()
end

local function sector_erase(address)
    connctrl.spi_transaction(0xD8)
    send_address(address)
    connctrl.end_spi_command()
end

local function bulk_erase()
    connctrl.spi_transaction(0xC7)
    connctrl.end_spi_command()
end

local function deep_power_down()
    connctrl.spi_transaction(0xB9)
    connctrl.end_spi_command()
end

local function release_from_deep_power_down()
    connctrl.spi_transaction(0xAB)
    connctrl.end_spi_command()
end

local function wait_for_command_completion(max_wait_time)
    connctrl.spi_transaction(0x05)
    local s, ns = u.clock_gettime()
    local start_time = s + ns / 1000000000
    local function time_spent()
        local s, ns = u.clock_gettime()
        local now = s + ns / 1000000000
        return now - start_time
    end
    local _time_spent
    repeat
        _time_spent = time_spent()
        local sr = connctrl.spi_transaction(0)
        if bit.band(sr, 1) == 0 then
            connctrl.end_spi_command()
            return _time_spent
        end
    until _time_spent > max_wait_time
    connctrl.end_spi_command()
    print("SR wait timeout!")
    return nil
end


------------------- API Functions ---------------------------

function spi_flash.verify_part()
    local id_bytes = read_identification()
    if id_bytes == nil then return false end
    if id_bytes.manf_id ~= 0xEF then return false end
    if id_bytes.mem_type ~= 0x40 then return false end
    if id_bytes.mem_cap ~= 0x18 then return false end
    return true
end

function spi_flash.read_block(address, byte_count)
    return read_data_bytes(address, byte_count)
end

function spi_flash.erase_all()
    write_enable()
    bulk_erase()
    return wait_for_command_completion(250)
end

function spi_flash.erase_region(address, byte_count)
    local sector_mask = bit.bnot(sector_size - 1)
    local start_sector = bit.band(address, sector_mask)
    local end_sector = bit.band(address + byte_count - 1, sector_mask)
    local sector_count = ((end_sector - start_sector) / sector_size) + 1
    for i=1,sector_count do
        write_enable()
        sector_erase(start_sector)
        if not wait_for_command_completion(3) then return false end
        start_sector = start_sector + sector_size
    end
    return true
end

function spi_flash.write_block(address, data_bytes)
    local start_address = bit.band(address, bit.bnot(page_size-1))
    local end_address = bit.band(address + #data_bytes, bit.bnot(page_size-1))
    local page_count = (end_address - start_address) / page_size
    local byte_count = #data_bytes
    local data_index = 1
    local page_data, max_page_bytes, page_bytes, readback
    while byte_count > 0 do
        max_page_bytes = page_size - bit.band(address, page_size - 1)
        page_bytes = byte_count
        if page_bytes > max_page_bytes then page_bytes = max_page_bytes end
        write_enable()
        page_data = data_bytes:sub(data_index, data_index + page_bytes - 1)
        page_program(address, page_data)
        if not wait_for_command_completion(0.1) then return false end
        readback = read_data_bytes(address, page_bytes)
        if readback ~= page_data then
            readback = read_data_bytes(address, page_bytes)
            if readback ~= page_data then
                print("Repeated readback failure at base address", address)
                return false
            end
        end
        address = address + page_bytes
        data_index = data_index + page_bytes
        byte_count = byte_count - page_bytes
    end
    return true
end

return spi_flash

