require 'strict'

-- directional_velocity.lua - Implements the Directional Velocity computed channel.

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local rational = require 'rational'
local log = require 'log'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_min = math.min
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Utils_frame_data_type = utils.frame_data_type
local Utils_put_frame = utils.put_frame

local SIGN_PLUS = 1
local SIGN_MINUS = 2

local mt = {}
    
-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

function mt:flush_output_frame()
    self.of.count = self.bufsize
    self.of.hdr.ts = self.tot_count
    self.of.ts = math.floor(self.tot_count)
    if not Utils_put_frame(self.of, self.pipe_list) then
        Frame_delete(self.of)
        self.of = nil
    end
    self.of = Frame_create(self.bufsize, self.out_frame_sample_size)
    self.out_vals = ffi_cast(Utils_frame_data_type(
        self.out_frame_sample_size), self.of.buf)
    self.tot_count = self.tot_count + self.bufsize
end

-- Processes incoming frames, produces outgoing frames.
function mt:process_in_frame_directional(f1,f2)
    local in_vals1 = ffi_cast(Utils_frame_data_type(f1.element_size), f1.buf)
    local in_vals2 = ffi_cast(Utils_frame_data_type(f2.element_size), f2.buf)
    
    for i = 0, f1.count-1 do
        self.out_vals[i] = in_vals1[i]*in_vals2[i];
    end
    self:flush_output_frame()
end

-- Processes incoming frames, produces outgoing frames.
function mt:process_in_frame(f1,f2)
    local in_vals1 = ffi_cast(Utils_frame_data_type(f1.element_size), f1.buf)
    local in_vals2 = ffi_cast(Utils_frame_data_type(f2.element_size), f2.buf)
    
    for i = 0, f1.count-1 do
        if i == 0 then
            if in_vals2[i] > self.last then
                self.state = SIGN_PLUS
            elseif in_vals2[i] < self.last then
                self.state = SIGN_MINUS
            end
        else
            if in_vals2[i] > in_vals2[i-1] then
                self.state = SIGN_PLUS
            elseif in_vals2[i] < in_vals2[i-1] then
                self.state = SIGN_MINUS
            end
        end

        if self.state == SIGN_PLUS then
           self.out_vals[i] = in_vals1[i];
        else 
            self.out_vals[i] = -in_vals1[i];
        end
    end

    -- Set sign on 1st output sample to the sign of the 2nd output sample.
    if self.first_frame then
        self.first_frame = false       
        if (f1.count > 1) then
            if self.out_vals[1] > 0 then
                if self.out_vals[0] < 0 then
                    self.out_vals[0] = self.out_vals[0] * -1.0
                end
            elseif self.out_vals[1] < 0 then
                if self.out_vals[0] > 0 then
                    self.out_vals[0] = self.out_vals[0] * -1.0
                end
            end
        end
    end
    self.last = in_vals2[f1.count-1]
    self:flush_output_frame()
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    while Pipe_has_frames(self.chan_info[1].pipe) and
            Pipe_has_frames(self.chan_info[2].pipe) do
        local f1 = Pipe_get_frame(self.chan_info[1].pipe)
        local f2 = Pipe_get_frame(self.chan_info[2].pipe)
        if self.directional_channel then
            self:process_in_frame_directional(f1,f2)
        else
            self:process_in_frame(f1,f2)
        end
        Frame_delete(f1)
        Frame_delete(f2)
        f1,f2 = nil,nil
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer for
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of pipe lists created by this instance; this will
--       be 1 for this module
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    local dir_v = setmetatable({}, {__index = mt})
    dir_v.sie_collect_idx = first_sie_index
    utils.set_common_meta(dir_v, stanza)
        -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    assert(meta_list, "Missing parameter 'meta_list'")
        -- pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    dir_v.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(dir_v.pipe_rate and dir_v.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(dir_v.pipe_rate))
        -- sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    dir_v.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(dir_v.sample_rate and dir_v.sample_rate > 0,
        "Invalid sample rate: " .. tostring(dir_v.sample_rate))
        -- Always two input channels for this module
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 2, "Directional Velocity computed channel " ..
        "requires exactly two input channels")
    dir_v.chan_info = {}
    for i = 1, #input_channels do
        local input_node = node_list[input_channels[i]]
        local input_stanza = meta_list[input_channels[i]]
        assert(input_node, "Node for Input Channel not found")
        assert(input_stanza, "Stanza for Input Channel not found")
        assert(input_node.is_logical ~= true, "Input channel must not be "..
            " a logical channel")
        dir_v.chan_info[i] =
            {name = input_channels[i],
             pipe = Pipe_attach_pipe(input_node.pipe_list)}
        assert(dir_v.chan_info[i].pipe and dir_v.chan_info[i].pipe ~= nil,
            "failed to create input pipe")
        assert(rational:new(input_stanza.sample_rate):tonumber() ==
            dir_v.sample_rate, 'Input channel "'..input_channels[i]..
            '" sample rate does not match this module\'s sample rate')
    end

    dir_v.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(dir_v.out_frame_sample_size and dir_v.out_frame_sample_size > 0,
        "Invalid frame sample size:  "..tostring(dir_v.out_frame_sample_size))
        -- create output pipe structures
    dir_v.pipe_list = Pipe_new_pipe_list()
    assert(dir_v.pipe_list and dir_v.pipe_list ~= nil,
        "failed to create (output) pipe list")
    --determine whether second channel is a directional channel
    dir_v.sign_channel = stanza.sign_channel

        --local variables needed not set from stanza.
    dir_v.tot_count = 0
    dir_v.bufsize = frame.bufsize(dir_v.sample_rate, dir_v.pipe_rate)
        -- output frame
    dir_v.of = Frame_create(dir_v.bufsize, dir_v.out_frame_sample_size)
        -- "pointer" to output frame data
    dir_v.out_vals = ffi_cast(Utils_frame_data_type(
        dir_v.out_frame_sample_size), dir_v.of.buf)
    dir_v.out_index = 0
    dir_v.last = 0
    dir_v.state = nil
    dir_v.first_frame = true

    return dir_v, 1  -- this module has one pipe list
end

return m