require 'strict'

-- expression.lua - Implements the Signal Calculator computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local lpeg = require 'lpeg'
local math = require 'math'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}

function mt:dump()
    dbg('tot_count=%lld, expression="%s"',
        self.tot_count, tostring(self.expression))
end

-- Performs any needed post instantiation initialization.
-- Parameter:
--     SIE writer (NOT USED for this module!)
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Loads the given chunk string as a lua function.
-- Parameters:
--   'chunk' is the string containing the lua code to be loaded
--   'name' is the function name as accepted by loadstring()
-- Returns:  the loaded and defined function, followed by
--   any error string (which will be nil for a successful load)
function mt:load_expr(chunk, name)
    -- Create a table to use as a metatable for translating expression
    -- functions to Lua "math." or locally defined functions.
    local xlat_tab = {}
    xlat_tab.fabs = math.abs
    xlat_tab.acos = math.acos
    xlat_tab.asin = math.asin
    xlat_tab.atan = math.atan
    xlat_tab.atan2 = math.atan2
    xlat_tab.ceil = math.ceil
    xlat_tab.cos = math.cos
    xlat_tab.cosh = math.cosh 
    xlat_tab.exp = math.exp
    xlat_tab.floor = math.floor 
    xlat_tab.log = math.log
    xlat_tab.log10 = math.log10
    xlat_tab.min = math.min
    xlat_tab.max = math.max
    xlat_tab.sin = math.sin
    xlat_tab.sinh = math.sinh
    xlat_tab.tan = math.tan
    xlat_tab.tanh = math.tanh
    xlat_tab.sqrt = math.sqrt
    xlat_tab.pi = function()
            return math.pi
        end
    xlat_tab.float = function(inval)
            if type(inval) == 'boolean' then
                return (inval) and 1 or 0
            else
                return ((inval) ~= 0) and 1 or 0
            end
        end
    xlat_tab.sgn = function(inval)
            if ((inval) == 0) then
                return 0.0
            elseif ((inval) > 0) then
                return 1.0
            elseif((inval) < 0) then
                return -1.0
            end
        end

    -- Create expression table
    self.e_tab = setmetatable(xlat_tab, {__mod = math.fmod, __index = _G})
    self.e_tab.cnt = 0
    local expr_fn, errmsg = loadstring(chunk, name)
    if expr_fn ~= nil then
        setfenv(expr_fn, self.e_tab)
        expr_fn() -- define it, per lua
        errmsg = nil
    end
    return expr_fn, errmsg
end

local special = { "%%", "%^", "%$", "%(", "%)", "%.",
                  "%[", "%]", "%*", "%+", "%-", "%?"}
local function escape_magic(str)
    for i=1,#special do
        str = (str:gsub(special[i], (special[i]:gsub("%%","%%%%"))))
    end
    return str
end

-- Parses the module's "expression" string.
-- Parameters:
--   'input_channels' is a list of names of all input channels
--   'expression' is the expression string from the stanza
--   'node_list' is a list of all nodes created to this point
-- Returns:  lua chunk loaded from the parsed expression string
function mt:parse(input_channels, expression, node_list)
    -- Attach input pipe(s), keeping the association between
    -- input channel name(s) and input pipe(s).
    self.in_pipe_info = {}
    local input_node
    for i = 1, #input_channels do
        input_node = node_list[input_channels[i]]
        assert(input_node, "Node for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        self.in_pipe_info[i] =
            {name = input_channels[i],
             is_logical = utils.xlate_meta_boolean(input_node.is_logical),
             pipe = Pipe_attach_pipe(input_node.pipe_list)}
    end

    -- Prepare 'expression' for loading:
    --   replace channel names in 'expression' with
    --   "dereferenced" frame buffer "pointers"
    --   NOTE:  channel names are always expected
    --   to be contained within single quotes!
    for pindex = 1, #self.in_pipe_info do
        if self.in_pipe_info[pindex].is_logical then
            expression = expression:gsub(escape_magic(
                "'" .. self.in_pipe_info[pindex].name .. "'"),
                ' (v[' .. pindex .. '][i] ~= 0) ')
        else
            expression = expression:gsub(escape_magic(
                "'" .. self.in_pipe_info[pindex].name .. "'"),
                ' v[' .. pindex .. '][i] ')
        end
    end
    --   make simple "language" substitutions
    expression = expression:gsub('&&', ' and ')
    expression = expression:gsub('||', ' or ')
    expression = expression:gsub('!=', ' ~= ')
    expression = expression:gsub('!', ' not ')
    --   construct string containing the loop code
    local chunk_str = [[
            if not cnt or cnt == 0 then return nil end
            for i = 0, cnt - 1 do
        ]]
    chunk_str = chunk_str ..
              ' outs[i] = ' .. expression .. '\n'
    chunk_str = chunk_str .. [[
            end
        ]]
    
    -- load the expression-evaluating function
    local expr_fn, errmsg = self:load_expr(chunk_str, "expr_fn")
    assert(errmsg ==  nil,
        string.format("expression loading error: \"%s\"", errmsg))

    -- return loaded expression-evaluating function
    -- and the parsed expression string
    return expr_fn, expression
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.in_pipe_info[<pipe index>].frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mt:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = #self.in_pipe_info > 0
    for i = 1, #self.in_pipe_info do
        got_frames = Pipe_has_frames(self.in_pipe_info[i].pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    if got_frames then
        for i = 1, #self.in_pipe_info do
            self.in_pipe_info[i].frame =
                Pipe_get_frame(self.in_pipe_info[i].pipe)
        end
    end
    return got_frames
end

-- Processes incoming frames, producing outgoing frames.
function mt:run()
    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- can't process any input frames yet
        return
    end

    -- Process each frame in input pipe(s).
    local of        -- will reference created output frame
    local out_vals  -- will reference "pointer" to output frame's data
    local _         -- we won't use 'in_vals' "pointer" returned by
                    --     create_output_frame()

    local types = {[4] = 'float *', [8] = 'double *'}

    while have_all_frames do
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        So it doesn't matter which input frame we use for
        --        obtaining frame count & size.
        -- create frame for output
        local reference_frame = self.in_pipe_info[1].frame

        -- Remember that self.e_tab serves as self.e_fun()'s global
        -- environment.
        of, _, self.e_tab.outs =
            utils.create_output_frame(reference_frame.maxsize,
                reference_frame, self.tot_count,
                self.out_frame_sample_size)
        -- Assign frame pointers to variables in e_fun()'s evironment
        -- table to allow e_fun() access to the input frame bufs.
        -- NOTE:  self.e_fun() itself contains a loop to cover all
        --        the input frames' data buf values.
        self.e_tab.cnt = reference_frame.count
        self.e_tab.v = {}
        for i, info in ipairs(self.in_pipe_info) do
            self.e_tab.v[i] = ffi_cast(types[info.frame.element_size],
                info.frame.buf)
        end
        -- fill output frame
        self.e_fun()
        of.count = reference_frame.count
        self.tot_count = self.tot_count + of.count

        -- put frame in output pipe(s)
        if not utils.put_frame(of, self.pipe_list) then
            Frame_delete(of)
        end
        -- don't need the processed input frames any longer
        for i = 1, #self.in_pipe_info do
            Frame_delete(self.in_pipe_info[i].frame)
            self.in_pipe_info[i].frame = nil
        end

        -- get a new set of input frames to process
        have_all_frames = self:get_frames()
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer for
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of pipe lists created by this instance; this will
--       be 1 for this module
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local exp_obj = setmetatable({}, {__index = mt})
    utils.set_common_meta(exp_obj, stanza)
    exp_obj.sie_collect_idx = first_sie_index
    -- validate needed metadata, and create input/output pipe
    -- structures
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1, "Signal Calculator computed channel" ..
        "must have at least one input channel")
    assert(stanza.sample_rate, "Missing sample_rate")
    exp_obj.sample_rate = stanza.sample_rate
    assert(stanza.data_type, "Missing data type")
    exp_obj.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(exp_obj.out_frame_sample_size and
        exp_obj.out_frame_sample_size > 0,
        string.format("Invalid frame sample size:  %s",
        tostring(exp_obj.out_frame_sample_size)))
    self.expression = stanza.expression
    assert(#self.expression >= 1, "Missing expression")
    exp_obj.pipe_list = Pipe_new_pipe_list()
    -- create loaded expression function
    exp_obj.e_fun, exp_obj.expression =
        exp_obj:parse(input_channels, self.expression,
            node_list)

    -- used for frames' time stamps (just a sample count for
    -- this module)
    exp_obj.tot_count = 0

    return exp_obj, 1  -- this module has one pipe list
end

return m
