require 'strict'

-- pulse_frequency.lua - Implements the Pulse Frequency computed channel.

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local rational = require 'rational'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_min = math.min
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Utils_frame_data_type = utils.frame_data_type
local Utils_put_frame = utils.put_frame

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

function mt:flush_output_frame()
    self.of.count = self.bufsize
    self.of.hdr.ts = self.tot_count
    self.of.ts = math.floor(self.tot_count)
    if not Utils_put_frame(self.of, self.pipe_list) then
        Frame_delete(self.of)
        self.of = nil
    end
    self.of = Frame_create(self.bufsize, self.out_frame_sample_size)
    self.out_vals = ffi_cast(Utils_frame_data_type(
        self.out_frame_sample_size), self.of.buf)
    self.tot_count = self.tot_count + self.bufsize
    self.out_index = 0
end

function mt:check_for_edge()
    local common = self.scale * self.factor * self.out_rate
    local state, ref = self.state, self.refstate
    --found an edge
    if (ref == 0 and state == 1) or (ref == 1 and state == 0) then
        if state == 0 and not self.first_rise_edge then --1st rise
            self.first_rise_edge = true
        elseif state == 1 and not self.first_fall_edge then --1st fall
            self.first_fall_edge = true
        elseif state == 0 then --falling edge
            if self.fall_cnt > self.reset_cnt then
                self.output = self.minfreq
            else
                self.output = common / self.fall_cnt
            end
            self.fall_ref = self.fall_cnt
        elseif state == 1 then --rising edge
            if self.rise_cnt >= self.reset_cnt then
                self.output = self.minfreq
            else
                self.output = common / self.rise_cnt
            end
            self.rise_ref = self.rise_cnt
        end

        if state == 1 then
            self.rise_cnt = 0
        else
            self.fall_cnt = 0
        end
    --check if pulse width longer than current output
    elseif  (self.first_rise_edge and self.rise_cnt > self.rise_ref) or
            (self.first_fall_edge and self.fall_cnt > self.fall_ref) then
        self.output = math.min(common / self.rise_cnt, common / self.fall_cnt)
    end
end

function mt:check_for_reset_count()
    --Check for reset_count
    if self.rise_cnt >= self.reset_cnt or self.fall_cnt >= self.reset_cnt then
        self.rise_ref = self.reset_cnt
        self.rise_cnt = self.reset_cnt
        self.fall_ref = self.reset_cnt
        self.fall_cnt = self.reset_cnt
        self.output = self.minfreq
    end
end

function mt:check_window_end()
    self.refstate = self.state
    self.wait = self.wait - 1
    
    if self.wait==0 then
        self.wait = self.factor
        --Pulse frequency cc needs to lag by one in order to
        --match the way CX hardware's pulse frequency works.
        if self.lag_output then
            --need to double push the first lag_output
            if not self.did_push_first_sample then
                self.out_vals[self.out_index] = self.lag_output
                self.out_index  = self.out_index + 1
                if self.out_index == self.bufsize then
                    self:flush_output_frame()
                end
                self.did_push_first_sample = true
            end
            self.out_vals[self.out_index] = self.lag_output
            self.out_index  = self.out_index + 1
        end
        self.lag_output = self.output > self.minfreq and self.output or self.minfreq
        if self.out_index == self.bufsize then
            self:flush_output_frame()
        end
    end
end

function mt:process_in_frame(f)
    local in_vals = ffi_cast(Utils_frame_data_type(f.element_size), f.buf)
    local inp_index = 0 -- Index to input frame data
    while inp_index < f.count do
        self.rise_cnt  = self.rise_cnt + 1
        self.fall_cnt = self.fall_cnt + 1
        self.state = in_vals[inp_index]
        inp_index = inp_index + 1
        
        self:check_for_edge()
        self:check_for_reset_count()
        self:check_window_end()
    end
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    while Pipe_has_frames(self.in_pipe) do
        local f = Pipe_get_frame(self.in_pipe) -- input frame
        self:process_in_frame(f)
        Frame_delete(f)
        f = nil
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local pulse_f = setmetatable({}, {__index = mt})
    pulse_f.sie_collect_idx = first_sie_index
    utils.set_common_meta(pulse_f, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Pulse Frequency computed channel " ..
        "only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    assert(input_node.is_logical, "Input channel must be a logical channel")
        --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    pulse_f.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(pulse_f.pipe_rate and pulse_f.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(pulse_f.pipe_rate))
        --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    pulse_f.out_rate = rational:new(stanza.sample_rate):tonumber()
    assert(pulse_f.out_rate and pulse_f.out_rate > 0,
        "Invalid sample rate: " .. tostring(pulse_f.out_rate))
    pulse_f.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(pulse_f.out_frame_sample_size and pulse_f.out_frame_sample_size > 0,
        "Invalid frame sample size:  "..tostring(pulse_f.out_frame_sample_size))

    --Module specific parameters
    pulse_f.factor = tonumber(stanza.output_rate_factor)
    assert(pulse_f.factor and pulse_f.factor%1 == 0 and pulse_f.factor >= 10,
        "Invalid output_rate_factor: "..tostring(stanza.output_rate_factor)..
        ", must be an integer 10 or greater.")
    pulse_f.minfreq = tonumber(stanza.minimum_output_frequency)
    assert(pulse_f.minfreq and pulse_f.minfreq >=0.001 and pulse_f.minfreq <= 1,
        "Invalid minimum_output_frequency: "..
        tostring(stanza.minimum_output_frequency)..
        ", must be value between 0.001 and 1.0 Hz.")
    pulse_f.scale = tonumber(stanza.output_scale_factor)
    assert(pulse_f.scale and pulse_f.scale ~= 0,
        "Invalid output_scale_factor: "..tostring(stanza.output_scale_factor)..
        ", must be a non-zero number.")

    -- create input/output pipe structures
    pulse_f.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(pulse_f.in_pipe and pulse_f.in_pipe ~= nil,
        "failed to create input pipe")
    pulse_f.pipe_list = Pipe_new_pipe_list()
    assert(pulse_f.pipe_list and pulse_f.pipe_list ~= nil,
        "failed to create (output) pipe list")

    --local variables needed not set from stanza.

    pulse_f.tot_count = 0
    pulse_f.in_rate   = pulse_f.out_rate * pulse_f.factor
    pulse_f.bufsize   = frame.bufsize(pulse_f.out_rate, pulse_f.pipe_rate)

    pulse_f.wait      = pulse_f.factor
    pulse_f.state     = nil
    pulse_f.refstate  = nil
    pulse_f.output    = pulse_f.minfreq
    pulse_f.reset_cnt = (pulse_f.in_rate * pulse_f.factor)/pulse_f.minfreq
    pulse_f.rise_cnt  = 0
    pulse_f.fall_cnt  = 0
    pulse_f.rise_ref  = math.huge
    pulse_f.fall_ref  = math.huge
    pulse_f.first_rise_edge = false
    pulse_f.first_fall_edge = false
    pulse_f.of = Frame_create(pulse_f.bufsize, pulse_f.out_frame_sample_size)
    pulse_f.out_vals = ffi_cast(
        Utils_frame_data_type(pulse_f.out_frame_sample_size), pulse_f.of.buf)
    pulse_f.out_index = 0

    return pulse_f, 1
end

return m
