require 'strict'

-- burst_history.lua - Implements the Burst History data mode.
--
-- Implementation note:
--     This data mode uses its own particular trigger mechanism,
--     not the "generic" datamode triggering mechanism.

local cjson = require 'cjson'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local C = ffi.C
local ffi_cast = ffi.cast
local ffi_new = ffi.new
local ffi_sizeof = ffi.sizeof
local math_max = math.max
local math_min = math.min
local math_floor = math.floor

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list

-- "constant" values for the possible runtime states
local ST_LOOK_FOR_TRIG           = 1
local ST_FIND_PRE_HISTORY_START  = 2
local ST_WRITE_PRE_HISTORY_START = 3
local ST_FINISH_BURST            = 4

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:
--     'sie_writer'  SIE writer
-- Returns:  true
function mt:setup(sie_writer)
    self.sw = sie_writer
    self:sie_xml()
    return true
end

-- Implements "gate" trigger, returning true whenever the input value
-- is non-zero.
-- Parameter:
--     'trig_data'  sample from trigger channel
-- Returns:  true if 'trig_data' is non-zero, false otherwise
function mt:gate_trigger(trig_data)
    return trig_data ~= 0
end

-- Implements "edge" trigger, returning true whenever the input value
-- is non-zero and the previous trigger channel value was zero.
-- Parameter:
--     'trig_data'  sample from trigger channel
-- Returns:  true if 'trig_data' is transitions from zero to non-zero,
--     false otherwise
function mt:edge_trigger(trig_data)
    if tonumber(trig_data) == 0 then
        self.prev_trig_on = false
        return false
    else  -- trig_data ~= 0  
        if self.prev_trig_on then
            return false
        else
            self.prev_trig_on = true
            return true
        end
    end
end

-- If possible, puts the start of each input channel's pre-trigger
-- history into output frames, copying the history from saved input
-- frames.  At most one set of input frames will be processed.
-- NOTE:  There will not necessarily be pre_count samples of history
--        output, since the trigger may have occurred too soon after
--        the previous burst to allow that.
-- Returns:  number of samples output, followed by index of next
--     input sample to process
function mt:write_start_pretrig()
    -- Note that the calculated 'start_index' may or may not be
    -- small enough to result in an index into the current set of
    -- input frames.
    local start_index = self.curr_trig.b_start - self.tot_count
    local have = true  -- will have at least one set of input frames
    local in_count = self.chan_info[1].in_frame.count

    if start_index >= in_count then
        -- pre-trigger history starts in a later set of input frames
        return 0, start_index
    end

    -- If "end of history (i.e. end of burst)" is in the current
    -- input frame the first parameter to math_min is used, otherwise
    -- the second is used.
    local out_count =
        math_min(self.curr_trig.b_end - self.tot_count - start_index + 1,
                 in_count - start_index)

    assert(out_count > 0, string.format("Invalid output count (%s): " ..
        "trig time = %s, self.curr_trig.b_end = %s, tot_count = %s, " ..
        "start_index = %s",
        tostring(out_count), tostring(self.curr_trig.trig),
        tostring(self.curr_trig.b_end), tostring(self.tot_count),
        tostring(start_index)))

    assert(start_index >= 0 and start_index < in_count,
           string.format("invalid start_index (%s), trig time = %s, " ..
               "pre-count = %s, earliest burst = %s, " ..
               "input frames ts = %s",
               tostring(start_index), tostring(self.curr_trig.trig),
               tostring(self.pre_count),
               tostring(self.earliest_burst_start),
               tostring(self.chan_info[1].in_frame.ts)))

    -- fill output frames from start of pre-trigger history
    for i = 1, #self.chan_info do
        local infr = self.chan_info[i].in_frame
        local sample_type = utils.frame_data_type(infr.element_size)
        local iv = ffi_cast(sample_type, infr.buf)
        local outfr = Frame_create(out_count, infr.element_size)
        outfr.hdr.ts = infr.hdr.ts + start_index
        outfr.ts = infr.ts + start_index
        outfr.offset = ffi_cast("int64_t", outfr.ts - self.curr_trig.trig)
        local ov = ffi_cast(sample_type, outfr.buf)
        local tmp_index = start_index
        for j = 0, out_count - 1 do
            ov[j] = iv[tmp_index]
            tmp_index = tmp_index + 1
        end
        outfr.count = out_count
        if not utils.put_frame(outfr, self.chan_info[i].pipe_list) then
            Frame_delete(outfr)
        end
    end

    return out_count, start_index + out_count
end

-- Puts as much as possible of each input channel's remaining
-- pre-trigger history into output frames, copying the history
-- from saved input frames.  NOTE that the remaining burst samples
-- start at index 0, due to previous call of self:write_start_pretrig().
-- Returns:  the number of samples output
function mt:write_remaining_burst()
    local count = self.chan_info[1].in_frame.count
    local nbr_to_write =
        math_min(self.burst_size - self.burst_written, count)

    for i = 1, #self.chan_info do
        local infr = self.chan_info[i].in_frame
        local sample_type = utils.frame_data_type(infr.element_size)
        local iv = ffi_cast(sample_type, infr.buf)
        local outfr = Frame_create(nbr_to_write, infr.element_size)
        outfr.hdr.ts = infr.hdr.ts
        outfr.ts = infr.ts
        outfr.offset = ffi_cast("int64_t", outfr.ts - self.curr_trig.trig)
        local ov = ffi_cast(sample_type, outfr.buf)
        for j = 0, nbr_to_write - 1 do
            ov[j] = iv[j]
        end
        outfr.count = nbr_to_write
        if not utils.put_frame(outfr, self.chan_info[i].pipe_list) then
            Frame_delete(outfr)
        end
    end

    return nbr_to_write
end

-- Creates the SIE buffers (pipes and XML) required.
-- There will be one channel description per input channel.
function mt:sie_xml()
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        -- write XML for this SIE channel
        info.stanza.description = self.stanza.description or ""
        info.stanza.datamode_name = self.name
        info.stanza.sie_name = info.name .. '@' .. self.name
        info.stanza.somat_input_channel = info.stanza.name
        -- the following line makes the schema info reflect the
        -- input channel's data bit size (?should this be different?)
        self.stanza.nbr_data_bits =
            tonumber(info.data_type:match('^.-(%d+)$'))
        assert(self.stanza.nbr_data_bits and self.stanza.nbr_data_bits > 0,
            "Missing data bit size")
        local schema_name = sie_schema.get_dm_schema_name(self.stanza.category)
        local schema_info = sie_schema.get_schema_info(schema_name, self.stanza)
        schema_info.dims[2]['core:label'] = info.stanza.name
        schema_info.dims[2]['core:units'] = info.stanza.output_units
        local md = sie_utils.write_sie_channel(self.sw, schema_info,
            info.stanza)
        -- Attach one end of pipe to this module for this input node
        local p = Pipe_attach_pipe(info.pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        self.sw:new_buffer(self.sie_collect_idx + i - 1,
            md._sie_group, 8192,
            ffi_sizeof(info.data_type), 8, p, nil, sie_utils.BH_DATAMODE)
    end
end

-- Determines if there is an input frame for each input pipe.
-- Returns:  true if a full set of frames exists, false otherwise
function mt:have_frames()
    local have_frames = #self.chan_info > 0
    if not have_frames then
        return false  -- short circuit return!
    end

    for i = 1, #self.chan_info do
        have_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not have_frames then
            break  -- don't have a frame from this input pipe
        end
    end

    return have_frames
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- NOTE:  self:have_frames() should be called before this function
--     is called, which means we don't have to check here for the
--     existence of a frame for every input channel.
function mt:get_frames()
    for i = 1, #self.chan_info do
        self.chan_info[i].in_frame =
            Pipe_get_frame(self.chan_info[i].in_pipe)
        assert(self.chan_info[i].in_frame, "Pipe_get_frame() " ..
            "failed for channel " .. self.chan_info[i].name)
    end
end

-- Checks trigger frame for next trigger occurrence, starting
-- at given index.
-- Paramater:
--     'start_index' is where to start looking in trigger frame
--     'max' is where to stop looking in trigger frame (inclusive)
-- Returns:  index of last trigger frame sample processed, and true
--     if a trigger was found (false otherwise for second value)
function mt:look_for_trig(start_index, trig_frame)
    local trigger_found = false
    local max = trig_frame.count - 1
    local index = start_index
    local trig_vals = ffi_cast(utils.frame_data_type(
        trig_frame.element_size), trig_frame.buf)
    -- find next trigger condition in trigger frame
    for i = start_index, max do
        if self.trigger(self, trig_vals[i]) then
            index = i  -- NOTE that returned 'index' points at trigger
            trigger_found = true
            break
        end
    end
    if not trigger_found then
        -- have "processed" all input samples
        index = max
    end
    return index, trigger_found
end

-- Collects any triggers from all currently available trigger frames.
-- The collected triggers are put in self.trigger_history.
function mt:collect_trigs()
    local trig_frame = Pipe_get_frame(self.trig_pipe)
    while trig_frame and (trig_frame ~= nil) do
        local index = 0
        local trigger_found
        while index < trig_frame.count do
            -- each trigger frame sample must be passed through
            -- 'look_for_trig', which calls the configured trigger
            -- function, since some trigger types rely on the history
            -- of trigger channel samples
            index, trigger_found = self:look_for_trig(index, trig_frame)
            -- ignore any triggers occurring while processing a burst
            if self.trig_tot_count + index < self.newest_burst_start or
               self.trig_tot_count + index > self.newest_burst_end then
                if trigger_found then
                    local this_trig = self.trig_tot_count + index
                    local this_start =
                        math_max(self.newest_burst_end + 1,
                            this_trig - self.pre_count)
                    -- NOTE:  the trigger information is stored as
                    --   timestamps, NOT as indices into frames
                    local t_his =
                        {trig = this_trig,
                         b_start = this_start,
                         b_end = this_start + self.burst_size - 1}
                    table.insert(self.trigger_history, t_his)
                    self.newest_burst_start = t_his.b_start
                    self.newest_burst_end = t_his.b_end
                end
                index = index + 1
            else
                index = self.newest_burst_end - self.trig_tot_count + 1
            end
        end
        self.trig_tot_count = self.trig_tot_count + trig_frame.count
        Frame_delete(trig_frame)
        trig_frame = Pipe_get_frame(self.trig_pipe)
    end
end

-- Processes incoming frames, producing outgoing frames.
-- Outline of logic:
--     if we have frames from all input pipes and the trigger pipe
--         set 'next input index' to 0 -- (same variable used for all frames)
--         set 'count' to first input frame's count value
--         loop until 'next input index' >= 'count'
--             if not in "triggered" state
--                 check for a trigger condition in the trigger frame
--                 if trigger condition found
--                     set "triggered" state on
--                     if pre-trigger history required
--                         for each input frame
--                             copy pre-trigger history from the
--                                 pre-trigger buffer to an output frame
--                             empty the pre-trigger buffer
--                             put the output frame in outgoing pipe list.
--                             NOTE:  pre-trigger history is never
--                                        duplicated in output frames!
--                         increment 'burst used' count by pre-trigger output
--                         set 'next input index' to the trigger index value
--                 else if a pre-trigger history is required
--                     for each input frame
--                         put as many input samples as necessary in
--                             the pre-trigger history buffer
--                     -- done with these frames
--                     set 'next input index' to 'count'
--             end
--             if in "triggered" state
--                 for each input frame
--                     copy as many samples as necessary into an output frame
--                         starting from the 'next' index
--                     put the output frame in outgoing pipe list.
--                 set 'next' to index of first input sample that
--                     hasn't been copied
--                 if number of samples output in the current "burst" 
--                         is the requested size of the burst
--                     reset to "not triggered" state
--         end  -- frame sample loop
--         get next set of input and trigger frames, process them
function mt:run()
    local have_frames = false

    -- get all available triggers
    self:collect_trigs()

    -- if we don't have leftover input frames, check for presence
    -- of a new set (one from each input pipe)
    if not self.leftover_samples then
        have_frames = self:have_frames()
        if not have_frames then
            -- can't process any input frames yet
            return
        end
    end

    -- Process input frames.
    while self.leftover_samples or have_frames do
        if not self.leftover_samples then
            self:get_frames()
        end
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        It doesn't matter which input frame we use for
        --        obtaining frame count.
        local in_count = self.chan_info[1].in_frame.count
        while self.in_index < in_count do
            -- Execute the current state.
            if self.state == ST_LOOK_FOR_TRIG then
                self.curr_trig = self.trigger_history[1]
                table.remove(self.trigger_history, 1)
                if self.curr_trig then
                    self.state = ST_FIND_PRE_HISTORY_START
                else
                    -- prune unneeded input frame sets
                    self.earliest_burst_start =
                        self.trig_tot_count - self.pre_count
                    if self.tot_count + in_count <
                       self.earliest_burst_start then
                        -- skip rest of input frame
                        self.in_index = in_count
                    else
                        -- need to wait until we get more trigger
                        -- channel samples
                        self.leftover_samples = true
                        return
                    end
                end
            elseif self.state == ST_FIND_PRE_HISTORY_START then
                if self.curr_trig.b_start - self.tot_count < in_count then
                    self.state = ST_WRITE_PRE_HISTORY_START
                else
                    -- skip rest of input frame
                    self.in_index = in_count
                end
            elseif self.state == ST_WRITE_PRE_HISTORY_START then
                self.burst_written, self.in_index =
                    self:write_start_pretrig()
                if self.burst_written == 0 then
                    -- should have been able to write at least
                    -- one sample, i.e. the sample at trigger time
                    assert(false, string.format(
                        "write_start_pretrig failed, in_index = %s, " ..
                        "trig time = %s", tostring(self.in_index),
                        tostring(self.curr_trig.trig)))
                elseif self.burst_written < self.burst_size then
                    self.state = ST_FINISH_BURST
                else
                    -- burst was wholly contained in this set
                    -- of frames so finish burst (flush sie bufs)
                    -- and look for start of next burst
                    self.sw:flush_these(self.sie_collect_idx,
                        #self.chan_info, true)
                    self.burst_written = 0
                    self.state = ST_LOOK_FOR_TRIG 
                end
            elseif self.state == ST_FINISH_BURST then
                local nbr_written = self:write_remaining_burst()
                self.in_index = self.in_index + nbr_written
                self.burst_written = self.burst_written + nbr_written
                if self.burst_written >= self.burst_size then
                    -- finish burst (flush sie bufs) and look for next
                    self.sw:flush_these(self.sie_collect_idx,
                        #self.chan_info, true)
                    self.burst_written = 0
                    self.state = ST_LOOK_FOR_TRIG
                end
            end
        end
        self.leftover_samples = false
        self.in_index = 0
        self.tot_count = self.tot_count + in_count
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end
        have_frames = self:have_frames()
    end
end

-- Cleans up at test conclusion.
-- Paramater:
--     'sie_writer' is SIE writer object - NOT USED HERE
function mt:finalize(sie_writer)
-- Nothing to do here.
end


local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer for
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
--   meta_list --> list of all created-to-this-point nodes' stanzas,
--       indexed by channel name
-- Return values:
--   the new instance object
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    assert(first_sie_index, "Missing parameter 'first_sie_index'")
    assert(meta_list, "Missing parameter 'meta_list'")
    local burst_history = setmetatable({}, {__index = mt})
    burst_history.stanza = stanza
    -- don't need to call set_common_meta() since there are
    -- no modules downstream
    burst_history.sie_collect_idx = first_sie_index
    burst_history.chan_info = {}
    -- validate and process needed metadata
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1, "Burst History data mode " ..
        "must have at least one input channel")
    --     node name
    burst_history.name = stanza.name
    assert(burst_history.name, "Missing node name")
    --     sample rate
    for i = 1, #input_channels do
        local node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        burst_history.sample_rate = node_meta.sample_rate
        if burst_history.sample_rate then
            burst_history.real_sample_rate =
                rational:new(burst_history.sample_rate):tonumber()
            burst_history.sample_rate =
                math_floor(burst_history.real_sample_rate)
            if burst_history.sample_rate > 0 then
                break
            end
        end
    end
    assert(burst_history.sample_rate and burst_history.sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(burst_history.sample_rate)))
    --     pre- and post-trigger counts
    assert(stanza.burst_pre_trigger_time, "Missing pre-trigger time")
    local pre_trig_time = tonumber(stanza.burst_pre_trigger_time)
    assert(pre_trig_time >= 0, "Invalid pre-trigger time: \"" ..
        tostring(stanza.burst_pre_trigger_time) .. "\"")
    burst_history.pre_count = pre_trig_time * burst_history.sample_rate
    burst_history.pre_count = math_floor(burst_history.pre_count)
    assert(stanza.burst_post_trigger_time, "Missing post-trigger time")
    local post_trig_time = tonumber(stanza.burst_post_trigger_time)
    assert(post_trig_time >= 0, "Invalid post-trigger time: \"" ..
        tostring(stanza.burst_post_trigger_time) .. "\"")
    burst_history.post_count = post_trig_time * burst_history.sample_rate
    burst_history.post_count = math_floor(burst_history.post_count)
    burst_history.burst_size = burst_history.pre_count +
        burst_history.post_count + 1
    --     trigger
    local trigger_mode = stanza.trigger_mode
    assert(trigger_mode and
        (trigger_mode == "when_true" or
         trigger_mode == "on_false_to_true_edge"),
        "Missing or invalid trigger mode: \"" ..
        tostring(trigger_mode) .. "\"")
    local trig_chan = stanza.trigger_channel
        assert(trig_chan, "Burst History requires a trigger channel")
    local trig_node = node_list[trig_chan]
    assert(trig_node, "Node for trigger channel " ..
        tostring(trig_chan) .. " not found")
    burst_history.trig_pipe = Pipe_attach_pipe(trig_node.pipe_list)
    assert(burst_history.trig_pipe and burst_history.trig_pipe ~= nil,
        "failed to create trigger pipe")
    burst_history.trigger = trigger_mode == "when_true" and
        burst_history.gate_trigger or burst_history.edge_trigger
    --     create input pipe structures
    local input_node
    local node_meta
    for i = 1, #input_channels do
        input_node = node_list[input_channels[i]]
        assert(input_node, "Node for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        --     data type
        local data_type = node_meta.data_type
        assert(data_type, "Missing data_type for \"" ..
            tostring(input_channels[i]) .. "\"")
        -- channel info array
        burst_history.chan_info[i] =
            {name = input_channels[i],
             -- in_node = input_node,
             in_pipe = Pipe_attach_pipe(input_node.pipe_list),
             pipe_list = Pipe_new_pipe_list(),
             stanza = node_meta,
             data_type = data_type}
        assert(burst_history.chan_info[i].in_pipe and
            burst_history.chan_info[i].in_pipe ~= nil,
            "failed to create input pipe")
        assert(burst_history.chan_info[i].pipe_list and
            burst_history.chan_info[i].pipe_list ~= nil,
            "failed to create pipe list")
    end

    -- initialize volatile values
    burst_history.burst_written = 0
    burst_history.newest_burst_start = -1
    -- the following is set to -1 so that edge triggers starting
    -- high will be captured
    burst_history.newest_burst_end = -1
    burst_history.in_index = 0
    burst_history.leftover_samples = false
    burst_history.state = ST_LOOK_FOR_TRIG
    burst_history.trigger_history = {}
    burst_history.curr_trig = nil
    burst_history.prev_trig_on = false
    burst_history.earliest_burst_start = nil

    -- used for input frames' time stamps (just a sample count
    -- for this module)
    burst_history.tot_count = 0
    -- used for keeping track of trigger times for trigger
    -- channel frames
    burst_history.trig_tot_count = 0

    return burst_history, #input_channels  -- 1 sie buf per input chan
end

return m
