require 'strict'

-- time_at_level_1d.lua - Implements one dimensional histogram
-- data mode.

local cjson = require 'cjson'
local ffi = require 'ffi'
local C = ffi.C 
local Pipe = require 'test_engine/pipe'
local Frame = require 'test_engine/frame'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'

local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_sizeof = ffi.sizeof
local ffi_string = ffi.string

local Pipe_attach_pipe = Pipe.attach_pipe
local Pipe_get_frame = Pipe.get_frame
local Pipe_has_frames = Pipe.has_frames
local Frame_delete = Frame.delete

local math_floor = math.floor
local math_min = math.min
local math_max = math.max

local ctor_cache = {}

local mt = {}

local config = require 'system_config'
local libte = ffi.load(config.test_engine_library)
local libteFlushFrames = libte.SIEBuffer_flush_frames
local libteFlushThese = libte.SIEWriter_flush_these
local libteFlush = libte.SIEBuffer_flush
local libteAmountWritten = libte.SIEWriter_amount_written
local libteAppend = libte.SIEBuffer_append
local libteEnq = libte.SIEBuffer_enq

-- function mt:run_evenly_divided_sequential(in_buf, count)
-- function mt:run_evenly_divided_message(in_buf, count)

function mt:run_evenly_divided_generic(num)
    local hist = self.hist[num]
    local f = Pipe_get_frame(hist.in_pipe)
    local bin
    local t
    while f ~= nil do
        local sptr = ffi_cast(f.element_size == 4 and 'float *' or 'double *',
            f.buf)
        for i = 0, f.count - 1 do
            t = sptr[i]
            bin = math_floor((t - hist.lo_bound) * hist.bound_scale) + 1
            bin = math_max(0, bin)
            bin = math_min(hist.bin_count - 1, bin)
            hist.bins[bin] = hist.bins[bin] + 1
            if hist.bins[bin] == 0 then
                hist.bins[bin] = bit.bnot(hist.bins[bin])
            end
        end
        Frame_delete(f)
        f = Pipe_get_frame(hist.in_pipe)
    end
end

function mt:run()
    local num = self.num_hist - 1
    for i = 0, num do
        if not Pipe_has_frames(self.hist[i].in_pipe) then return end
    end
    for i = 0, num do
        self:run_evenly_divided_generic(i)
    end
end

function mt:finalize(sie_writer)
    if self.final_called ~= 0 then return end
    self.final_called = 1
    local num = self.num_hist - 1
    local bin_bits = 64
    local hdr = ffi.new('uint32_t[2]')
    hdr[0] = 3
    for i = 0, num do
        local hist = self.hist[i]
        local buf = sie_writer.bufs[hist.sie_idx]
        hdr[1] = hist.bin_count
        libteEnq(buf, hdr, ffi_sizeof('uint32_t') * 2)
        libteEnq(buf, hist.bins, hdr[1] * bin_bits / 8)
        libteFlush(buf, 1)
    end
end

-- Writes the given data to the SIE file using sie_writer's
-- "scratch" sie buf.
function mt:scratch_write(sw, hist)
        local bounds_bits = 64  -- all bounds use 64 bits (for now!)
        sw:new_buffer(sw.scratch_buf_idx, nil,
            (hist.bin_count + 1) * bounds_bits / 8, bounds_bits / 8, 0)
        local bbuf = sw.bufs[sw.scratch_buf_idx]
        local ptr = ffi.cast('const char *', hist.bounds)
        libteAppend(bbuf, ptr, (hist.bin_count + 1) * bounds_bits / 8)
        libteFlush(bbuf, 0)
    return bbuf.group
end

-- Create all of the SIE buffers needed by this sink
function mt:setup(sie_writer, is_restart, dm_tab)
    for i = 0, self.num_hist - 1 do
        local hist = self.hist[i]

        -- get channel id for the appropriate generic histogram
        -- lookup table channel, writing it if necessary
        local bin_tab = {}
        bin_tab._base_id =
            sie_utils.ensure_sie_bounds_channel(sie_writer, 64)
        bin_tab._dm = ffi.string(hist.hist_name)
        -- write bin min/max values, which the histogram SIE channel
        -- will use as a lookup table
        bin_tab._sie_group = self:scratch_write(sie_writer, hist)
        --     write the channel which will use the lookup table
        local index_chan_id =
            sie_utils.write_bin_channel(sie_writer, bin_tab)
        -- write XML for this SIE channel
        local sch_tab = {}
        sch_tab.description = ffi.string(hist.description) or ""
        sch_tab.datamode_name = ffi.string(hist.hist_name)
        sch_tab.sie_name = ffi.string(hist.name) .. '@' ..
            sch_tab.datamode_name
        sch_tab.somat_input_channel = ffi.string(hist.name)
        sch_tab.category = ffi.string(hist.category)
        sch_tab.data_type = dm_tab.data_type
        sch_tab.sample_rate = hist.sample_rate
        sch_tab.nbr_data_bits = 64
        sch_tab.name = ffi_string(hist.name)
        sch_tab.units = ffi_string(hist.units)
        sch_tab.low_bound = hist.lo_bound
        sch_tab.high_bound = hist.hi_bound
        local schema_name = sie_schema.get_dm_schema_name(sch_tab.category)
        local schema_info = sie_schema.get_schema_info(schema_name, sch_tab)
        -- assign index channel ids
        schema_info.dims[2]._index_ch = index_chan_id
        schema_info.dims[3]._index_ch = index_chan_id
        local st_tab = {}
        st_tab.sie_name = sch_tab.sie_name
        st_tab.category = sch_tab.category
        st_tab.data_type = dm_tab.data_type
        st_tab.datamode_name = sch_tab.datamode_name
        st_tab.connector = ""
        st_tab.somat_input_channel = sch_tab.somat_input_channel
        st_tab.description = sch_tab.description
        st_tab.sample_rate = sch_tab.sample_rate
        local md = sie_utils.write_sie_channel(sie_writer, schema_info,
            st_tab)
        -- Create the SIE buffer and attach the pipe to it
        local dims = 1
        local bin_bits = 64
        sie_writer:new_buffer(hist.sie_idx, md._sie_group,
            hist.bin_count * bin_bits / 8, bin_bits / 8,
            ffi_sizeof('uint64_t') * (dims + 1))

        -- Ensure there is file space to write the channel data
        -- at end of test.
        sie_writer:reserve_space(
            ((hist.bin_count + 2) * (bin_bits / 8)) +   -- bin space
            (ffi_sizeof('uint32_t') * 2) + -- header space
            (ffi_sizeof('uint32_t') * 5), -- SIE overheader
            function() self:finalize(sie_writer) end)
    end
end

local function meta_ctor(t)
    if not ctor_cache[t] then
        local tal_type = string.format('struct %s_tal', t)
        local spec = string.gsub([[
            struct $t_hist_sie {
                $t dimcount; // KLUDGE placeholder
                $t bin_count; // KLUDGE placeholder
            };
            struct $t_hist {
                int             bin_count;
                $t              *bins;
                double          *bounds;
                /* For evenly_divided special case */
                double          lo_bound;
                double          hi_bound;
                double          bound_scale;
                double          bound_span;
                int             sie_idx;
                char            *name;
                char            *description;
                char            *hist_name;
                char            *category;
                char            *units;
                double          sample_rate;
                struct pipe     *in_pipe;
            };
            struct $t_tal {
                int             num_hist;
                int             final_called;
                struct $t_hist  *hist;
            };
        ]], '%$(%w+)', {t = t})
        ffi.cdef(spec)
        ctor_cache[t] = ffi.metatype(tal_type, {__index = mt})
    end
    return ctor_cache[t]
end

require 'middleclass'

local m = class('time_at_level_1d')

function m:setup(sie_writer, is_restart)
    self.tal:setup(sie_writer, is_restart, self.dm_tab)
end

function m:run()
    if self.trigger_module then self.trigger_module:run() end
    self.tal:run()
end

--[[
function m:finalize(sie_writer)
    self.tal:finalize(sie_writer)
end
]]

function m:initialize(stanza, node_list, sie_idx, stanza_list)
    local bin_type = stanza.bin_type or 'uint64_t'
    local ctor = meta_ctor(bin_type)

    self.sie_buffers_needed = #stanza.input_channels
    self.dm_tab = {}
    self.dm_tab.data_type = ""  -- data_type irrelevant for histogram

    self.tal = ctor()
    self.tal.num_hist = #stanza.input_channels
    self.tal.hist = ffi_cast(string.format('struct %s_hist *', bin_type),
        C.calloc(self.tal.num_hist,
        ffi_sizeof(string.format('struct %s_hist', bin_type))))
    local objs = self.tal.hist

    -- If triggered then instantiate a new trigger module
    self.trigger_module = stanza.trigger_mode ~= 'always_on' and
        require('test_engine/modules/trigger/' .. stanza.trigger_mode):new(
            stanza.input_channels, stanza.trigger_channel, node_list)

    for i = 0, #stanza.input_channels - 1 do
        local hb = stanza.histogram_bins[i + 1]
        local obj = objs[i]
        local ic = stanza.input_channels[i + 1]

        obj.sie_idx = sie_idx + i

        -- Add 2 extra bins for underflow and overflow
        obj.bin_count = hb.number_of_bins + 2
        obj.bins = ffi_cast(string.format('%s *', bin_type),
            C.calloc(obj.bin_count, ffi_sizeof(bin_type)))
        obj.bounds = ffi_cast('double *',
            C.calloc(obj.bin_count + 1, ffi_sizeof('double')))
        obj.name = ffi_cast('char *',
            C.calloc(ic:len() + 1, ffi_sizeof('char')))
        ffi_copy(obj.name, ic)
        local units = stanza_list[ic].output_units or ''
        obj.units = ffi_cast('char *', C.calloc(units:len() + 1,
            ffi_sizeof('char')))
        ffi_copy(obj.units, units)

        local desc = stanza_list[ic].description or ic or
            'No channel description'

        obj.description = ffi_cast('char *', C.calloc(desc:len() + 1,
            ffi_sizeof('char')))

        ffi_copy(obj.description, desc)

        obj.hist_name = ffi_cast('char *',
            C.calloc(stanza.name:len() + 1, ffi_sizeof('char')))
        ffi_copy(obj.hist_name, stanza.name)

        obj.category = ffi_cast('char *',
            C.calloc(stanza.category:len() + 1, ffi_sizeof('char')))
        ffi_copy(obj.category, stanza.category)

        if not self.trigger_module then
            -- Not triggered...attach directly to input channel
            obj.in_pipe = Pipe_attach_pipe(node_list[ic].pipe_list)
        else
            -- Triggered...attach to gated channel
            obj.in_pipe =
                Pipe_attach_pipe(self.trigger_module.chan_info[i + 1].pipe_list)
        end

        obj.sample_rate = rational(stanza_list[ic].sample_rate):tonumber()

        -- Evenly divided bins
        if hb.type == 'evenly_divided' then
            obj.lo_bound = hb.low_bound
            obj.hi_bound = hb.high_bound
            obj.bound_span = obj.hi_bound - obj.lo_bound
            obj.bound_scale = (obj.bin_count - 2) / obj.bound_span
            obj.bounds[0] = -1/0
            obj.bounds[obj.bin_count] = 1/0
            for i = 0, obj.bin_count - 2 do
                obj.bounds[i + 1] = obj.lo_bound + i / obj.bound_scale
            end
    
        -- User defined bins
        elseif hb.type == 'user_defined' then
            error("TAL Datamode with user defined bins is not yet supported")
    
        -- Unknown bin type
        else
            error(string.format("TAL with unknown bin type: %s", hb.type))
        end
    end

    return self
end

return m
