require 'strict'

-- trig_common.lua - Provides a framework for triggers
-- used by a data mode.

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local msg_can = require 'message_can'
local sample_types = require 'sample_types'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local log = require 'log'

local ffi_cast = ffi.cast
local ffi_new = ffi.new
local ffi_sizeof = ffi.sizeof
local ffi_string = ffi.string

local Frame_create = frame.create
local Frame_delete = frame.delete
local Pipe_add_frame = pipe.add_frame
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Pipe_flush_frames = pipe.flush_frames

local mt = {}

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mt:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = #self.chan_info > 0
    if not got_frames then
        return false  -- short circuit return!
    end

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    got_frames = got_frames and Pipe_has_frames(self.trig_pipe)
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame =
                Pipe_get_frame(self.chan_info[i].in_pipe)
        end
        self.trig_frame = Pipe_get_frame(self.trig_pipe)
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
        self.trig_frame = nil
    end
    return got_frames
end

-- Filters input frame(s) based on the value of the trigger frame.
--     If a trigger frame sample is not zero the corresponding
--     samples from the input frames are copied to their output
--     frames.  If the trigger sample is zero, the corresponding
--     samples from the input frames are ignored.
function mt:run()
    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- can't process any input frames yet
        return
    end

    -- Process each frame in trigger and input pipe(s).
    while have_all_frames do
    
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        So it doesn't matter which input frame we use for
        --        obtaining frame count.
        local count = self.chan_info[1].in_frame.count

        -- get input and trigger frames data "pointers"
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local sample_type = utils.frame_data_type(infr.element_size)
            self.chan_info[i].invals =
                ffi_cast(sample_type, infr.buf)
        end
        local trig_vals =
            ffi_cast(utils.frame_data_type(self.trig_frame.element_size),
                     self.trig_frame.buf)
        -- create output frames and get data "pointers"
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local outfr = Frame_create(infr.maxsize, infr.element_size)
            local sample_type = utils.frame_data_type(
                self.chan_info[i].in_frame.element_size)
            outfr.hdr.ts = infr.hdr.ts
            outfr.ts = infr.ts
            self.chan_info[i].out_frame = outfr
            self.chan_info[i].outvals =
                ffi_cast(sample_type, outfr.buf)
        end

        -- process each trigger and input sample
        local out_index = 0
        for in_index = 0, count - 1 do
            if self.core_trig:triggered(trig_vals[in_index]) then
                for findex = 1, #self.chan_info do
                    self.chan_info[findex].outvals[out_index] =
                        self.chan_info[findex].invals[in_index]
                end
                out_index = out_index + 1
            end
        end

        -- only send output frames if they have some data
        if out_index > 0 then
            -- set count for each output frame and put it in pipe;
            -- also create a new output frame for each input channel
            for i = 1, #self.chan_info do
                local of = self.chan_info[i].out_frame
                of.count = out_index
                if not utils.put_frame(of, self.chan_info[i].pipe_list) then
                    Frame_delete(of)
                end
            end
        else
            -- delete output frames, since we aren't relying on the
            -- next output frames being the same size
            for i = 1, #self.chan_info do
                Frame_delete(self.chan_info[i].out_frame)
            end
        end

        -- don't need the processed input frames any longer
        Frame_delete(self.trig_frame)
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end

        have_all_frames = self:get_frames()
    end
end

-------------------------------
-- Message Triggering functions

local function search(low, high, list, ts)
    if low > high then return false end
    local mid = math.floor((low+high)/2)
    if list[mid].start > ts then
        return search(low,mid-1,list,ts)
    end
    if list[mid].stop < ts then
        return search(mid+1,high,list,ts)
    end
    return (list[mid].start <= ts and ts <= list[mid].stop)
end

-- Get state of trigger at time ts, implements a binary search on known periods.
function mt:get_triggered(ts)
    return search(1, #self.trigger_periods, self.trigger_periods,ts)
end

-- With CAN, input frames can hold multiple CAN messages.
-- In cases where not all CAN messages can be processed, a remainder is noted.
function mt:process_can_frame(chan)
    local infr = chan.in_frame
    chan.out_frame = Frame_create(infr.maxsize, infr.element_size)
    local outfr = chan.out_frame
    chan.outvals = ffi_cast("message_can *", outfr.buf)
    local iv = chan.invals
    local ov = chan.outvals
    local out_index = 0
    local start = infr.count - (chan.frame_remainder or infr.count)
    chan.frame_remainder = nil
    chan.waiting_on_its_pipe = nil
    
    for i = start, infr.count-1 do
        chan.last_ts = iv[i].timestamp[0]
        if chan.last_ts > self.trigger_periods.last then
            chan.frame_remainder = infr.count - i
            break
        elseif self:get_triggered(chan.last_ts) then
            ov[out_index] = ffi_new('message_can', iv[i])
            out_index = out_index + 1
        end
    end
    
    if out_index > 0 then
        outfr.count = out_index
        if not utils.put_frame(outfr, chan.pipe_list) then
            chan.outvals = nil
            Frame_delete(outfr)
            chan.out_frame = nil
        end
    else
        chan.outvals = nil
        Frame_delete(outfr)
        chan.out_frame = nil
    end
    
    if not chan.frame_remainder then
        chan.invals = nil
        Frame_delete(infr)
        chan.in_frame = nil
    end
end

-- get trigger frame if available
function mt:have_trigger_frame()
    if Pipe_has_frames(self.trig_pipe) then
        self.trig_frame = Pipe_get_frame(self.trig_pipe)
        return true
    else
        self.trig_frame = nil
        return false
    end
end

-- Used to maintain size of periods data
function mt:remove_periods_before(ts)
    if #self.trigger_periods == 0 then
        return
    end
    local list = self.trigger_periods
    local function search(low, high)
        if low > high then return low end
        local mid = math.floor((low+high)/2)
        if list[mid].start > ts then
            return search(low,mid-1)
        end
        if list[mid].stop < ts then
            return search(mid+1,high)
        end
        return mid
    end
    local keep = search(1, #self.trigger_periods)
    for i = 1, keep-1 do
        local t = table.remove(self.trigger_periods, 1)
    end
end

function mt:compress_trigger_data(trig_vals, find_start, last_sample_ts)
    local TICKSEC = ffi_new('int64_t', 40000000)
    local rate    = ffi_new('int64_t',tonumber(self.trig_node.sample_rate))
    local persec  = (TICKSEC / rate)
    local ts      = ffi_new('int64_t',tonumber(self.trig_frame.hdr.ts)*persec)
    
    for i = 0, tonumber(self.trig_frame.count)-1 do
        local triggered = self.core_trig:triggered(trig_vals[i])
        if triggered and find_start then
            -- Trigger marked continuing through frame until disproven.
            self.trigger_periods[#self.trigger_periods+1] =
                { start = ts + i * persec,
                  stop  = last_sample_ts }
            find_start = false
        elseif not triggered and not find_start then
            self.trigger_periods[#self.trigger_periods].stop=ts+i*persec
            find_start = true
        end
    end
end

function mt:process_trig_frame()
    if not self:have_trigger_frame() then return false end
    assert(self.trig_node.sample_rate, "Sample rate for trigger channel \"" ..
        tostring(self.trig_node.name) .. "\" not found")

    local TICKSEC = ffi_new('int64_t', 40000000)
    local rate    = ffi_new('int64_t',tonumber(self.trig_node.sample_rate))
    local persec  = (TICKSEC / rate)
    local count   = tonumber(self.trig_frame.count)
    local ts      = ffi_new('int64_t',tonumber(self.trig_frame.hdr.ts)*persec)
    local last_sample_ts = ffi_new('int64_t', ts + count*persec -1)
    local trig_vals      = ffi_cast(utils.frame_data_type(
        self.trig_frame.element_size), self.trig_frame.buf)
    local find_start = true
    -- If previous frame's last sample and this first sample are
    -- triggered, extend the most recent trigger period.
    if #self.trigger_periods > 0 and
            self.trigger_periods[#self.trigger_periods].stop == ts-1 and
            self.core_trig:triggered(trig_vals[0]) then
        self.trigger_periods[#self.trigger_periods].stop = last_sample_ts
        find_start = false
    end
    self.trigger_periods.last = last_sample_ts
    
    -- Remove unneeded trigger periods from list
    local huge = ffi_new('int64_t', math.huge)    
    local oldest_ts = huge
    for i = 1, #self.chan_info do
        if self.chan_info[i].last_ts < oldest_ts then
            oldest_ts = self.chan_info[i].last_ts
        end
    end
    if oldest_ts < huge then
        self:remove_periods_before(oldest_ts)
    end
    
    self:compress_trigger_data(trig_vals, find_start, last_sample_ts)
    Frame_delete(self.trig_frame)
    return true
end

function mt:have_frame(chan)
    local got_frames = Pipe_has_frames(chan.in_pipe)
    if got_frames then
        chan.in_frame = Pipe_get_frame(chan.in_pipe)
    end
    return got_frames
end

-- Process each chan until its frames are N/A or waiting on trigger data
function mt:process_all_messages_from_channel(chan)
    assert(chan.data_type, 'Missing data_type for "'..tostring(chan.name)..'"')
    while chan.last_ts <= self.trigger_periods.last do
        if not chan.frame_remainder and not self:have_frame(chan) then
            chan.waiting_on_its_pipe = true
            break
        end
        chan.waiting_on_its_pipe = nil
        local infr = chan.in_frame
        local sample_type = chan.data_type == "message_can" and
            "message_can *" or "message_raw_buf *"
        chan.invals = ffi_cast(sample_type, infr.buf)

        -- message_raw - single message, put in out_pipe unaltered
        if chan.data_type == "message_raw" then
            chan.frame_remainder = nil --RAW's is a flag, CAN's is index
            chan.last_ts = chan.invals[0].timestamp[0]
            if chan.last_ts > self.trigger_periods.last then
                chan.frame_remainder = true
            elseif self:get_triggered(chan.last_ts) and
                utils.put_frame(infr, chan.pipe_list) then
                -- If put_frame fails, infr is deleted in else case.
            else
                Frame_delete(infr)
                chan.in_frame = nil
            end
        -- message_can - may have multiple messages per frame
        elseif chan.data_type == "message_can" then
            self:process_can_frame(chan)
        end
        if chan.last_ts > self.trigger_periods.last and
                not self:process_trig_frame() then
            break
        end
    end
end

function mt:process_messages()
    if #self.chan_info == 0 then return end

    local all_channels_waiting = false
    while self:process_trig_frame() and not all_channels_waiting do
        for i = 1, #self.chan_info do
            self:process_all_messages_from_channel(self.chan_info[i])
        end

        all_channels_waiting = true
        for i = 1, #self.chan_info do
            all_channels_waiting = all_channels_waiting and
                self.chan_info[i].waiting_on_its_pipe
        end
    end
    collectgarbage('collect')
end

local m = {}

-- Creates a new trigger object.
-- Parameters:
--   'in_chans'  input channel names
--   'trig_chan' trigger channel name
--   'node_list' list of previously instantiated channel nodes
-- Returns:
--   the new instance object
function m:new(in_chans, trig_chan, node_list, core_trig)
    assert(in_chans, "missing first parameter")
    assert(trig_chan, "missing second parameter")
    assert(node_list, "missing third parameter")
    assert(core_trig, "missing core trigger object parameter")

    local trig_common = setmetatable({}, {__index = mt})

    -- get input trigger pipe
    local trig_node = node_list[trig_chan]
    assert(trig_node, "Node for trigger channel " ..
        tostring(trig_chan) .. " not found")
    trig_common.trig_node = trig_node
    trig_common.trig_pipe = Pipe_attach_pipe(trig_node.pipe_list)
    assert(trig_common.trig_pipe and trig_common.trig_pipe ~= nil,
        "failed to create trigger pipe")

    -- set info for each input channel
    trig_common.chan_info = {}
    for i = 1, #in_chans do
        local input_node = node_list[in_chans[i]]
        assert(input_node, "Node for Input Channel \"" ..
            tostring(in_chans[i]) .. "\" not found")

        trig_common.chan_info[i] =
            {name = in_chans[i],
             data_type = input_node.buf_type and 
                ffi_string(input_node.buf_type),            -- used for messages
             last_ts = ffi_new('int64_t', 0),               -- used for messages
             in_pipe = Pipe_attach_pipe(input_node.pipe_list),
             pipe_list = Pipe_new_pipe_list()}
        assert(trig_common.chan_info[i].in_pipe and
            trig_common.chan_info[i].in_pipe ~= nil,
            "failed to create input pipe")
        assert(trig_common.chan_info[i].pipe_list and
            trig_common.chan_info[i].pipe_list ~= nil,
            "failed to create pipe list")
    end

    trig_common.core_trig = core_trig
    trig_common.trigger_periods = {}                        -- used for messages
    trig_common.trigger_on = false
    return trig_common
end

return m
