require 'strict'

-- simple_trigger.lua - Implements a basic trigger for use by
-- computed channels.
--

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local Frame_delete = frame.delete
local Pipe_get_frame = pipe.get_frame

local mt = {}

-- Gets the next trigger frame, storing the reference internally
-- to the trigger object.
function mt:get_frame()
    assert(not self.f, "ERROR:  a trigger frame is already stored")
    self.f = Pipe_get_frame(self.in_pipe)
    assert(self.f, "ERROR:  cannot get the next trigger frame")
    self.in_vals =
        ffi_cast(utils.frame_data_type(self.f.element_size), self.f.buf)
end

-- Deletes the currently stored trigger frame.
function mt:delete_frame()
    if self.f then
        Frame_delete(self.f)
        self.f = nil
    end
end

-- State definitions for 'after_true_to_false_edge' mode.
local STATE_NOT_ARMED  = 0
local STATE_ARMED      = 1
local STATE_TRIG_FIRE  = 2

-- Determines whether the trigger sample value at the given index
-- has caused the trigger to be on or off.  The three possible
-- trigger modes are:  "when_true", "on_false_to_true_edge", and
-- "after_true_to_false_edge".
--
-- NOTE:  This function considers each call to be for a different
--        trigger input value.  This is important for modes that
--        keep state information, which at this point is only the
--        'after_true_to_false_edge' mode.
--
-- Parameter:
--     'index'  the offset into the current trigger frame for
--              the sample to be tested
-- Returns:  'true' if trigger is on, false otherwise
function mt:triggered(index)
    if self.mode == 'when_true' then
        return self.in_vals[index] ~= 0
    elseif self.mode == 'on_false_to_true_edge' then
        local retval = self.prev_false and self.in_vals[index] ~= 0
        self.prev_false = self.in_vals[index] == 0
        return retval
    else  -- is 'after_true_to_false_edge'
          -- (note that this triggers AFTER the edge)
        local retval = false
        if self.after_true_to_false_state == STATE_NOT_ARMED then
            -- check for "arming" of trigger
            if self.in_vals[index] ~= 0 then
                self.after_true_to_false_state = STATE_ARMED
            end
        elseif self.after_true_to_false_state == STATE_ARMED then
            -- check for start of "triggering" condition
            if self.in_vals[index] == 0 then
                self.after_true_to_false_state = STATE_TRIG_FIRE
            end
        else  -- self.after_true_to_false_state == STATE_TRIG_FIRE
            -- delay accomplished, return "triggered" and reset trigger
            retval = true
            self.after_true_to_false_state = STATE_NOT_ARMED 
        end
        return retval
    end
end


local m = {}

-- Creates a new "simple trigger" object.
-- Parameters:
--   'in_pipe'  input frame pipe
--   'mode'     trigger mode
-- Returns:
--   the new instance object
function m:new(in_pipe, mode)
    assert(in_pipe and in_pipe ~= nil, "missing first parameter")
    assert(mode, "missing second parameter")

    local trig_obj = setmetatable({}, {__index = mt})
    trig_obj.in_pipe = in_pipe
    trig_obj.mode = mode
    trig_obj.after_true_to_false_state = STATE_NOT_ARMED
    trig_obj.f = nil

    return trig_obj
end

return m
