local ffi = require 'ffi'
local C = ffi.C
local quorum_latch = require 'quorum_latch'
local ev = require 'ev'
local config = require 'system_config'

ffi.cdef[[
enum {
    SIGPWR = 30,
    SIGTERM = 15,
    SIGQUIT = 3,
    SIGSTOP = 19,
};

typedef int     pid_t;

int kill(pid_t pid, int sig);
]]

-- Function used to stop tests in a normal shutdown situation.  The callback is
-- called after all tests have been successfully stopped.
local function killall_tests_normal(cb, pid_list, timeout)
    local ql = quorum_latch:new(#pid_list)

    -- Kill all of the tests running on the system
    for i, pid in pairs(pid_list) do
        C.kill(ffi.new('pid_t', tonumber(pid)), C.SIGQUIT)
    end

    -- Adjust timeout by rate at which the timer is running.
    timeout = timeout * 10

    -- Timer used to monitor for when all of the tests die.
    ev.Timer.new(function(l, t, e)
            -- If we've waited long enough, use SIGPWR to kill off the tests
            -- that remain and call the callback.
            if timeout <= 0 then
                for i, pid in pairs(pid_list) do
                    local p = ffi.new('pid_t', tonumber(pid))
                    C.kill(p, C.SIGPWR)
                end
                t:stop(l)
                if cb then cb('timeout waiting for running tests to stop') end
                return
            end

            -- Check to see which tests have exited.
            for i, pid in pairs(pid_list) do
                local p = ffi.new('pid_t', tonumber(pid))
                if C.kill(p, 0) < 0 then
                    if ql:set_vote(i) then
                        t:stop(l)
                        if cb then cb('all running tests stopped') end
                    end
                end
            end
            timeout = timeout - 1
        end, .1, .1):start(ev.Loop.default)
end

-- Function used to kill tests in an emergency stop situation
local function killall_tests_emergency(cb, pid_list)
    for _, pid in ipairs(pid_list) do
        C.kill(ffi.new('pid_t', tonumber(pid)), C.SIGPWR)
    end
    if cb then cb('emergency stop complete') end
end

-- Function used to stop tests
local function killall_tests(emergency, cb, timeout)
    local plist = io.popen('ps -o pid,group | grep acq')
    if plist then
        local pid_list = {}
        for p in plist:lines() do
            local pid = p:match('(%d+) ')
            if pid then
                pid_list[#pid_list + 1] = pid
            end
        end
        plist:close()
        if #pid_list > 0 then
            if emergency then
                killall_tests_emergency(cb, pid_list)
            else
                killall_tests_normal(cb, pid_list, timeout or 20)
            end
        elseif cb then
            cb('No running tests to stop')
        end
    elseif cb then
        cb('No running tests to stop')
    end
end

return killall_tests
